//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

import dec.trek.*;
import java.io.*;
import java.util.Vector;

//
//
// Jtu -- Utility methods for JTrek console apps
//
//
final class Jtu
{
	static String outSpec;	// Filespec for report file. Null means System.out

	private static final String spaces =
		"                                                            ";

	// Watch item types.
	// First method entry/exit related stuff
	public static final int wit_first = RunLog.ITEM_EXTEND;
	public static final int wit_enter = wit_first + 0;
	public static final int wit_exit = wit_first + 1;
	public static final int wit_exitvalue = wit_first + 2;
	public static final int wit_arg = wit_first + 3;
	public static final int wit_stack = wit_first + 4;
	public static final int wit_thread = wit_first + 5;

	// Tracing stmts
	public static final int wit_action = wit_first + 10;
	public static final int wit_if = wit_action + 0;
	public static final int wit_stmt = wit_action + 1;
	public static final int wit_switch = wit_action + 2;
	public static final int wit_while = wit_action + 3;

	// Values
	public static final int wit_anyval = wit_action + 8;
	public static final int wit_element = wit_anyval + 0;
	public static final int wit_value = wit_anyval + 1;

	// Find item types
	public static final int fit_first = wit_anyval + 8;
	public static final int fit_class = fit_first + 0;
	public static final int fit_field = fit_first + 1;
	public static final int fit_inmethod = fit_first + 2;
	public static final int fit_method = fit_first + 3;
	public static final int fit_ref = fit_first + 4;
										// Stmt with match
	public static final int fit_refnear = fit_first + 5;
										// Stmt near a match
	public static final int fit_local = fit_first + 6;
										// Match of code-less local def
	public static final int fit_separator = fit_first + 7;
	public static final int fit_reflocal = fit_first + 8;
										// Stmt with local-defining assign


	// A utility method to allow search-string to involve simple wildcarding.
	// For more advanced stuff, use one of the publicly available regular
	// expression packages for java.
	// candidate = the string to lookup (presumaby name from program)
	// wild = the pattern being looked for. May have * at start and/or end
	// exact = true says do case-sensitive match
	private static boolean equalsWild(String candidate, String wild, boolean exact)
	{
		// Lookup for cases where start is wild
		boolean wildstart = wild.startsWith("*");
		boolean wildend = wild.endsWith("*");
		int wildlen = wild.length();

		// The anywhere match case cant be done with regionMatches
		if (wildstart && wildend)
		{
			if (wildlen==1) return true;		// wild is simply *
			wild = wild.substring(1, wildlen-1);
			if (exact)
				return candidate.indexOf(wild) >= 0;
			else return candidate.toLowerCase().indexOf(wild.toLowerCase()) >= 0;
		}

		// Other cases lookup wild at specific point in candidate
		int candlen = candidate.length();
		int candpos = 0;					// Presume no-wild values
		int wildpos = 0;
		if (!wildstart && !wildend)
		{
			if (candlen!=wildlen)
				return false;					// Simple match requires same len
		}
		else if (wildstart)
		{
			wildpos = 1;						// Exclude leading *
			wildlen -= 1;
			candpos = candlen - wildlen;		// Compare end of candidate
		}
		else wildlen -= 1;						// Exclude trailing *

		// If left with more specific chars in wild than in candidate,
		// cant match. Else compare the two regions.
		if (candlen<wildlen)
			return false;
		return candidate.regionMatches(!exact, candpos, wild, wildpos, wildlen);
	}

	// Process one of the common args if present
	static int getCmdLineArg(String argv[], int i)
	{
		if (argv[i].equals("-of"))
		{
			if (i>argv.length-2)
				throw new RuntimeException("No value specified for -of");
			outSpec = argv[i+1];
			return i + 2;
		}

		return 0;					// Not a custom arg
	}

	// Returns which match term that matches or -1
	// matchit = list of match strings. A match string may be
	//	method-name().local-name. Either piece may contain wildcarding.
	// exact = true says do case-sensitive match
	// meth = method in which to look for loc
	// loc = null for any, non-null for specific local
	static int hasMatch(Vector matchit, boolean exact, Method meth, Local loc)
	{
		Vector vec1 = new Vector(1);
		vec1.setSize(1);

		// Scan thru match terms
		int cnt = matchit.size();
		for (int matchidx=0; matchidx<cnt; matchidx++) {
			String name = (String)matchit.elementAt(matchidx);
			if (name==null) continue;

			// Term is local-var match term & it matches "this"s method?
			int dotidx = name.lastIndexOf(").");
			if (dotidx < 0) continue;
			String mname = name.substring(0, dotidx+1);

			// Generic or wild method name
			if (name.charAt(dotidx-1)=='(')
			{
				vec1.setElementAt(mname, 0);
				if (isMatchFor(vec1, exact, meth) < 0)
					continue;
			}

			// Specific method
			else {
				Trek trek = meth.getClassFile().getTrek();
				if (trek.getObject(mname)!=meth)
					continue;
			}

			// If matching any, return match term. Else get and 
			// compare isolated name. Ret match term if it matches.
			// But dont match "this" unless explicitly specified
			if (loc==null) return matchidx;
			name = name.substring(dotidx+2);
			if (loc.isThis())
			{
				if (name.equals("this"))
					return matchidx;
			}
			else if (equalsWild(loc.toName(), name, exact))
				return matchidx;
		}								// End loop
		return -1;
	}

	// Search for classname
	static int isMatchFor(Vector matchit, boolean exact, ClassFile cf)
	{
		// Get name info of class looking to match
		String pkg = cf.toPackage();
		String cname = cf.toName();

		// Scan thru match terms
		int cnt = matchit.size();
		for (int matchidx=0; matchidx<cnt; matchidx++) {
			String name = (String)matchit.elementAt(matchidx);
			if (name==null) continue;

			// Chk if right pkg if one specified
			int dotidx = name.lastIndexOf('.');
			if (dotidx >= 0)
			{
				String currpkg = name.substring(0, dotidx);
				name = name.substring(dotidx+1);
				if (!equalsWild(pkg, currpkg, exact))
					continue;
			}

			// Compare isolated class. If dont match and more left, try again.
			// Else exit and return false.
			if (equalsWild(cname, name, exact))
				return matchidx;
		}							// End loop
		return -1;
	}

	// Returns match term used by instruction, or -1 if none
	// matchit = list of match strings. A match string may be
	//	method-name().local-name. Either piece may contain wildcarding.
	// exact = true says do case-sensitive match
	// Note that matching is name-based. You dont have to access actual
	// objects (and read other classfies) to do matches.
	static int isMatchFor(Vector matchit, boolean exact, Instruction inst)
	{
		// Chk for member ref. If there, see if its name and class match
		String qcname = inst.toMemberClass();
		if (qcname!=null)
		{
			String memname = inst.toMemberName();
			boolean fldref = inst.getType()!=Trek.IN_CALL;
			return isMatchFor(matchit, exact, qcname, memname, fldref);
		}
		
		// Chk for local ref
		Local loc = inst.getLocal();
		if (loc==null) return -1;
		Method meth = inst.getStatement().getMethod();
		return hasMatch(matchit, exact, meth, loc);
	}

	// Returns which match term that matches or -1
	// matchit = list of match strings. A match string may be
	//   name, class.name, package.class.name. The member-name portion of a
	//   match string may start and/or end with *, to indic a wildcard match.
	// exact = true says do case-sensitive match
	static int isMatchFor(Vector matchit, boolean exact, Member mem)
	{
		// Setup stuff needed to compare name-list to member's name
		String qcname = mem.getClassFile().toQualifiedName();
		return isMatchFor(matchit, exact, qcname, mem.toName(), mem instanceof Field);
	}

	// Returns which match term that matches or -1
	// matchit = list of match strings. A match string may be
	//   name, class.name, package.class.name. The member-name portion of a
	//   match string may start and/or end with *, to indic a wildcard match.
	// exact = true says do case-sensitive match
	// qcname = external-fmt qualified classname of a member
	// memname = this member's name
	// fldref = true if this member is a field, or false if its a method
	static int isMatchFor(Vector matchit, boolean exact,
		String qcname, String memname, boolean fldref)
	{
		// Scan thru match terms
		int cnt = matchit.size();
		for (int matchidx=0; matchidx<cnt; matchidx++) {
			String name = (String)matchit.elementAt(matchidx);
			if (name==null) continue;

			// If term is not name or name(), skip it since it cant match.
			// Also skip if (fldref && name()) or (!fldref && name).
			if (name.indexOf('(')>=0)
			{								// Better be method name
				if (!name.endsWith(")")) continue;
				if (fldref) continue;
				int len = name.length();
				if (name.charAt(len-2)=='(')
					name = name.substring(0, len-2);
				else continue;				// Specific method not supported
			}
			else if (!fldref)				// Field name
				continue;

			// If name is class specific, chk if this is right class
			int dotidx = name.lastIndexOf('.');
			if (dotidx >= 0)
			{
				String cname = name.substring(0, dotidx);
				if (exact)
				{
					if (qcname.equals(cname) || qcname.endsWith("." + cname))
						;					// Same pkg/class or same class
					else continue;
				}
				else {
					if (sameClassIgnoreCase(qcname, cname, dotidx))
						;					// Same pkg/class or same class
					else continue;
				}
				name = name.substring(dotidx+1);
			}

			// Compare isolated name. If dont match and more left, try again.
			// Else exit and return false.
			if (equalsWild(memname, name, exact))
				return matchidx;
		}							// End loop
		return -1;
	}

	// Open a text output file, defaulting to console if outSpec null
	static PrintWriter openReportFile()
	{
		// Setup output stream. Default to console
		try {
			if (outSpec==null)
				return new PrintWriter(System.out, true);
			else return new PrintWriter(new FileWriter(outSpec));
		}
		catch (Exception e) {
			throw new RuntimeException("Could not open output file");
		}
	}

	// Parse opd1 opr opd2
	// ifstr = thing to parse
	// opds = 2-elem array to ret opd1/2 in, also returns opr code
	// However if bad syntax, returns -1. If empty term, returns -2.
	static int parseIf(String ifstr, String opds[])
	{
		int opridx, ch1=0, which=-1;

		// If it's empty, just return
		int iflen = ifstr.length();
		if (iflen<=0)
			return -2;

		// Get to operator of Condition.
		for (opridx=0; opridx<iflen; opridx++) {
			ch1 = ifstr.charAt(opridx);
			which = "\\\'\"=!<>{}".indexOf(ch1);
			if (which>=0) break;	// Past variable ref
		}

		// If not a opr b, ret just an a
		opds[0] = ifstr;
		if (which<=2 || opridx==0 || opridx>=iflen-1)
			return -1;

		// Now identify which operator
		int ifopr = 0, oplen = 2;
		int ch2 = ifstr.charAt(opridx+1);
		switch (ch1) {
		case '=':
			ifopr = Trek.DOIF_EQ;
			if (ch2!='=') oplen = 1;
			break;
		case '!':
			if (ch2=='=')
				ifopr = Trek.DOIF_NE;
			else return -1;
			break;

		// Less than/less than or equal
		case '<':
		case '{':
			if (ch2=='=')
				ifopr = Trek.DOIF_LE;
			else {
				oplen = 1;
				ifopr =	Trek.DOIF_LT;
			}
			break;

		// Gtr than/gtr than or equal
		case '>':
		case '}':
			if (ch2=='=')
				ifopr = Trek.DOIF_GE;
			else {
				oplen = 1;
				ifopr =	Trek.DOIF_GT;
			}
			break;
		}

		// Return computed results
		opds[0] = ifstr.substring(0, opridx).trim();
		opds[1] = ifstr.substring(opridx+oplen).trim();
		return ifopr;
	}

	// Insure new lines dont start at column where prev ends (LF is only
	// LF -- not implicitly CRLF -- on some systems)
	static void println(PrintWriter chan)
	{
		chan.println('\r');
	}
	static void println(PrintWriter chan, String str)
	{
		chan.print(str);
		chan.println('\r');
	}

	// Returns true if qualified class name = class name or ends with .class-name
	// (with case ignored either way)
	private static boolean sameClassIgnoreCase(String qcname, String cname, int cnlen)
	{
		if (qcname.equalsIgnoreCase(cname))
			return true;

		// Qcname end with .<cnlen>
		int qcnlen = qcname.length();
		int tailpos = qcnlen-cnlen;
		if (tailpos<1 || qcname.charAt(tailpos-1)!='.')
			return false;

		// End of qcname case-less match of cname
		String qctail = qcname.substring(tailpos);
		return qctail.equalsIgnoreCase(cname);
	}

	// Pad and/or truncate val to specified field layout, and return it
	// val = string to size
	// tail = immutable piece to suffix to val, BUT before padding
	// left = if doing text, num of padding chars to left of string
	// width = if doing text, width to put lpad+val+tail into. 0 says no limit
	static String toPaddedLabel(String val, String tail, int left, int width)
	{
		// Setup left padding, and chk for simple case
		String lpad = spaces.substring(0, Math.min(20, left));
		if (width<=0)
			return lpad+val+tail;

		// Pad or truncate on right
		width -= tail.length();
		int len = val.length();
		if (len>=width)
			return lpad + val.substring(0, width) + tail;
		else return lpad + val + tail + spaces.substring(0, width-len);
	}

	// Output watch context with approp padding
	static String toPaddedPrefix(String context, int itype, int valstart)
	{
		// The stuff between context and value is watch-case specific
		String sfx;
		switch (itype) {
		case Jtu.wit_element:
		case Jtu.wit_if:
		case Jtu.wit_value:
			sfx = "==>  ";
			break;
		case Jtu.wit_switch:
			sfx = "==>  Case ";
			break;
		case Jtu.wit_while:
			sfx = "Iter ";
			break;
		default:
			sfx = "";
			break;
		}

		// Need formatted form for text log.
		// Shrink context string, if not room for all of it
		int ctxlen = context.length();
		if (ctxlen > valstart-1)
			return context.substring(0, valstart-5) + "... " + sfx;
		else return context + spaces.substring(0, valstart-ctxlen) + sfx;
	}

}

