//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

import dec.trek.*;
import java.io.*;
import java.util.*;

//
//
// cref. Do a cross-reference listing of program's members
//
//
public class cref extends Trek
{
	private boolean lastHadRefs;		// Control formatting around member lines
	private PrintWriter outChan;		// Identies report file
	private int pass;					// cref does 2 scans thru program

	protected void atStartOf(ClassFile cf)
	{
		// Formatting policy is indent inferiors and 1 blank line between
		// everything except (1) consecutive ref-less members
		// and (2) a class and its first member
		if (pass==2)
		{
			outChan.println(cf.toQualifiedName());
			lastHadRefs = true;
		}
	}

	protected void atEndOf(ClassFile cf)
	{
 		if (pass==1) return;
		if (!lastHadRefs) outChan.println();
	}

	protected void at(Field field)
	{
		if (pass==1) return;
		printRefs(field);
	}

	protected void atStartOf(Method method)
	{
		if (pass==1) return;
		printRefs(method);
	}

	protected void at(Instruction inst)
	{
		// Ret immed if this inst doesnt ref a member or mem cant be accessed
		Member mem;
		try {
			mem = inst.getMember();
			if (mem == null) return;
		}
		catch (TrekException trex) {
			return;
		}

		// Have a ref. Create place to store it if havent yet.
		// Conversely if dup ref from same statement, dont insert it
		Statement stmt = inst.getStatement();
		Vector refs = (Vector)mem.userDefined;
		if (refs == null)
		{
			refs = new Vector();
			mem.userDefined = refs;		// Permanize vector
		}
		else if (stmt == refs.lastElement())
			return;

		// Add this stmt to mem's list of refs
		refs.addElement(stmt);
		return;
	}

	// Process cref-specific cmd-line stuff
	protected int getCmdLineArg(String argv[], int i)
	{
		// Process one of the common args if present
		int newi = Jtu.getCmdLineArg(argv, i);
		if (newi > 0) return newi;		// Ate this arg

		return 0;					// Not a custom arg
	}

	public static void main(String argv[])
	{
		cref ctrek = new cref();
		try {
			ctrek.getCmdLine(argv);
			ctrek.makeCref();
		}
		catch (Throwable e) {
			System.out.println(ctrek.toErrorMessage(e));
			System.exit(1);
		}
	}

	// Do real work
	public void makeCref()
	{
		// Setup classes
		outChan = Jtu.openReportFile();
		initTrek();

		// 1) Find the refs  2) Print them
		pass = 1;
		doTrek();			// Go thru each instruction, finding refs
		pass = 2;
		doTrek(2);			// Scan thru member level to process collected refs

		// Cleanup
		endTrek();
		outChan.close();
	}

	// Process refs that have been stored with this member in 1st pass
	private final static int maxfrom = 20;
	private void printRefs(Member mem)
	{

		// Display diff line if member not referenced
		Vector vec = (Vector)mem.userDefined;
		if (vec==null)
		{
			outChan.println("  " + mem.toString() + " [No references]");
			lastHadRefs = false;
			return;
		}

		// Print member dcl and init loop
		if (!lastHadRefs) outChan.println();
		outChan.println("  " + mem.toString());
		String lastfrom = "";

		// Print the refs to this member
		for (int i=0; i<vec.size(); i++) {
			Statement stmt = (Statement)vec.elementAt(i);
			outChan.println("    " + stmt.toLocation()+ ":  " + stmt.toString());
		}
		lastHadRefs = true;
		outChan.println();
	}

}

/*
		// Print "class.its-method statement" for the ref
		String from = meth.getClassFile().getName() + "." + meth.getName();
		String frnum = "(" + stmt.getStatementNumber() + ")";
		int fromlen = from.length() + frnum.length();

		// If name too long, put at least half of it on a preceding line
		if (fromlen > maxfrom)
		{
			int prelen;
			if (fromlen > (maxfrom-2)*2)
				prelen = fromlen - (maxfrom-2);
			else prelen = fromlen/2 + 1;

			// Dont want to do same pre-line mult times in a row
			if (!from.equals(lastfrom))
			{
				String prefrom = from.substring(0, prelen);
				outChan.println("    " + prefrom);
			}

			// Construct suffix and updated length
			lastfrom = from;
			from = "  " + from.substring(prelen);
			fromlen = from.length() + frnum.length();
		}
		else lastfrom = from;

		// Cause 1st columns of stmts to align
		from = from + frnum + ":                  ".substring(0, 20+2-fromlen);
*/

