//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

package dec.trek;

//
//
// Call. For building a method call to be inserted in user's code
//
//
public final class Call extends Change
{
	// If user sets this, RunLog.sysInited chk is not generated
	// and call can be inserted in jdk initer methods
	public static boolean systemMethodsSafe;

	// Stuff for inserted code and related stuff.
	// Hack: just create big array to build call in
	private String callName;		// Instrum method being called (must be static)
	private String callDesc = "(";	// Build signature of instrumentation method
	private String ipStackDesc[];	// Data types info for stack args if not 1st
	private int jumps[] = new int[5];// Offsets where to insert doif jump targets
	private int jumpIdx;			// # of such offsets
	private int passedIndex = -1;	// Temp for passing array element's index
	private int passedValue = -1;	// Temp for passing ins pt's value (needed
									// only when it is not 1st thing passed)
	private int passedObject = -1;	// Temp for passing ins pt's object
	private int passedStack[];		// Temps for non-invoke passStack args
	private int temps;				// Temp locals needed so far
	private int wild;				// Type of insert-multiple call

	// Flavors of methods
	private static final int Flav_vanilla = 0;
	private static final int Flav_sysmethod = 1;
	private static final int Flav_sysiniter = 2;

	// For doIf comparisons
	private final static int if0inst[] = {
		Code.ifne, Code.ifeq,
		Code.ifge, Code.ifle,
		Code.ifgt, Code.iflt};
	private final static int compinst[] = {
		Code.if_icmpne, Code.if_icmpeq,
		Code.if_icmpge, Code.if_icmple,
		Code.if_icmpgt, Code.if_icmplt};

	// add call at end of main class
	// mname = name of method to insert call to
	// trek = the trek to insert call for
	public static Call addAfter(String mname, Trek trek)
	{
		// Is there a main class?
		Method meth = trek.getMainMethod();
		if (meth == null) return null;
		Call call = addAfter(mname, meth);
		if (call!=null)
			call.priority = Statement.plev_afttrek;
		return call;
	}

	// Add call at finalize method of class
	public static Call addAfter(String mname, ClassFile cf)
	{
		// After-class is defined as end of finalizer
		Method meth = cf.lookupMethod("finalize", null);
		if (meth == null) return null;
		Call call = addAfter(mname, meth);
		if (call!=null)
			call.priority = Statement.plev_aftclass;
		return call;
	}

	// Insert call at each exit point of the specified method
	public static Call addAfter(String mname, Method meth)
	{
		// Find some Return stmt to use as hook for insert pt
		for (Statement stmt=meth.lastStatement(); stmt!=null; stmt=stmt.prev()) {
			if (stmt.type==Trek.ST_RETURN || stmt.type==Trek.ST_THROW)
			{
				Call call = addCall(mname, stmt, null, Trek.ST_EXIT);
				if (call!=null)
					call.priority = Statement.plev_aftmethod;
				return call;
			}
		}
		return null;					// No returns, no can insert
	}

	// How handle code-less stmts???

	// Add call after specified stmt.
	public static Call addAfter(String mname, Statement stmt)
	{
		// Chk if user trying to insert code at inaccessible location
		// Note: after Return/Throw stmt is treated as after retval setup
		//		if any, and after Switch stmt is treated as after key setup.
		switch (stmt.type) {

		// After-if/endloop is inaccess and context is imprecise (vs Return)
		case Trek.ST_IF:
		case Trek.ST_ENDDO:
		case Trek.ST_ENDFOR:
		case Trek.ST_ENDWHILE:

		// Unconditional goto
		case Trek.ST_BREAK:
		case Trek.ST_CONTINUE:
			throw new TrekException("Code cannot be inserted after a jumping statement");

		// User would equate after these to start of content
		case Trek.ST_TRY:
		case Trek.ST_CATCH:
		case Trek.ST_FINALLY:
		case Trek.ST_LABEL:
		case Trek.ST_CASE:
		case Trek.ST_DO:
		case Trek.ST_FOR:
		case Trek.ST_WHILE:
		case Trek.ST_THEN:
		case Trek.ST_ELSE:
			return addBefore(mname, stmt.next());
		}

		// Do the requested work
		return addCall(mname, stmt, null, Trek.ST_AFTER);
	}

	// Add call after specified path of if-stmt, or return null if no such path.
	// stmt = an if-stmt
	// truepath = true says want to insert on Then-path, false Else-path
	public static Call addAfter(String mname, Statement stmt, boolean truepath)
	{
		Statement path;

		// If If-stmt optimized to branch directly to Then, there is no
		// real Then or Else path. Of course, even at src-level, there may
		// be no Else path. However optimized code this occurs more often
		if (truepath)
			path = stmt.firstOfThen();
		else path = stmt.firstOfElse();

		// Have start of desired path if any, init call to insert there
		if (path!=null)
			return addBefore(mname, path);
		else return null;
	}

	// Add call at specified instruction (ie. after its stack args are set up)
	public static Call addAt(String mname, Instruction inst)
	{
		return addCall(mname, inst.parent, inst, Trek.ST_AT);
	}

	// add call at start of main classfile
	public static Call addBefore(String mname, Trek trek)
	{
		// Is there a main class?
		Method meth = trek.getMainMethod();
		if (meth == null) return null;
		Call call = addBefore(mname, meth);
		if (call!=null)
			call.priority = Statement.plev_beftrek;
		return call;
	}

	// Add call at class's constructor
	public static Call addBefore(String mname, ClassFile cf)
	{
		// Before-class is defined to mean: start of all constructors
		for (Method meth=cf.firstMethod(); meth!=null; meth=meth.next()) {
			Statement stmt = meth.firstStatement();
			if (meth.isConstructor() && stmt!=null)
			{
				Call call = addCall(mname, stmt, null, Trek.ST_ENTER);
				if (call!=null)
					call.priority = Statement.plev_befclass;
				return call;
			}
		}
		return null;
	}

	// Add call at start of method
	public static Call addBefore(String mname, Method meth)
	{
		Statement stmt = meth.firstStatement();
		if (stmt==null) return null;

		// Build the call object. Call is conditional for java.lang methods
		// because jvm inits that occur before 1st user code reached cant
		// tolerate calls from jdk to outside jdk (ie. when the class being
		// instrumented is in java.lang)
		Call call = addBefore(mname, stmt);
		if (call!=null)
			call.priority = Statement.plev_befmethod;
		return call;
	}

	// Add call at start of statement
	public static Call addBefore(String mname, Statement stmt)
	{
		return addCall(mname, stmt, null, Trek.ST_BEFORE);
	}

	// add a call to a method either before or after specified ins pt
	// mname = fully qualified name of the static method to call
	//			if not preceded by xxx: is also changeId to give add.
	// st = stmt to modify
	// in = ins pt is specific instruction
	// typ = <=Trek.ST_INSERTED, or _ENTER or _EXIT
	private static Call addCall(String mname,
		Statement st, Instruction in, int typ)
	{
		// Chk arg consistency
		if (st==null && in==null)
			throw new TrekException("Null insert point specified");
		if (mname==null || mname.indexOf('.')<0)
			throw new TrekException(
				"Method to call not 'qualified-class-name.method-name'");

		// JVM's cant handle instrumented jdk constructors in all cases
		int flavor = getMethodFlavor(st.parent);
		if (flavor==Flav_sysiniter)
			return null;

		// Create call object.
		// Insure system methods arent called until logger inited.
		// This prevents pre-init recursion/crashes when system
		// methods are themselves instrumented
		Call call = new Call(mname, st, in, typ);
		if (flavor==Flav_sysmethod)
			call.doIfInt("/dec/trek/RunLog/sysInited", Trek.DOIF_NE, 0);
		return call;
	}
	
/*
	// Init a call that starts with unconditional goto (if havent yet)
	// mname = method to call
	// targstmt = stmt to prepend the goto to
	// Notes:
	//	the adjustPc logic recognizes this as special case and the branches
	//	in the if-stmt that branch to targstmt are forced to branch past the
	//	inserted goto (rather than to it as would normally occur)
	//	The goto can be needed even if Then blk ends in goto. For ex, some
	//	stmt other than If could branch to stmt past Then blk.
	private static Call addGoto(String mname, Statement targstmt)
	{
		Call call = addCall(mname, targstmt, null, Trek.ST_BEFORE);
		if (call!=null && !targstmt.hasAddedGoto())
		{
			call.priority = Statement.plev_else;
			call.jumps[call.jumpIdx++] = 0;
			call.writeShortInst(Code.goto0, 0);
		}
		return call;
	}
*/

	// Add var with specified desc to arg list of call
	private void addToArgList(String desc)
	{
		// If double or long, need 2 stack words for this arg
		// Objects must be passed generically
		if (desc.startsWith("[")
			|| (desc.startsWith("L") && !desc.equals("Ljava/lang/String;")))
			desc = objectDesc;
		callDesc += desc;
	}
	private void addToArgList(String desc, String supname)
	{
		// Arg must be obj ref
		if (!desc.startsWith("L"))
			throw new TrekException("Value must be an object reference");

		// If instrumentation method has broader type than desc, verify
		// that it is superclass of desc
		if (supname!=null)
		{
			ClassFile supercf = cf.getTrek().getClassFile(supname);
			String subname = Datum.getDescClass(desc);
			ClassFile subcf = cf.getTrek().getClassFile(subname);

			// Got the classes, now verify
			if (!supercf.isAncestor(subcf))
				throw new TrekException(
					supname + " is not a superclass of " + subname);
			callDesc += "L" + supname.replace('.', '/') + ";";
		}
		else callDesc += desc;

	}

	// Alloc room for storing info for stack args -- if havent yet
	private void allocStackAttr(int num)
	{
		if (ipStackDesc!=null) return;
		ipStackDesc = new String[num];
		passedStack = new int[num];
		for (int i=0; i<num; i++) passedStack[i] = -1;	// -1 = unassigned
	}

	// Create Call object.
	// Args -- see addCall (AND it can only be invoked from addCall)
	private Call(String mname, Statement st, Instruction in, int typ)
	{
		// Canonicalize wildcard insert
		if (typ==Trek.ST_EXIT)
		{
			wild = typ;
			typ = Trek.ST_AFTER;
		}
		else if (typ==Trek.ST_ENTER)
		{
			wild = typ;
			typ = Trek.ST_BEFORE;
		}

		// Save method name and change ID
		int icolon = mname.indexOf(':');
		if (icolon<0)
			changeId = callName = mname;
		else {
			changeId = mname.substring(0, icolon);
			callName = mname.substring(icolon + 1);
		}

		// Save other input properties
		stmt = st;
		inst = in;
		meth = st.parent;			// Convenience
		cf = meth.parent;			// Convenience
		priority = Statement.plev_normal;
		setInsertPoint(typ);
	}									// End addCall

	// Gen code to compare the 2 values placed on stack
	// desc1 = descriptor of 1st val
	// opr = the compare to do (DOIF_*)
	// desc2 = descriptor of 2nd val
	private void doCompare(String desc1, int opr, String desc2)
	{
		// Chk if 2 descriptors consis
		if (desc1.equals("Z")) desc1 = "I";
		if (desc2.equals("Z")) desc2 = "I";
		if (!desc1.equals(desc2))
			throw new TrekException("Data types of compared values must be the same");

		// save where to adjust target, and do the if-inst
		// Do integer compare
		if (desc1.equals("I"))
		{
			jumps[jumpIdx++] = newCodeLen;
			writeShortInst(compinst[opr], 0);
		}

		// Do long compare
		else if (desc1.equals("J"))
		{
			newCode[newCodeLen++] = (byte)Code.lcmp;
			jumps[jumpIdx++] = newCodeLen;
			writeShortInst(if0inst[opr], 0);
		}

		// Do double compare
		else if (desc1.equals("D"))
		{
			newCode[newCodeLen++] = (byte)Code.dcmpg;
			jumps[jumpIdx++] = newCodeLen;
			writeShortInst(if0inst[opr], 0);
		}

		// Do float compare
		else if (desc1.equals("F"))
		{
			newCode[newCodeLen++] = (byte)Code.fcmpg;
			jumps[jumpIdx++] = newCodeLen;
			writeShortInst(if0inst[opr], 0);
		}

		// Do object compare (which is optimized for strings because
		// String has equals method that overrides Object.equals)
		else {
			// Insure equals method in pool and do the equals call
			MethodRef ref = MethodRef.add("java/lang/String/equals",
				"(Ljava/lang/Object;)Z", cf);
			writeShortInst(Code.invokevirtual, ref.idx);

			// Do the actual (inverted) condit jump
			jumps[jumpIdx++] = newCodeLen;
			if (opr==Trek.DOIF_EQ)
				writeShortInst(Code.ifeq, 0);		// Jump if false
			else if (opr==Trek.DOIF_NE)
				writeShortInst(Code.ifne, 0);		// Jump if true
			else throw new TrekException("String comparison supports only EQ and NE");
		}
	}

	// Do compare of object to null
	private void doCompareNull(String desc1, int opr)
	{
		if (desc1.charAt(0)!='L')
			throw new TrekException("Only objects can be compared to null");
		jumps[jumpIdx++] = newCodeLen;
		int opcode = opr==Trek.DOIF_EQ ? Code.ifnull : Code.ifnonnull;
		writeShortInst(opcode, 0);
	}

	// Gen code to compare variable and another term,
	// var1 = variable reference
	// opr = compare to do
	// term2 = var2, "xxx", xxx, true, false, or numeric constant
	public void doIf(String var1, int opr, String term2)
	{
		// Load 1st term, and chk if special of compare to null
		String desc1 = doIfLoad1(var1);
		if (term2==null)
		{
			doCompareNull(desc1, opr);
			return;
		}

		// Decide if 2nd term is var or constant
		String desc2 = doConstant(desc1, term2);
		if (desc2==null)
			desc2 = loadVarRef(term2);
		doCompare(desc1, opr, desc2);
	}

	// Compare variable with boolean value
	public void doIfBoolean(String var1, boolean val)
	{
		String desc1 = doIfLoad1(var1);
		String desc2 = loadInteger(val ? 1 : 0);
		doCompare(desc1, Trek.DOIF_EQ, desc2);
	}

	// Compare variable with integer value
	public void doIfInt(String var1, int opr, int val)
	{
		String desc1 = doIfLoad1(var1);
		String desc2 = loadInteger(val);
		doCompare(desc1, opr, desc2);
	}

	// Compare variable with String value (incl null)
	public void doIfString(String var1, int opr, String val)
	{
		// Load var1, (create string constant &) load string, & do compare
		String desc1 = doIfLoad1(var1);
		if (val==null)
		{
			doCompareNull(desc1, opr);
			return;
		}
		String desc2 = doString(val);
		doCompare(desc1, opr, desc2);
	}

	// Load 1st term of doIf
	private String doIfLoad1(String var1)
	{
		if (callDesc.length() > 1)
			throw new TrekException("IFs must be done before any arguments passed");
		return loadVarRef(var1);
	}

	// Done building call, insert the code
	// Notes:
	//	Use of ins pt value or object cant happen if wildcarding
	public void done()
	{
		// Insure classfile has cp entries needed to do this call
		// and add the call to the instruction stream
		if (callName!=null)
		{
			String mname = callName.replace('.', '/');
			callDesc += ")V";
			MethodRef ref = MethodRef.add(mname, callDesc, cf);
			writeShortInst(Code.invokestatic, ref.idx);
		}

		// Build the access code for insert pt data if applic.
		// However such code is not allowed if insert pt is wild.
		int delta = 0;
		if (passedObject>=0 || passedValue>=0 || passedIndex>=0
			|| ipStackDesc!=null)
		{
			delta = mergeObjectAccessCode();
			if (delta!=0 && wild!=0)
				throw new TrekException("Insert point data cannot be "
					+ "passed when code auto-inserted at multiple points");
		}

		// Insert code at each target statement
		if (wild==Trek.ST_EXIT)
		{
			for (stmt=meth.lastStatement(); stmt!=null; stmt=stmt.prev()) {
				if (stmt.type==Trek.ST_RETURN || stmt.type==Trek.ST_THROW)
					doneCall(0);
			}
		}
		else if (wild==Trek.ST_ENTER)
		{
			for (meth=cf.firstMethod(); meth!=null; meth=meth.next()) {
				stmt = meth.firstStatement();
				if (meth.isConstructor() && stmt!=null)
					doneCall(0);
			}
		}
		else doneCall(delta);
	}

	// Do done-logic for a specific insert point
	// delta = delta for use of insert-pt context
	private void doneCall(int delta)
	{
		// Insure do-if jumps go to end of new code.
		for (int i=0; i<jumpIdx; i++) {
			jumps[i] += delta;			// Obj access code moves If's
			Instruction.writeShort(newCodeLen-jumps[i], newCode, jumps[i]+1);
		}

		// Insert the new code
		doneInsert();
	}

	// Write insts to load a member, and return its data type
	// name = name of field to load, or method(...)
	// qcname = fully-qualified internal-fmt of name's class,
	//			or null for ins pt's class
	// refdesc = if ref term already exists, its data type
	protected String doMember(String name, String qcname, String refdesc)
	{
		String desc;

		// Special case RunLog.sysInited so that user doest have to put
		// RunLog in input path of trek in order to use sysInited
		if (name.equals("sysInited")
			&& qcname!=null && qcname.equals("dec/trek/RunLog"))
		{
			desc = "I";
			loadField(qcname + "/sysInited", name, desc, refdesc, true);
			return desc;
		}
		else return super.doMember(name, qcname, refdesc);
	}									// End loadMember

	// Use data type to determine load inst to use
	private int getLoadInst(String desc)
	{
		switch (desc.charAt(0)) {
		case '[':
		case 'L':
			return Code.aload;
		case 'B':
		case 'C':
		case 'I':
		case 'S':
		case 'Z':
			return Code.iload;
		case 'J':
			return Code.lload;
		case 'D':
			return Code.dload;
		case 'F':
			return Code.fload;
		}
		throw new TrekException("Data type not recognized");
	}

	// Returns method's flavor (Flav_*)
	private static int getMethodFlavor(Method meth)
	{
		if (systemMethodsSafe) return Flav_vanilla;

		// Chk if method in system class
		ClassFile cf = meth.parent;
		boolean sysmeth = cf.isSystemClass();
		if (!sysmeth) return Flav_vanilla;

		// Chk for <clinit> or constructor
		if (cf.clinit==meth || meth.isConstructor())
			return Flav_sysiniter;
		else return Flav_sysmethod;
	}

	// Use data type to determine store inst to use
	private int getStoreInst(String desc)
	{
		switch (desc.charAt(0)) {
		case '[':
		case 'L':
			return Code.astore;
		case 'B':
		case 'C':
		case 'I':
		case 'S':
		case 'Z':
			return Code.istore;
		case 'J':
			return Code.lstore;
		case 'D':
			return Code.dstore;
		case 'F':
			return Code.fstore;
		}
		throw new TrekException("Data type not recognized");
	}

	// (Calculate and) Return the lv idx to use for this temp local
	// offset = value to return, or -1 to calculate
	// desc = for size of temp local
	private int getTempIdx(int offset, String desc)
	{
		// Already calculated?
		if (offset>=0)
			return offset;

		// Calculate now
		if (meth.callLocalsBase<0)
			meth.callLocalsBase = meth.addLocals;
		offset = meth.maxLocals + meth.callLocalsBase + temps;

		// Update amt of temp local slots in use, and return offset
		if (desc==null || Datum.getDataSize(desc)==1)
			temps += 1;
		else temps += 2;
		return offset;
	}

	// Write an inst to load a field
	// qname = fully qualified name of field to load
	// name = name of field to load
	// desc = its data type
	// refdesc = descriptor of object reference, or null if leftmost term
	// is_static = true if static field
	protected void loadField(String qname, String name, String desc,
		String refdesc, boolean is_static)
	{
		// Simple case is loading a static field (no obj ptr). Conversely
		// specifying an object is inconsistent with loading a static field
		FieldRef ref = FieldRef.add(qname, desc, cf);
		if (is_static)
		{
			if (refdesc!=null)
				throw new TrekException(
					"Object reference for " + name
					+ " not allowed because it is static");
			writeShortInst(Code.getstatic, ref.idx);
		}

		// If no object yet, method of ins pt better have a "this".
		// If no problems, actually load the instance field.
		else {
			if (refdesc==null)
			{
				if (meth.hasModifier(Trek.ACC_STATIC))
					throw new TrekException(
						"No object reference for " + name
						+ " and " + meth.toName() + " is static");
				writeByteInst(Code.aload, 0);
			}
			writeShortInst(Code.getfield, ref.idx);
		}
	}

	// Chk if term is ins pt. Returns desc if is, false if not
	protected String loadInsertPoint(String term, String refdesc)
	{
		// 1) Ins pt's value
		if (term.equals("@"))
		{
			// 1a) Pseudo-ins pt value for load inst... do actual var ref
			if (ipName!=null)
			{
				int subpos[] = {0};
				refdesc = doName(ipName, subpos, refdesc);
				if (subpos[0]>=0)
					refdesc = doName(ipName, subpos, refdesc);
			}

			// 1b) Insert pt's value (but not once-only case when pass() used).
			else {
				loadIpValue(term);
				refdesc = ipDesc;
			}
			return refdesc;
		}

		// 2) It can be @@ (ie. insert point's object)
		else if (term.equals("@@"))
		{
			if (ipObject==null)
				throw new TrekException("No object identified by insert point");
			if (inst.type==Vmi.it_call)
				passInvoke(true);
			else {
				passedObject = getTempIdx(passedObject, null);
				writeByteInst(Code.aload, passedObject);
			}
			return "L" + ipObject + ";";
		}

		// Something else
		else return null;
	}

	// Write an instruction to load an integer constant
	protected String loadInteger(int ival)
	{
		// See if common special instruction
		if (ival==0)
			newCode[newCodeLen++] = (byte)Code.iconst_0;
		else if (ival==1)
			newCode[newCodeLen++] = (byte)Code.iconst_1;

		// See if can do it with immed constant
		else if (ival > -128 && ival <128)
			writeByteInst(Code.bipush, ival&0xff);
		else if (ival > -32382 && ival <32382)
			writeShortInst(Code.sipush, ival);

		// Need a pool constant.
		else loadPoolConstant(IntegerConstant.add(ival, cf));
		return "I";
	}

	// Load the insert pt's value
	private void loadIpValue(String name)
	{
		// Cant be done if ins pt has no value
		if (ipDesc==null)
			throw new TrekException("Insert point does not have a value");

		// The really fast load if possible
		if (newCodeLen==0 && name==null)
			writeStackInst(Code.dup, Code.dup2);

		// Cant. Get it from local (see MergeObjectAccessCode)
		else {
			passedValue = getTempIdx(passedValue, ipDesc);
			writeByteInst(getLoadInst(ipDesc), passedValue);
		}
	}

	// Load a local var
	protected void loadLocal(Local loc, String desc)
	{
		writeByteInst(getLoadInst(desc), loc.idx);
	}

	// Write an inst to load a method
	// qname = fully qualified name of method to load
	// name = unqualified name of method (for excep msg's)
	// desc = its full descriptor
	// thisload = index in newCode where "this" loaded,
	//			or -1 if name is the leftmost term of variable ref
	// is_static = true if static method
	protected void loadMethod(String qname, String name, String desc,
		int thisload, boolean is_static)
	{
		// Simple case is loading a static field (no obj ptr). Conversely
		// specifying an object is inconsistent with loading a static field
		MethodRef ref = MethodRef.add(qname, desc, cf);
		if (is_static)
		{
			if (thisload<0)
				throw new TrekException(
					"Object reference for " + name
					+ " not allowed because it is static");

			// Undo loadArgs hack of presuming this-load was needed
			newCodeLen -= 1;
			int tail = newCodeLen - thisload;
			if (tail>0)
			System.arraycopy(newCode, thisload+1, newCode, thisload, tail);
			writeShortInst(Code.invokestatic, ref.idx);
		}

		// If no object yet, method of ins pt better have a "this".
		// If no problems, invoke the method
		else {
			if (thisload>=0 && meth.hasModifier(Trek.ACC_STATIC))
				throw new TrekException(
					"No object reference for " + name
					+ " and " + meth.toName() + " is static");
			writeShortInst(Code.invokevirtual, ref.idx);
		}
	}

	// Load element from constant pool
	// refidx = element's index in cp
	protected void loadPoolConstant(CpInfo cpi)
	{
		// Chk for 64 bit constant
		if (cpi instanceof DoubleConstant || cpi instanceof LongConstant)
			writeShortInst(Code.ldc2_w, cpi.idx);

		// If low index, use shorter instruction
		else if (cpi.idx < 256)
			writeByteInst(Code.ldc, cpi.idx);
		else writeShortInst(Code.ldc_w, cpi.idx);
	}

	// Pass an inst's stack args -- up to 2
	// dt1 = desc of 1st arg
	// dt2 = desc or 2nd arg, or null if no 2nd arg
	// Notes:
	//	If inst's stack input <=64 bits, it is passed directly from stack
	private void loadStackFast(String dt1, String dt2)
	{
		// Use general form if args already passed
		if (newCodeLen!=0)
		{
			loadTemps(dt1, dt2, null);
			return;
		}

		// Do fast by not storing to a local before use, just dup
		addToArgList(dt1);
		if (dt2==null)
		{
			if (Datum.getDataSize(dt1)==2)
				newCode[newCodeLen++] = (byte)Code.dup2;
			else newCode[newCodeLen++] = (byte)Code.dup;
		}

		// Do two 32 bit values with a single dup2, or use general form
		else if (Datum.getDataSize(dt1)!=1 || Datum.getDataSize(dt2)!=1)
			loadTemps(dt1, dt2, null);
		else {
			addToArgList(dt2);
			newCode[newCodeLen++] = (byte)Code.dup2;
		}
	}

	// Load stack item of specified type into specified passStack temp
	private void loadTemp(String desc, int idx)
	{
		loadTempAdd(desc, idx, true);
	}
	private void loadTempAdd(String desc, int idx, boolean add)
	{
		if (add)
			addToArgList(desc);
		ipStackDesc[idx] = desc;
		passedStack[idx] = getTempIdx(passedStack[idx], desc);
		writeByteInst(getLoadInst(desc), passedStack[idx]);
	}

	// Pass one or more prev stored stack items (slave of loadStackFast)
	private void loadTemps(String dt1, String dt2, String dt3)
	{
		// Load 1st item, and 2nd and third items if they exist
		allocStackAttr(dt3!=null ? 3 : dt2!=null?2:1);
		loadTemp(dt1, 0);
		if (dt2!=null) loadTemp(dt2, 1);
		if (dt3!=null) loadTemp(dt3, 2);
	}

	// Append code for loading a "this". Note that this method may be called
	// before we even know if call site has a "this" or not
	protected void loadThis()
	{
		newCode[newCodeLen++] = (byte)Code.aload_0;	// Load "this"
	}

	// Generate instructions to load the value of the specified variable ref.
	// name = variable reference to analyze and load
	// Also returns its descriptor
	private String loadVarRef(String name)
	{
		// Extire expression is insert pt's value.
		if (name==null)
		{
			loadIpValue(name);
			return ipDesc;
		}

		// Get & load term at pos if any
		return doVarRef(name);
	}

	// Save the value(s)/object assoc with insert pt, or inst's stack args.
	// Returns how many insts were prepended to code stream to do the saves
	// Notes:
	//	After the saves done, stack is always as was before the saves
	//	The ins-pt stuff is more complicated than necessary so as to 
	//	insert the fewest number of bytecodes into the user's program.
	//	The passStack stuff is less predictable so it just does the brute
	//	force store/load for each value needed.
	private int mergeObjectAccessCode()
	{
		// Done with call-building insts, save them to side
		byte callcode[] = newCode;
		int callen = newCodeLen;
		newCodeLen = 0;				// Bytes of insts in setup/call
		newCode = new byte[200];

		// Store index/object and possibly ins pt value in temp locals
		if (passedIndex>=0)
			mergeObjectOrIndex(Code.istore, passedIndex);
		else if (passedObject>=0)
			mergeObjectOrIndex(Code.astore, passedObject);

		// If only ins-pt info used is the value (and it is not the 1st
		// arg of the instrumentation-method), dup & save it in temp loc.
		else if (passedValue>=0)
		{
			writeStackInst(Code.dup, Code.dup2);
			writeByteInst(getStoreInst(ipDesc), passedValue);
		}

		// Setup temp locs for instruction's stack args if they were passed
		if (ipStackDesc!=null)
			mergeStack();

		// Update cnt of temp locals used, method-wide
		// Append call setup to value-access code
		meth.addLocals = Math.max(meth.callLocalsBase+temps, meth.addLocals);
		int delta = newCodeLen;
		System.arraycopy(callcode, 0, newCode, newCodeLen, callen);
		newCodeLen += callen;
		return delta;
	}										// mergeObjectAccessCode

	// Save ins pt's object or index, and its val too if requested.
	// opcode = write-inst to use to save this thing
	// locidx = temp local idx for the object or index
	private void mergeObjectOrIndex(int opcode, int locidx)
	{
		// At a store, so OI V at top of stack (OI=shorthand for the 
		// stack loc containing either an object-ref or array-index)
		// Create V OI V on stack, saving top V if ins pt val also needed
		// Finally go from V OI to OI V OI on stack to prep to save OI
		if (ipDesc!=null)
		{
			writeStackInst(Code.dup_x1, Code.dup2_x1);
			if (passedValue>=0)
				writeByteInst(getStoreInst(ipDesc), passedValue);
			else writeStackInst(Code.pop, Code.pop2);
			writeStackInst(Code.dup_x1, Code.dup_x2);
		}

		// At a read, thus OI already at top of stack.
		// So can simply dup it to prepare for saving it
		else newCode[newCodeLen++] = (byte)Code.dup;

		// Save the now-duplicated object or index
		writeByteInst(opcode, locidx);
	}

	// Setup temp locs for inst's stack args, and restore its stack
	private void mergeStack()
	{
		int len = ipStackDesc.length;

		// Setup temp locals
		for (int i=len-1; i>=0; i--) {
			if (ipStackDesc[i]!=null)
				writeByteInst(getStoreInst(ipStackDesc[i]), passedStack[i]);
		}

		// Restore inst's stack
		for (int i=0; i<len; i++) {
			if (ipStackDesc[i]!=null)
				writeByteInst(getLoadInst(ipStackDesc[i]), passedStack[i]);
		}
	}

	// Normalize various source integer data types to the one internal
	// integer 32-bit type (ie. int).
	// Add insert pt's value to arglist.
	// Notes:
	//	if called for a value-less return stmt, there is no ipDesc so error occurs
	public void pass()
	{
		if (ipName!=null)
			pass(ipName);
		else {
			loadIpValue(null);
			addToArgList(ipDesc);
		}
	}

	// Pass a variable ref. Same syntax as in Java source except for 2
	// addit shorthands: @ = insert pt's value if any. @@ = its obj if any
	// Also support "passing" side effects... assume a method call whose
	// ret desc is V was called for the side-effects it performs
	public void pass(String name)
	{
		String desc = loadVarRef(name);
		if (desc.equals("V"))
			desc = doString("<Side Effects>");
		addToArgList(desc);
	}

	// Add insert pt's value to arglist.
	// clname = class to pass it as in method ref
	public void passAs(String clname)
	{
		if (ipName!=null)
			passAs(clname, ipName);
		else {
			loadIpValue(null);
			addToArgList(ipDesc, clname);
		}
	}

	// Pass an object reference using a explicit data type
	// clname = class to pass it as in method ref
	// name = the variable reference
	public void passAs(String clname, String name)
	{
		addToArgList(loadVarRef(name), clname);
	}

	// Pass 64-bit floating value
	public void passDouble(double dval)
	{
		addToArgList("D");
		loadPoolConstant(DoubleConstant.add(dval, cf));
	}

	// Pass 32-bit floating
	public void passFloat(float fval)
	{
		addToArgList("F");
		loadPoolConstant(FloatConstant.add(fval, cf));
	}

	// Process the types that are internally integers. Except for
	// boolean -- which compiler wont allow integer ops to be done on,
	// this is implicit because byte, char, short all promote to ints.
	public void passBoolean(boolean ival) {passInteger(ival?1:0, "Z");}
	public void passInt(int ival) {passInteger(ival, "I");}
	private void passInteger(int ival, String desc)
	{
		addToArgList(desc);
		loadInteger(ival);
	}

	// Pass the index of the insert point's array element
	public void passIndex()
	{
		// Chk if at array element
		if (!ipElement)
			throw new TrekException(
				"No array element is associated with insert point");

		// Pass the array index
		addToArgList("I");
		passedIndex = getTempIdx(passedIndex, null);
		writeByteInst(getLoadInst("I"), passedIndex);
	}

	// Pass the args of invoke*
	// onlythis = true, pass only this. false, all args including "this" if any
	// Notes:
	//	onlythis supports use of @@ for call insts. Passing "this" requires
	//	all the work of passing invoke's stack except the loads of the args.
	private void passInvoke(boolean onlythis)
	{
		// Get number of args, via methodref pool entry
		CpInfo cpi = cf.constant(inst.readUnsignedShort(inst.start+1));
		int numargs = cpi.getArgCount();

		// Determ if ref to static method. If has neither obj nor args, done
		int hasthis = inst.getOpcode()==Code.invokestatic ? 0 : 1;
		numargs += hasthis;
		if (numargs<=0)
			return;

		// Alloc temps, and pass object ref if any
		allocStackAttr(numargs);
		if (hasthis>0)
			loadTempAdd("L" + cpi.toInternalClassName() + ";", 0, !onlythis);

		// Get data type of each arg and pass it
		int len;
		String desc = Method.toArgsDesc(cpi.toSignature());
		for (int i=hasthis, pos=0; i<numargs; i++, pos+=len) {
			len = Datum.getDescLength(desc, pos);
			String argdesc = desc.substring(pos, pos+len);
			if (onlythis)
			{
				ipStackDesc[i] = argdesc;
				passedStack[i] = getTempIdx(passedStack[i], argdesc);
			}
			else loadTemp(argdesc, i);
		}
	}

	// Pass 64-bit integer
	public void passLong(long lval)
	{
		addToArgList("J");
		loadPoolConstant(LongConstant.add(lval, cf));
	}

	// Pass instruction's stack arguments -- must be untyped
	// desc = type(s) to give to stack arg(s)
	public void passStack(String desc)
	{
		// Instruction is insert pt
		if (inst==null)
			throw new TrekException("Insert point not set at an instruction");
		switch (inst.opcode) {

		// Untyped args: 1 32-bit value
		case Code.pop: case Code.swap:
		case Code.dup: case Code.dup_x1: case Code.dup_x2:
			if (Datum.getDescLength(desc, 0)!=desc.length()
				|| Datum.getDataSize(desc)!=1)
				throw new TrekException(
					"'desc' not consistent with " + inst.toName());
			loadStackFast(desc, null);
			break;
		
		// Untyped args: two 32-bit values, or a double or long
		case Code.pop2:
		case Code.dup2: case Code.dup2_x1: case Code.dup2_x2:
			int len1 = Datum.getDescLength(desc, 0);
			String desc1 = desc.substring(0, len1);
			int size1 = Datum.getDataSize(desc1);

			// The double or long case
			if (size1==2)
			{
				if (len1==desc.length())
				{
					loadStackFast(desc1, null);
					break;
				}
			}

			// The 2 values case
			else if (size1==1)
			{
				int len2 = Datum.getDescLength(desc, len1);
				String desc2 = desc.substring(len1, len2);
				int size2 = Datum.getDataSize(desc2);
				if (size2==1 && len1+len2==desc.length())
				{
					loadStackFast(desc1, desc2);
					break;
				}
			}
			throw new TrekException("'desc' not consistent with " + inst.toName());

		// Inst doesnt have untyped arg(s)
		default:
			throw new TrekException(
				"passStack(String desc) only for untyped stack values");
		}
	}

	// Pass instruction's stack arguments -- if any and typed
	// Notes:
	//	wide is seen as the underlying opcode
	public void passStack()
	{
		// Instruction is insert pt
		if (inst==null)
			throw new TrekException("Insert point not set at an instruction");
		switch (inst.opcode) {

		// Loading array element
		case Code.iaload: case Code.laload: case Code.faload: case Code.daload:
		case Code.aaload: case Code.baload: case Code.caload: case Code.saload:
			pstkBaseIndex();
			break;

		// Storing a result in a local variable
		case Code.istore: case Code.lstore: case Code.fstore:
		case Code.dstore: case Code.astore:
		case Code.istore_0: case Code.lstore_0: case Code.fstore_0:
		case Code.dstore_0: case Code.astore_0:
		case Code.istore_1: case Code.lstore_1: case Code.fstore_1:
		case Code.dstore_1: case Code.astore_1:
		case Code.istore_2: case Code.lstore_2: case Code.fstore_2:
		case Code.dstore_2: case Code.astore_2:
		case Code.istore_3: case Code.lstore_3: case Code.fstore_3:
		case Code.dstore_3: case Code.astore_3:
			pstkDataType(1);
			break;

		// Store array element
		case Code.iastore: case Code.lastore: case Code.fastore: case Code.dastore:
		case Code.aastore: case Code.bastore: case Code.castore: case Code.sastore:
			pstkBaseIndexValue();
			break;

		// Arithmetic operations
		case Code.iadd: case Code.ladd: case Code.fadd: case Code.dadd:
		case Code.isub: case Code.lsub: case Code.fsub: case Code.dsub:
		case Code.imul: case Code.lmul: case Code.fmul: case Code.dmul:
		case Code.idiv: case Code.ldiv: case Code.fdiv: case Code.ddiv:
		case Code.irem: case Code.lrem: case Code.frem: case Code.drem:
			pstkDataType(2);
			break;
		case Code.ineg: case Code.lneg: case Code.fneg: case Code.dneg:
			pstkDataType(1);
			break;

		// Shifting operations
		case Code.ishl: case Code.ishr: case Code.iushr:
			pstkIntData(2);
			break;
		case Code.lshl: case Code.lshr: case Code.lushr:
			pstkLongInt();
			break;

		// Bitwise boolean oprs
		case Code.iand: case Code.land:
		case Code.ior: case Code.lor:
		case Code.ixor: case Code.lxor:
			pstkDataType(2);
			break;

		// Casting
		case Code.i2l: case Code.i2f: case Code.i2d:
		case Code.l2i: case Code.l2f: case Code.l2d:
		case Code.f2i: case Code.f2l: case Code.f2d:
		case Code.d2i: case Code.d2l: case Code.d2f:
		case Code.int2byte: case Code.int2char: case Code.int2short:
			pstkDataType(1);
			break;

		// Bunch of if-related instructions. 
		case Code.lcmp:
		case Code.fcmpl: case Code.fcmpg:
		case Code.dcmpl: case Code.dcmpg:
			pstkDataType(2);
			break;

		// 0-based ifs
		case Code.ifeq: case Code.ifne:
		case Code.iflt: case Code.ifge:
		case Code.ifgt: case Code.ifle:
			pstkIntData(1);
			break;

		// Full a op b if's
		case Code.if_icmpeq: case Code.if_icmpne:
		case Code.if_icmplt: case Code.if_icmpge:
		case Code.if_icmpgt: case Code.if_icmple:
			pstkIntData(2);
			break;
		case Code.if_acmpeq: case Code.if_acmpne:
			pstkObjectData(2);
			break;
		case Code.lookupswitch: case Code.tableswitch:
			pstkIntData(1);
			break;
		case Code.ireturn: case Code.lreturn: case Code.freturn:
		case Code.dreturn: case Code.areturn:
			pstkDataType(1);
			break;

		// Data members
		case Code.getfield:
			pstkObjectData(1);
			break;
		case Code.putstatic:
			pstkFieldData();
			break;
		case Code.putfield:
			pstkObjectField();
			break;

		// New objects
		case Code.newarray:
		case Code.anewarray:
			pstkIntData(1);
			break; 

		// Miscellaneous
		case Code.arraylength:
		case Code.athrow:
		case Code.checkcast:
		case Code.instanceof_:
		case Code.ifnull:
		case Code.ifnonnull:
		case Code.monitorenter:
		case Code.monitorexit:
			pstkObjectData(1);
			break;

		// Calls (variable length arg lists)
		case Code.invokestatic: case Code.invokespecial:
		case Code.invokeinterface: case Code.invokevirtual:
			passInvoke(false);
			break;

		// Untyped args: 1 32-bit value
		case Code.pop: case Code.swap:
		case Code.dup: case Code.dup_x1: case Code.dup_x2:
		
		// Untyped args: two 32-bit values, or a double or long
		case Code.pop2:
		case Code.dup2: case Code.dup2_x1: case Code.dup2_x2:
			throw new TrekException(
				"passStack(String desc) must be used for untyped stack values");

		// Instructions that have no stack input, so passStack is noop.
		// No stack usage at all: nop, goto, iinc, jsr, ret
		// Output args only: xload* (non-array), xconst*, ldc*, new

		// new* should be defined as having insert point value???
		}											// End switch
	}

	// Pass a string
	public void passString(String str)
	{
		addToArgList("Ljava/lang/String;");

		// Non-null string requires pool. Null string, just load a null
		if (str!=null)
			loadPoolConstant(StringConstant.add(str, cf));
		else newCode[newCodeLen++] = (byte)Code.aconst_null;
	}

	// Work methods for passStack fixed-length arg lists.
	// For multi-value cases, name order matches
	// push order. That is, rightmost term is thing on top of stack.
	private void pstkBaseIndex()
	{
		loadStackFast(objectDesc, "I");
	}
	private void pstkBaseIndexValue()
	{
		String desc = inst.toDesc(objectDesc);
		loadTemps(objectDesc, "I", desc);
	}
	private void pstkDataType(int num)
	{
		String desc = inst.toDesc(objectDesc);
		loadStackFast(desc, num==1 ? null : desc);
	}
	private void pstkFieldData()
	{
		CpInfo cpi = inst.getMemberCp();
		loadStackFast(Datum.promote(cpi.toDesc()), null);
	}
	private void pstkIntData(int num)
	{
		loadStackFast("I", num==1 ? null : "I");
	}
	private void pstkLongInt()
	{
		loadStackFast("J", "I");
	}
	private void pstkObjectData(int num)
	{
		loadStackFast(objectDesc, num==1 ? null : objectDesc);
	}
	private void pstkObjectField()
	{
		CpInfo cpi = inst.getMemberCp();
		loadStackFast(objectDesc, Datum.promote(cpi.toDesc()));
	}

	// Gen either 32 or 64 bit stack-manipulating inst
	// Also returns false if 32 and true if 64
	private boolean writeStackInst(int opcode32, int opcode64)
	{
		if (Datum.getDataSize(ipDesc)==2)
		{
			newCode[newCodeLen++] = (byte)opcode64;
			return true;
		}
		else {
			newCode[newCodeLen++] = (byte)opcode32;
			return false;
		}
	}

}
