//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

package dec.trek;
import java.util.Vector;
import java.io.*;

//
//
// Case. Captures case-label pc's
//
//
final class Case implements Sortable
{
	private static final int dfault = -42380147;
									// Hack: dont match real case
	int match;						// Switch value that triggers this case
	int pc;							// PC this case starts at

	Case(int match, int pc)
	{
		this.match = match;
		this.pc = pc;
	}

	// Returns std 1,0,-1 from comparing This to arg
	// this = the object that is in the vector
	// obj = the object/value outside the vector
	public int compareTo(Object obj)
	{
		// Only inserts, no lookups
		Case cas = (Case)obj;
		if (pc < cas.pc)
			return -1;
		if (pc == cas.pc)
			return 0;
		else return 1;
	}


	// Find the next target address in a switch statement
	static Vector makeCases(Instruction inst)
	{
		Vector cases;
		int sw_offset = inst.parent.start + inst.start;
										// Offset of sw inst in its method
		int len;
		int pad = 3 - sw_offset % 4;
		int defoff = inst.readInt(inst.start + 1 + pad + 0);

		// Table switch (pc's for consec indexes)
		if (inst.opcode == Code.tableswitch)
		{
			int low = inst.readInt(inst.start + 1 + pad + 4);
			int high = inst.readInt(inst.start + 1 + pad + 8);
			int pc0 = inst.start + 1 + pad + 12;
			len = high - low + 1;
			cases = new Vector(len+1);
			for (int i=0; i<len; i++) {
				int casoff = inst.readInt(pc0 + i*4);
				Trek.insertSorted(cases, new Case(low+i, sw_offset+casoff));
			}
		}

		// Lookup switch (match value/pc pairs)
		else {
			int pc0 = inst.start + 1 + pad + 8;
			len = inst.readInt(inst.start + 1 + pad + 4);
			cases = new Vector(len+1);
			for (int i=0; i<len; i++) {
				int match = inst.readInt(pc0 + i*8);
				int casoff = inst.readInt(pc0 + i*8 + 4);
				Trek.insertSorted(cases, new Case(match, sw_offset+casoff));
			}
		}

		// Insert the default-branch pc and return computed vector
		Trek.insertSorted(cases, new Case(dfault, sw_offset+defoff));
		return cases;
	}

	// Build the src code: case n or default
	public String toString()
	{
		if (match != Case.dfault)
			return "case " + match + ":";
		else return "default:";
	}

	// Write a switch inst (excluding opcode byte), updating pc's in table
	// chan = where to write to
	// inst = the switch inst (its padding has already been updated)
	static void writeSwitch(DataOutput chan, Instruction inst)
		throws IOException
	{
		Statement stmt = inst.parent;
		int sw_offset = stmt.start + inst.start;
										// Orig offset of sw inst in its method

		// Write out already-adjusted pad and adjusted target of default:
		int pad = inst.getPad(0);		// Get adjusted padding
		int target = inst.readInt(inst.start + 1 + pad) + sw_offset;
		chan.write(stmt.insts, inst.start+1, pad);
		chan.writeInt(stmt.adjustJump(inst, target));

		// Table switch (pc's for consec indexes). Write out size info
		// and the table (with updated pc's)
		if (inst.opcode == Code.tableswitch)
		{
			int low = inst.readInt(inst.start + 1 + pad + 4);
			int high = inst.readInt(inst.start + 1 + pad + 8);
			int pc0 = inst.start + 1 + pad + 12;
			int len = high - low + 1;
			chan.write(stmt.insts, inst.start+1+pad+4, 8);

			// Write table of relative jumps
			for (int i=0; i<len; i++) {
				target = inst.readInt(pc0 + i*4) + sw_offset;
				chan.writeInt(stmt.adjustJump(inst, target));
			}
		}

		// Lookup switch (match value/pc pairs)
		else {
			int len = inst.readInt(inst.start + 1 + pad + 4);
			int pc0 = inst.start + 1 + pad + 8;
			chan.write(stmt.insts, inst.start+1+pad+4, 4);

			// Write table of match values and relative-jumps
			for (int i=0; i<len; i++) {
				chan.write(stmt.insts, pc0 + i*8, 4);	// Write the match val
				target = inst.readInt(pc0 + i*8 + 4) + sw_offset;
				chan.writeInt(stmt.adjustJump(inst, target));
			}
		}
	}

}

