//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

package dec.trek;

//
//
// Change. Parent to Code and Call
//
//
public abstract class Change
{
	// Allow load* to be used to get last program obj in chain or to load it
	private Object wrtObject;

	// Stuff for inserted code and related stuff.
	// Hack: just create big array to build code in
	protected ClassFile cf;			// Stmt's classfile
	protected String changeId;		// Id for locating this add's code
	protected Instruction inst;		// Ins pt is this instruction

	// Info about ins pt
	protected String ipDesc;		// Data type info for ins pt's value if any
	protected boolean ipElement;	// Ins pt has an array element
	protected String ipName;		// A name to access load-inst inspt var by
	protected String ipObject;		// Class of ins pt's object if any

	protected Method meth;			// Insert pt's method
	protected int newCodeLen;				// Bytes of insts in setup/code
	protected byte newCode[] = new byte[1000];
	int offset;						// Byte offset of exactly known insert pos
	int priority;					// Priority of new code, see Statement.plev_*
	protected Statement stmt;		// Statement being modified

	protected static final String objectDesc = "Ljava/lang/Object;";

	// Discard the code being built. Currently this is just garbage collection,
	// so java's auto gc makes this method a noop.
	public void discard()
	{
	}

	// Done building code, insert the code
	public abstract void done();

	// Shared part of done method
	protected void doneInsert()
	{
		// Stmt-ize new code to insert it, and permanize its change ID
		int type = offset<0 ? offset : Trek.ST_AT;
		Statement stadd = new Statement(meth, newCodeLen, newCode, priority, type);
		stadd.string = changeId;

		// Determine its impact on method's max stack size
		int stkdelta = 0;
		int maxdelta = 0;
		for (Instruction in=stadd.firstInstruction(); in!=null; in=in.next()) {
			stkdelta += in.getStackDelta();
			if (stkdelta>maxdelta)
				maxdelta = stkdelta;
		}
		meth.addStack = Math.max(maxdelta, meth.addStack);

		// Determine successsor to new code, splitting stmt if needed.
		// Add the new code to the method & mark classfile as modified.
		Statement succ = stmt.split(this);
		stadd.insert(succ);
		cf.modified = true;
	}

	// Generate instructions to load args and return "(arglist desc)"
	// args = set of var-refs separated by commas, and term by rparen
	// refdesc = non-null if name accessed via an object
	// thisload = where to return where "this" loaded if done
	protected String doArgs(String args, String refdesc, int thisload[])
	{
		// Leftmost term. Could be static or could be accessed via "this".
		// Wont know til lookup method, and cant do that until have desc
		// of arglist to determine proper incarnation of overloaded name.
		// So do hack of presuming instance-ref now, and fixing up in 
		// loadMethod
		if (refdesc==null)
		{
			thisload[0] = newCodeLen;
			loadThis();
		}

		// Analyze/load the arglist
		String argdesc = "";
		int nextpos;
		for (int pos=0; pos<args.length(); pos=nextpos+1) {
			nextpos = Datum.smartIndexOf(args, 0, ",)", pos);
			if (nextpos<0)
				throw new TrekException("Bad arglist");
			String arg = args.substring(pos, nextpos);

			// Strip spaces. Empty arg is ok only if arglist is ()
			arg = arg.trim();
			if (arg.length()==0)
			{
				if (argdesc.length()!=0)
					throw new TrekException("Empty argument");
				break;
			}

			// Process curr arg, deciding if constant or variable ref
			String desc = doConstant(null, arg);
			if (desc!=null)
				argdesc += desc;
			else argdesc += doVarRef(arg);
		}
		return "(" + argdesc + ")";
	}

	// Load a constant value, deriving it from its string representation
	// according to the data type in desc -- if term2 is a constant.
	// desc1 = descriptor for var to compare to
	// term2 = the value to analyze: var, true, false, "xxx", 'xxx', number
	// Returns descriptor for term2 if it is constant, and null otherwise
	protected String doConstant(String desc1, String term2)
	{
		// String constant?
		int ch1 = term2.charAt(0);
		if ("\'\"".indexOf(ch1) >= 0)
		{
			int len = term2.length();
			if (ch1!=term2.charAt(len-1))
				throw new TrekException("String constant not terminated properly");
			return doString(term2.substring(1, len-1));
		}

		// Numeric constant? If so, desc1 determines what to convert to.
		// But if it doesnt exist, constant must be explicitly typed
		// Also allowing leading + even tho java doesnt
		else if ("0123456789.-+".indexOf(ch1) >= 0)
		{
			if (ch1=='+') term2 = term2.substring(1);
			if (desc1!=null)
				ch1 = desc1.charAt(0);
			else ch1 = term2.charAt(term2.length()-1);

			// Floating numbers are most general numeric format
			try {
				switch (ch1) {
				case 'd':
				case 'D':
					double dval = Double.valueOf(term2).doubleValue();
					loadPoolConstant(DoubleConstant.add(dval, cf));
					return "D";
				case 'f':
				case 'F':
					float fval = Float.valueOf(term2).floatValue();
					loadPoolConstant(FloatConstant.add(fval, cf));
					return "F";
				}
			}
			catch (Exception e) {
				throw new TrekException("Value to compare is bad numeric constant");
			}

			// For integers, src format and desc nomenclature not in sync
			if (desc1==null)
			{
				if (ch1=='l' || ch1=='L')
					ch1 = 'J';
				else ch1 = 'I';
			}

			// Must be integer for int or long
			try {
				switch (ch1) {
				case 'I':
					return loadInteger(Integer.valueOf(term2).intValue());
				case 'J':
					long lval = Long.valueOf(term2).longValue();
					loadPoolConstant(LongConstant.add(lval, cf));
					return "J";
				}
			}
			catch (Exception e) {
				throw new TrekException("Value to compare is bad integer constant");
			}

			// Fell thru switches. That is, var1 is not a numeric variable
			throw new TrekException(
				"Data types of compared values must be the same");
		}

		// Boolean constant
		else if (term2.equals("true"))
		{
			loadInteger(1);
			return "Z";
		}
		else if (term2.equals("false"))
		{
			loadInteger(0);
			return "Z";
		}

		// term2 is a name
		return null;
	}

	// Write insts to load a member or just get mem, and return its data type
	// name = name of field to load, or method(...)
	// qcname = fully-qualified internal-fmt of name's class,
	//			or null for ins pt's class
	// refdesc = if ref term already exists, its data type
	protected String doMember(String name, String qcname, String refdesc)
	{
		// Get owner's classfile or use insert pt's classfile
		ClassFile owner;
		if (qcname!=null)
			owner = cf.parent.getClassFile(qcname);
		else owner = cf;

		// Look up name in its owner's class (& superclasses).
		int thisload[] = {-1};
		Member mem;
		boolean ismeth = name.endsWith(")");
		if (ismeth)
		{
			int lparen = name.indexOf("(");
			if (lparen<0)
				throw new TrekException(name + " has bad argument list");

			// But first process arglist for method name. argdesc is
			// part of identifying method: to deal with overloading and
			// to verify signature.
			String arglist = name.substring(lparen+1);
			String argdesc = doArgs(arglist, refdesc, thisload);
			name = name.substring(0, lparen);
			mem = getMember(name, owner, argdesc, refdesc);
		}
		else mem = getMember(name, owner, null, refdesc);
		wrtObject = mem;

		// Member must be accessed from its actual class. That is, if owner
		// inherited field from a superclass, need fld-ref pool entry that
		// identifies superclass rather than owner. For load, need
		// full desc, for use of member's value need just retval desc.
		boolean is_static = mem.hasModifier(Trek.ACC_STATIC);
		String qname = mem.parent.toInternalName() + "/" + name;
		String desc = mem.toDesc();
		if (ismeth)
			loadMethod(qname, name, mem.toSignature(), thisload[0], is_static);
		else loadField(qname, name, desc, refdesc, is_static);

		// Make name's descriptor -- in actions as opposed to ref -- be canon
		// value. Note above the fieldref created needed the actual reference
		// type so that the pool entry of the real field is used.
		return Datum.promote(desc);
	}									// End doMember

	// Process name at specified position (slave of doVarRef)
	// name = overall name expression string
	// pos = index of next term to process (and updated to position of 
	//		next name in the name expression, or -1)
	// refdesc = data type of ref obj, or null for leftmost name
	// Returns data type of the name
	protected String doName(String name, int pos[], String refdesc)
	{
		// Access name at current position
		int initpos = pos[0];
		String term = getNextTerm(name, pos);

		// If there is already owner, then name must be a member of it.
		// So owner had better be an object rather than primitive type
		// Get owner's classfile and load the specified member
		if (refdesc!=null)
		{
			String qcname = Datum.getDescClass(refdesc);
			if (qcname==null)
				throw new TrekException(
					name.substring(0, initpos) + " does not identify an object");
			return doMember(term, qcname, refdesc);
		}

		// At leftmost name. 5 cases:
		// 1) Ins pt's value
		// 2) It can be @@ (ie. insert point's object)
		if (term.equals("@") || term.equals("@@"))
			return loadInsertPoint(term, null);

		// 3) A local
		Local loc;
		if ((loc=getLocal(term))!=null)
		{
			String desc = Datum.promote(loc.desc);
			wrtObject = loc;
			loadLocal(loc, desc);
			return desc;
		}

		// 4) A fully qualified member name. Classname must end with / and
		// components of package must be separated by /'s too.
		if (term.startsWith("/"))
		{
			int aft = term.indexOf('(');
			int bef = (aft>=0) ? term.lastIndexOf('/', aft) : term.lastIndexOf('/');
			if (bef<=1)
				throw new TrekException(term + " is not /class/member");
			String memname = term.substring(bef+1);
			String qcname = term.substring(1, bef);
			return doMember(memname, qcname, null);
		}

		// 5) Member of insert pt's class. However cant be non-static member
		// unless ins pt's method has a "this" (which is always local 0)
		return doMember(term, null, refdesc);
	}											// End doName

	// Load a string and returns descriptor for it
	// val = string to load
	protected String doString(String val)
	{
		loadPoolConstant(StringConstant.add(val, cf));
		return "Ljava/lang/String;";
	}

	// Analyze specified variable ref, returing last term's descriptor
	// name = variable reference to analyze
	protected String doVarRef(String name)
	{
		String desc = null;
		int pos[] = {0};
		while (pos[0]>=0) desc = doName(name, pos, desc);
		return desc;
	}

	// Returns the local def for specified name if it exists at insert pt
	// name = local name to lookup
	protected Local getLocal(String name)
	{
		int pc = stmt.start;

		// For store at end, at start otherwise
		if (inst!=null)
		{
			pc += inst.start;
			if (inst.type==Vmi.it_storelv)
				pc += inst.size;
		}

		// Access pt is either at start or end of insert pt's stmt
		else if (offset!=Trek.ST_BEFORE)
				pc += stmt.insts.length;
		return meth.getLocal(name, pc);
	}

	// Lookup name in current owner class, and its superclasses if neces
	// name = name of member to load
	// owner = classfile of its class
	// argdesc = non-null (and desc of arglist) if name is method
	// refdesc = non-null if name accessed via an object
	// Notes:
	//	One diff between source code and this logic. User has access to
	//	all private members since he can set context to the superclass
	//	directly if he wished to.
	protected Member getMember(String name, ClassFile owner,
		String argdesc, String refdesc)
	{
		// Setup string to identify member if lookup fails.
		// Note that bad arglist can cause lookup failure. In particular,
		// if an arg is itself a void method, the V that was placed in
		// argdesc will obviously match no method being looked up.
		String sig = name;
		if (argdesc!=null)
			sig += argdesc;

		// Recursively lookup name in its superclass chain
		ClassFile cowner = owner;
		do {
			Member mem;
			if (argdesc!=null)
				mem = cowner.lookupMethod(name, argdesc);
			else mem = (Member)Trek.lookup(cowner.fields, name);
			if (mem!=null) return mem;

			// Recurse to owner class if can
			try {
				cowner = cowner.getSuperClass();
			}
			catch (Exception e) {
				throw new TrekException(
					"Could not determine if " + name 
					+ " was defined (super class lookup failed)");
			}
		} while (cowner!=null);

		// No such member. If no owner yet, just say undefined.
		// Else say "not a member"
		if (refdesc==null)
			throw new TrekException(name + " is not defined");
		else throw new TrekException(
			sig + " is not a member of " + owner.toQualifiedName());
	}

	// Get next dot-separated term from variable-ref
	// name = the overall ref string
	// pos = 1-item array containing where to get next term from
	protected String getNextTerm(String name, int pos[])
	{
		int start = pos[0];
		int dotidx = Datum.smartIndexOf(name, 0, ".", start);
		pos[0] = dotidx<0 ? -1 : dotidx + 1;

		// Process name at current position
		if (dotidx<0)
			return name.substring(start);
		else return name.substring(start, dotidx);
	}

	// Get target obj of variable ref. Same syntax as in Java source except
	// for 2 addit shorthands: @ = insert pt's value if any. @@ = its obj if any
	Object getObject(String name)
	{
		doVarRef(name);
		return wrtObject;
	}

	// Work method for stmt.hasChange(chid) and stmt.undoChange(chid)
	// chid = change id to look for: null=any, exact, or xxx* for starts-with
	// haschk = true if chking, false if undoing
	// Returns true if stmt has matching change
	static boolean handleChange(String chid, Statement stmt, boolean haschk)
	{
		// Estab marker past curr stmt
		Statement stnew = stmt.next();
		stnew = stnew==null ? stnew : stnew.firstChunk();

		// Setup to process wild or exact chid
		boolean exact = true;
		if (chid!=null && chid.endsWith("*"))
		{
			exact = false;
			chid = chid.substring(0, chid.length()-1);
		}

		// Scan for fragments that match specified change id
		boolean matched = false;
		boolean stillsplit = false;
		Statement stn = null;
		for (Statement st=stmt.firstChunk(); st!=stnew; st=stn) {
			stn = st.next;

			// Ignore real stmt content (orig stmt or undel piece)
			if (st.type>Trek.ST_PIECE
				|| (st.type==Trek.ST_PIECE && st.insts.length>0))
				continue;

			// If curr chunk matches chid, ret true or remove change
			if (chid==null
				|| (exact && chid.equals(st.string))
				|| (!exact && st.string.startsWith(chid)))
			{
				if (haschk) return true;

				// Remove the change:
				// 1) For insert or del-stmt, just remove it. (In latter
				// case, st will be mate since by def it is only piece)
				// 2) for del-inst, get the inst back from loan.
				else {
					matched = true;
					if (st.loan==null)
					{
						st.remove();
						if (stmt.mate==st) stmt.mate = null;
					}
					else {
						st.insts = st.loan.insts;
						st.loan = null;
						st.addr = 0;
					}
				}
			}

			// If at least 1 at-insertion or deleted piece remains,
			// then stmt must remain split into pieces
			else if (st.type==Trek.ST_PIECE || st.type==Trek.ST_AT)
				stillsplit = true;
		}

		// No matches or finished undoing matching changes. Treat undo as a
		// change since changes, saveTrek, undo a change, saveTrek is possible
		if (haschk)
			return false;
		if (matched)
			stmt.parent.parent.modified = true;

		// If undid all interior changes, go back to unsplit stmt
		if (stillsplit) return matched;
		for (Statement st=stmt.next; st!=stnew && st.type!=Trek.ST_AFTER; st=stn) {
			stn = st.next;
			st.remove();
		}
		stmt.mate = null;
		return matched;
	}

	// Load or do noop acton, depending of type of object
	protected abstract String loadInsertPoint(String term, String refdesc);
	protected abstract void loadField(String qname, String name, String desc,
		String refdesc, boolean is_static);
	protected abstract String loadInteger(int ival);
	protected abstract void loadLocal(Local loc, String desc);
	protected abstract void loadMethod(String qname, String name, String desc,
		int thisload, boolean is_static);
	protected abstract void loadPoolConstant(CpInfo cpi);
	protected abstract void loadThis();

	// Set ip* from instruction if can
	protected void setAtIp()
	{
		String desc;
		Local loc;
		CpInfo cpi = null;

		// See if this inst has a value
		if ((loc = inst.getLocal()) != null)
			desc = loc.desc;
		else if ((cpi = inst.getMemberCp()) != null)
			desc = cpi.toDesc();
		else return;

		// See if at instruction that is assoc with an object.
		// See if at array-element rather than at fld or local. If so,
		// strip leading [ to get data type of array elements.
		if (inst.opcode==Code.getfield || inst.opcode==Code.putfield
			|| inst.opcode==Code.invokeinterface
			|| inst.opcode==Code.invokespecial
			|| inst.opcode==Code.invokevirtual)
			ipObject = cpi.toInternalClassName();
		else if (inst.type==Vmi.it_loadel || inst.type==Vmi.it_storel)
		{
			ipElement = true;
			desc = desc.substring(1);
		}
		desc = Datum.promote(desc);

		// Store insts have value on stack, loads dont (they're about to
		// put it there). So ins pt truly has a value for store's only.
		if (inst.type==Vmi.it_storel
			|| inst.type==Vmi.it_storefld
			|| inst.type==Vmi.it_storelv)
			ipDesc = desc;

		// But orig value can simply be accessed directly in that case via
		// @@.name (or just name for a local).
		// However treat iinc instruction (which doesnt use stack) as
		// a store by special casing it to set ins pt after the inst.
		else if (inst.type==Vmi.it_loadlv || inst.opcode==Code.iinc)
			ipName = loc.name;
		else if (inst.type==Vmi.it_loadfld && ipObject!=null)
			ipName = "@@." + cpi.toName();
		return;
	}

	// Set context associated with this insert pt
	// typ = st_before/after/at
	protected void setInsertPoint(int typ)
	{
		// Setup for code-insertions within stmt
		switch (typ) {
		case Trek.ST_BEFORE:
			offset = Trek.ST_BEFORE;
			break;

		// An offset of 0 is by def BEFORE. Special case iinc as
		// store by setting offset to after the inst.
		case Trek.ST_AT:
			setAtIp();
			offset = inst.start;
			if (this instanceof Call && inst.opcode==Code.iinc)
				offset += inst.size;
			if (offset==0)
				offset = Trek.ST_BEFORE;
			break;

		// Handle intra-switch/return/throw special cases
		case Trek.ST_AFTER:
			if (stmt.type==Trek.ST_RETURN || stmt.type==Trek.ST_THROW)
				setReturnIp();
			else if (stmt.type==Trek.ST_SWITCH)
			{
				offset = Trek.ST_AT;
				ipDesc = "I";
			}
			else offset = Trek.ST_AFTER;
		}
	}

	// Set ip* for addAfter(return-stmt);
	protected void setReturnIp()
	{
		priority = Statement.plev_return;

		// Find the return inst
		offset = Trek.ST_AT;
		for (int i=0; i<stmt.insts.length; i+=inst.size) {
			inst = new Instruction(stmt, i);

			// Get exception's/retval's data type, if any
			switch (inst.opcode) {
			case Code.athrow:
			case Code.areturn: ipDesc = objectDesc; return;
			case Code.ireturn: ipDesc = "I"; return;
			case Code.lreturn: ipDesc = "J"; return;
			case Code.freturn: ipDesc = "F"; return;
			case Code.dreturn: ipDesc = "D"; return;
			case Code.return_: offset = Trek.ST_BEFORE; return;
			}
		}
		throw new TrekException("No Return instruction found");
	}

	// Returns string representation of code
	String toString(int notpub)
	{
		if (inst!=null)
			return "At " + inst.toString() + " in " + stmt.toString();

		// With respect to a statement
		switch (offset) {
		case Trek.ST_BEFORE:
			return "Before " + stmt.toString();
		case Trek.ST_AFTER:
			return "After " + stmt.toString();
		default:
			return "Within " + stmt.toString();
		}
	}

	// Add instruction with 1 nominal unsigned 8-bit arg into inst stream.
	// opcode = inst's opcode
	// val = if value larger than 8 bits, must be an op supported by "wide"
	// Notes:
	//	Caller can pass signed small value by passing value&0xff
	protected void writeByteInst(int opcode, int val)
	{
		if ((val&0xffffff00)==0)
		{						// Normal case
			newCode[newCodeLen++] = (byte)opcode;
			newCode[newCodeLen++] = (byte)val;
			return;
		}

		// Wide value
		switch (opcode) {
		case Code.iload:
		case Code.istore:
		case Code.fload:
		case Code.fstore:
		case Code.aload:
		case Code.astore:
		case Code.dload:
		case Code.dstore:
		case Code.lload:
		case Code.lstore:
		case Code.ret:
			newCode[newCodeLen++] = (byte)Code.wide;
			writeShortInst(opcode, val);
			return;
		default:
			throw new TrekException("Argument has more than 8 significant bits");
		}
	}

	// Add (possibly wide) iinc to inst stream
	// locidx = loc being incr
	// delta = amt to incr by (signed)
	protected void writeIincInst(int locidx, int delta)
	{
		if (delta==(byte)delta && locidx<256)
		{
			newCode[newCodeLen++] = (byte)Code.iinc;
			newCode[newCodeLen++] = (byte)locidx;
			newCode[newCodeLen++] = (byte)delta;
		}
		else {
			newCode[newCodeLen++] = (byte)Code.wide;
			writeShortInst(Code.iinc, locidx);
			newCodeLen = Instruction.writeShort(delta, newCode, newCodeLen);
		}
	}

	// Add instruction with 16-bit arg into instruction stream.
	// opcode = inst's opcode
	// val = 16-bit value to hi-endian-ize
	protected void writeShortInst(int opcode, int val)
	{
		// Put high-order byte of ref, then low-order byte in newCode
		newCode[newCodeLen++] = (byte)opcode;
		newCodeLen = Instruction.writeShort(val, newCode, newCodeLen);
	}

}
