//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

package dec.trek;
import java.io.*;

class DoubleConstant extends ValueRef
{
	private double doubleVal;

	// Add scalar value entry
	DoubleConstant(double dval, ClassFile cf)
	{
		super(cf);
		doubleVal = dval;
		cf.addConstant(this);
		cf.addConstant(null);			// Acct for wasted-entry after double entry
	}

	// Add scalar value entry if doesnt exist. Either way, return ptr
	static DoubleConstant add(double dval, ClassFile cf)
	{
		DoubleConstant dc = get(dval, cf);
		if (dc!=null) return dc;
		return new DoubleConstant(dval, cf);
	}

	// Returns true if specified object has same type/value(s) as "this".
/*
	private boolean equals(CpInfo cpi)
	{
		return isSameType(cpi) && cpi.doubleVal==doubleVal;
	}
*/

	// Return matching scalar value entry, or null
	private static DoubleConstant get(double dval, ClassFile cf)
	{
		return (DoubleConstant)CpInfo.get(new Double(dval), cf);
	}

	// Returns object containing value of cpi, or null if non-value cp entry
	public Object getObject()
	{
		return new Double(doubleVal);
	}

	// Returns true if specified object has same type as "this".
	protected boolean isSameType(CpInfo cpi)
	{
		return cpi instanceof DoubleConstant;
	}

	// Returns unique key for specified object
	protected Object keyOf()
	{
		return new Double(doubleVal);	
	}

	// Returns descriptor of pool entry
	String toDesc()
	{
		return "D";
	}

	// Return a source-oriented string repr of the specified cp entry
	public String toString()
	{
		if (Double.isNaN(doubleVal))
			return "Double.NaN";
		else if (doubleVal==Double.NEGATIVE_INFINITY)
			return "Double.NEGATIVE_INFINITY";
		else if (doubleVal==Double.POSITIVE_INFINITY)
			return "Double.POSITIVE_INFINITY";
		else return "" + doubleVal + "D";
	}

	// For dump report
	protected String toTag()
	{
		return "(Double)   ";
	}

	// Write pool entry for this constant
	protected void write(DataOutput chan) throws IOException
	{
		// Write type of constant and dispatch to write rest
		chan.writeByte(Cp_double);
		chan.writeDouble(doubleVal);
	}

}

