//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

package dec.trek;
import java.io.*;

// Handles field ref constants.

public class FieldRef extends MemberRef
{
	// Add field ref
	FieldRef(int clidx, int natidx, ClassFile cf)
	{
		super(clidx, natidx, cf);
	}

	// Construct a start-scan cp entry
	FieldRef(ClassFile cf)
	{
		super(cf);
	}

	// Add a field ref, if it doesnt already exist.
	// Either way, return a matching pool entry.
	// name = member's name (including internal qualified class name)
	// desc = member's descriptor
	// cf = classfile want to know if ref is defined within
	protected static FieldRef add(String name, String desc, ClassFile cf)
	{
		return (FieldRef)MemberRef.add(Cp_fieldref, name, desc, cf);
	}

	// Scan for matching ref entry
	// name = its fully qualified internal name
	// desc = its internal data-type info
	// cf = classfile to search
	static FieldRef get(String name, String desc, ClassFile cf)
	{
		return (FieldRef)MemberRef.get(Cp_fieldref, name, desc, cf);
	}

	// Return the fieldref's field, or null
	public Field getObject()
	{
		return (Field)getObject(0);
	}

	// Returns true if specified object has same type as "this".
	protected boolean isSameType(CpInfo cpi)
	{
		return cpi instanceof FieldRef;
	}

	// Returns unique key for specified object
	protected Object keyOf()
	{
		return (char)Cp_fieldref + toInternalString();
	}

	// Scan for next entry of specified type of constant
	// Returns matching constant or null
	public FieldRef next()
	{
		return (FieldRef)nextOf();
	}

	// Return a source-oriented string repr of the specified cp entry
	public String toString()
	{
		NameAndType nat = (NameAndType)parent.constant(natIdx);
		Field field = new Field(parent, nat.nameIdx, nat.descIdx);
		return field.toString(getTypeRef().toString());
	}

	// For dump report
	protected String toTag()
	{
		return "(Field)    ";
	}

	// Write a method ref constant, preserving if from interface
	protected void write(DataOutput chan) throws IOException
	{
		chan.writeByte(Cp_fieldref);
		super.write(chan);
	}

}

