//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

package dec.trek;
import java.io.*;

class FloatConstant extends ValueRef
{
	private float floatVal;

	// Add scalar value entry
	FloatConstant(float fval, ClassFile cf)
	{
		super(cf);
		floatVal = fval;
		cf.addConstant(this);
	}

	// Add scalar value entry if doesnt exist. Either way, return ptr
	static FloatConstant add(float fval, ClassFile cf)
	{
		FloatConstant fc = get(fval, cf);
		if (fc!=null) return fc;
		return new FloatConstant(fval, cf);
	}

	// Returns true if specified object has same type/value(s) as "this".
/*
	boolean equals(CpInfo cpi)
	{
		return isSameType(cpi) && cpi.floatVal==floatVal;
	}
*/

	// Return matching scalar value entry, or null
	private static FloatConstant get(float fval, ClassFile cf)
	{
		return (FloatConstant)CpInfo.get(new Float(fval), cf);
	}

	// Returns object containing value of cpi, or null if non-value cp entry
	public Object getObject()
	{
		return new Float(floatVal);
	}

	// Returns true if specified object has same type as "this".
	protected boolean isSameType(CpInfo cpi)
	{
		return cpi instanceof FloatConstant;
	}

	// Returns unique key for specified object
	protected Object keyOf()
	{
		return new Float(floatVal);	
	}

	// Returns descriptor of pool entry
	String toDesc()
	{
		return "F";
	}

	// Return a source-oriented string repr of the specified cp entry
	public String toString()
	{
		if (Float.isNaN(floatVal))
			return "Float.NaN";
		else if (floatVal==Float.NEGATIVE_INFINITY)
			return "Float.NEGATIVE_INFINITY";
		else if (floatVal==Float.POSITIVE_INFINITY)
			return "Float.POSITIVE_INFINITY";
		else return "" + floatVal + "F";
	}

	// For dump report
	protected String toTag()
	{
		return "(Float)    ";
	}

	// Write pool entry for this constant
	protected void write(DataOutput chan) throws IOException
	{
		// Write type of constant and dispatch to write rest
		chan.writeByte(Cp_float);
		chan.writeFloat(floatVal);
	}

}

