//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

package dec.trek;
import java.util.Vector;

//
//
// Instruction
//
//
public final class Instruction 
{

	// Instruction cache for a primordial stmt (owned by cache())
	private Vector icache;
	private int idx;

	// Info "pointed to" by inst
	private CpInfo cpi;					// Member pool entry reffed, if any
	private Local loc;					// Local reffed by inst, if any

	// Properties of an instruction
	Statement parent;					// Its stmt
	int opcode;							// Opcode of this instruction
	int size;							// # of bytes in opcode & inst's args
	int start;							// Offset of inst within parent's insts
	int type;							// Type of instruction (it_*)

	// Types of instruction targets (ge 0 means the target itself)
	static final int tar_multi = -2;
	static final int tar_none = -1;

	// Types of instruction arglists.
	// Used to determine instruction size and for toString-ing an instruction
	private static final byte ia_absent = 0;		// No args
	private static final byte ia_br2 = 1;			// PC-relative imm value
	private static final byte ia_br4 = 2;
	private static final byte ia_cpbyte = 3;		// <128 constant pool entry idx
	private static final byte ia_cp = 4;			// Any constant pool index
	private static final byte ia_imm1 = 5;			// A 1-byte immediate abs value
	private static final byte ia_imm2 = 6;
	private static final byte ia_local = 7;			// Local variable
	private static final byte ia_c1i1u = 8;			// Cpi + unsigned imm byte
	private static final byte ia_l1i1 = 9;			// Local + imm byte
	private static final byte ia_tableswitch = 10;
	private static final byte ia_lookupswitch = 11;

	// Describes args of instructions that are part of instruction
	// (as opposed to on operand stack).
	private static final byte arglists[] = {
		ia_absent, ia_absent, ia_absent, ia_absent, ia_absent,
		ia_absent, ia_absent, ia_absent, ia_absent, ia_absent,	//9
		ia_absent, ia_absent, ia_absent, ia_absent, ia_absent,
		ia_absent, ia_imm1, ia_imm2, ia_cpbyte, ia_cp,	//19
		ia_cp, ia_local, ia_local, ia_local, ia_local, 
		ia_local, ia_absent, ia_absent, ia_absent, ia_absent,	//29
		ia_absent, ia_absent, ia_absent, ia_absent, ia_absent, 
		ia_absent, ia_absent, ia_absent, ia_absent, ia_absent,	//39
		ia_absent, ia_absent, ia_absent, ia_absent, ia_absent, 
		ia_absent, ia_absent, ia_absent, ia_absent, ia_absent,	//49
		ia_absent, ia_absent, ia_absent, ia_absent, ia_local, 
		ia_local, ia_local, ia_local, ia_local, ia_absent,	//59
		ia_absent, ia_absent, ia_absent, ia_absent, ia_absent, 
		ia_absent, ia_absent, ia_absent, ia_absent, ia_absent,	//69
		ia_absent, ia_absent, ia_absent, ia_absent, ia_absent, 
		ia_absent, ia_absent, ia_absent, ia_absent, ia_absent,	//79
		ia_absent, ia_absent, ia_absent, ia_absent, ia_absent, 
		ia_absent, ia_absent, ia_absent, ia_absent, ia_absent,	//89
		ia_absent, ia_absent, ia_absent, ia_absent, ia_absent, 
		ia_absent, ia_absent, ia_absent, ia_absent, ia_absent,	//99
		ia_absent, ia_absent, ia_absent, ia_absent, ia_absent, 
		ia_absent, ia_absent, ia_absent, ia_absent, ia_absent,	//109
		ia_absent, ia_absent, ia_absent, ia_absent, ia_absent, 
		ia_absent, ia_absent, ia_absent, ia_absent, ia_absent,	//119
		ia_absent, ia_absent, ia_absent, ia_absent, ia_absent, 
		ia_absent, ia_absent, ia_absent, ia_absent, ia_absent,	//129
		ia_absent, ia_absent, ia_l1i1, ia_absent, ia_absent, 
		ia_absent, ia_absent, ia_absent, ia_absent, ia_absent,	//139
		ia_absent, ia_absent, ia_absent, ia_absent, ia_absent, 
		ia_absent, ia_absent, ia_absent, ia_absent, ia_absent,	//149
		ia_absent, ia_absent, ia_absent, ia_br2, ia_br2, 
		ia_br2, ia_br2, ia_br2, ia_br2, ia_br2,	//159
		ia_br2, ia_br2, ia_br2, ia_br2, ia_br2, 
		ia_br2, ia_br2, ia_br2, ia_br2, ia_local,	//169
		ia_tableswitch, ia_lookupswitch, ia_absent, ia_absent, ia_absent, 
		ia_absent, ia_absent, ia_absent, ia_cp, ia_cp,	//179
		ia_cp, ia_cp, ia_cp, ia_cp, ia_cp, 
		ia_c1i1u, ia_absent, ia_cp, ia_imm1, ia_cp,	//189
		ia_absent, ia_absent, ia_cp, ia_cp, ia_absent, 
		ia_absent, 0, ia_c1i1u, ia_br2, ia_br2,	//199
		ia_br4, ia_br4, ia_absent, ia_absent, ia_absent,		// Rest for future
		ia_absent, ia_absent, ia_absent, ia_absent, ia_absent,	//209
		ia_absent, ia_absent, ia_absent, ia_absent, ia_absent,
		ia_absent, ia_absent, ia_absent, ia_absent, ia_absent,	//219 
		ia_absent, ia_absent, ia_absent, ia_absent, ia_absent, 
		ia_absent, ia_absent, ia_absent, ia_absent, ia_absent,	//229
		ia_absent, ia_absent, ia_absent, ia_absent, ia_absent, 
		ia_absent, ia_absent, ia_absent, ia_absent, ia_absent,	//239
		ia_absent, ia_absent, ia_absent, ia_absent, ia_absent, 
		ia_absent, ia_absent, ia_absent, ia_absent, ia_absent,	//249
		ia_absent, ia_absent, ia_absent, ia_absent, ia_absent, ia_absent
	};

	// For toString of Instruction
	private static final String names[] = {
		"nop", "aconst_null", "iconst_m1", "iconst_0", "iconst_1",
		"iconst_2", "iconst_3", "iconst_4", "iconst_5", "lconst_0",	//9
		"lconst_1", "fconst_0", "fconst_1", "fconst_2", "dconst_0",
		"dconst_1", "bipush", "sipush", "ldc", "ldc_w",				//19
		"ldc2_w", "iload", "lload", "fload", "dload",
		"aload", "iload_0", "iload_1", "iload_2", "iload_3",			//29
		"lload_0", "lload_1", "lload_2", "lload_3", "fload_0",
		"fload_1", "fload_2", "fload_3", "dload_0", "dload_1",			//39
		"dload_2", "dload_3", "aload_0", "aload_1", "aload_2",
		"aload_3", "iaload", "laload", "faload", "daload",				//49
		"aaload", "baload", "caload", "saload", "istore",
		"lstore", "fstore", "dstore", "astore", "istore_0",			//59
		"istore_1", "istore_2", "istore_3", "lstore_0", "lstore_1",
		"lstore_2", "lstore_3", "fstore_0", "fstore_1", "fstore_2",	//69
		"fstore_3", "dstore_0", "dstore_1", "dstore_2", "dstore_3",
		"astore_0", "astore_1", "astore_2", "astore_3", "iastore",		//79
		"lastore", "fastore", "dastore", "aastore", "bastore",
		"castore", "sastore", "pop", "pop2", "dup",					//89
		"dup_x1", "dup_x2", "dup2", "dup2_x1", "dup2_x2",
		"swap", "iadd", "ladd", "fadd", "dadd",						//99
		"isub", "lsub", "fsub", "dsub", "imul",
		"lmul", "fmul", "dmul", "idiv", "ldiv",						//109
		"fdiv", "ddiv", "irem", "lrem", "frem",
		"drem", "ineg", "lneg", "fneg", "dneg",						//119
		"ishl", "lshl", "ishr", "lshr", "iushr",
		"lushr", "iand", "land", "ior", "lor",							//129
		"ixor", "lxor", "iinc", "i2l", "i2f",
		"i2d", "l2i", "l2f", "l2d", "f2i",								//139
		"f2l", "f2d", "d2i", "d2l", "d2f",
		"int2byte", "int2char", "int2short", "lcmp", "fcmpl",			//149
		"fcmpg", "dcmpl", "dcmpg", "ifeq", "ifne",
		"iflt", "ifge", "ifgt", "ifle", "if_icmpeq",					//159
		"if_icmpne", "if_icmplt", "if_icmpge", "if_icmpgt", "if_icmple",
		"if_acmpeq", "if_acmpne", "goto", "jsr", "ret",				//169
		"tableswitch", "lookupswitch", "ireturn", "lreturn", "freturn",
		"dreturn", "areturn", "return", "getstatic", "putstatic",		//179
		"getfield", "putfield", "invokevirtual", "invokespecial", "invokestatic",
		"invokeinterface", "op186", "new", "newarray", "anewarray",	//189
		"arraylength", "athrow", "checkcast", "instanceof", "monitorenter",
		"monitorexit", "wide", "multianewarray", "ifnull", "ifnonnull",//199
		"goto_w", "jsr_w", "breakpoint", "op203", "op204",
		"op205", "op206", "op207", "op208", "op209",					//209
		"op210", "op211", "op212", "op213", "op214",
		"op215", "op216", "op217", "op218", "op219",					//219
		"op220", "op221", "op222", "op223", "op224",
		"op225", "op226", "op227", "op228", "op229",					//229
		"op230", "op231", "op232", "op233", "op234",
		"op235", "op236", "op237", "op238", "op239",					//239
		"op240", "op241", "op242", "op243", "op244",
		"op245", "op246", "op247", "op248", "op249",					//249
		"op250", "op251", "op252", "op253", "imdep1", "imdep2"
	};

	// The amount that stack size changes during instruction's execution.
	// Note: call insts (incl jsr) eventually fall thru to next inst. It
	// is stack size when next inst is reached that is used in setting delta
	private static final int stackDelta[] = {
		0, 1, 1, 1, 1,
		1, 1, 1, 1, 2,		//9
		2, 1, 1, 1, 2,
		2, 1, 1, 1, 1,		//19
		2, 1, 2, 1, 2,
		1, 1, 1, 1, 1,		//29
		2, 2, 2, 2, 1,
		1, 1, 1, 2, 2,		//39
		2, 2, 1, 1, 1,
		1, -1, 0, -1, 0,	//49
		-1, -1, -1, -1, -1,
		-2, -1, -2, -1, -1,	//59
		-1, -1, -1, -2, -2,
		-2, -2, -1, -1, -1,	//69
		-1, -2, -2, -2, -2,
		-1, -1, -1, -1, -3,	//79
		-4, -3, -4, -3, -3,
		-3, -3, -1, -2, 1,	//89
		1, 1, 2, 2, 2,
		0, -1, -2, -1, -2,	//99
		-1, -2, -1, -2, -1,
		-2, -1, -2, -1, -2,	//109
		-1, -2, -1, -2, -1,
		-2, 0, 0, 0, 0,		//119
		-1, -1, -1, -1, -1,
		-1, -1, -2, -1, -2,	//129
		-1, -2, 0, 1, 0,
		1, -1, -1, 0, 0,	//139
		1, 1, -1, 0, -1,
		0, 0, 0, -3, -1,	//149
		-1, -3, -3, -1, -1,
		-1, -1, -1, -1, -2,	//159
		-2, -2, -2, -2, -2,
		-2, -2, 0, 0, 0,	//169
		-1, -1, -1, -2, -1,
		-2, -1, 0, 9, -9,	//179
		9, -9, -9, -9, -9,
		-9, 0, 1, 0, 0,		//189
		0, -1, 0, 0, -1,
		-1, 0, -9, -1, -1,	//199
		0, 0, 0, 0, 0,
		0, 0, 0, 0, 0,		//209
		0, 0, 0, 0, 0,
		0, 0, 0, 0, 0,		//219
		0, 0, 0, 0, 0,
		0, 0, 0, 0, 0,		//229
		0, 0, 0, 0, 0,
		0, 0, 0, 0, 0,		//239
		0, 0, 0, 0, 0,
		0, 0, 0, 0, 0,		//249
		0, 0, 0, 0, 0, 0
	};

	// The # of bytes in instruction (ie. 1 for opcode & n for imbedded args),
	// wrt to each type of arglist.
	private static final int sizes[] = {
		1,		// No args			[0]
		3,		// 2 branch bytes
		5,		// 4 branch bytes
		2,		// Narrow cp		[3]
		3,		// Normal cp
		2,		// 1 imm byte
		3,		// 2 imm bytes		[6]
		2,		// Local var
		4,		// cp + 1 imm byte
		3,		// Lv + 1 imm byte	[9]
		-1,		// Variable size
		-1,		// Variable size
		-1		// Variable size	[12]
	};

	// Find the instruction that starts at target in parent.insts
	// inst = where to start scanning for desired inst
	static Instruction at(Instruction inst, int target)
	{
		// Target is later. Scan forward
		if (target>inst.start)
		{
			int len = inst.icache.size();
			for (int i=inst.idx+1; i<len; i++) {
				inst = (Instruction)inst.icache.elementAt(i);
				if (inst.start==target) return inst;
			}
		}

		// Target is earlier. Scan backward
		else {
			for (int i=inst.idx; i>=0; i--) {
				inst = (Instruction)inst.icache.elementAt(i);
				if (inst.start==target) return inst;
			}
		}
		return null;
	}

	// Cache all the insts of specified stmt, and return 1st
	// stmt = stmt to process
	// getlast = true if want to ret last
	static Instruction cache(Statement stmt, Vector cache)
	{
		// If empty method, done. If no cache, just ret 1st isnt
		int instlen = stmt.insts.length;
		if (instlen<=0) return null;

		// Cache each inst, using old storage where can
		Instruction inst;
		int cachelen = cache.size();
		int start = 0;
		for (int i=0; start<instlen; i++) {

			// If slot already allocated, use it. Else create new slot
			if (i<cachelen)
				inst = (Instruction)cache.elementAt(i);
			else {
				inst = new Instruction();
				cache.addElement(inst);
			}

			// Fill in this inst
			inst.inits(stmt, start);
			inst.idx = i;
			inst.icache = cache;
			start += inst.size;
		}

		// Return 1st inst
		return (Instruction)cache.firstElement();
	}

	// Sets the instruction's variable to its array element's array if applic
	// Notes:
	//	The get* calls are in this method cant cause infinitely recursive
	//	calls to checkArrayLoad since calls here guaran not for an element inst
	private void checkArrayLoad()
	{
		Instruction inst = getArrayLoad();
		if (inst==null) return;

		// Found the inst that loads the array base, assign its info to
		// inst that manipulates the array element
		cpi = inst.getMemberCp();
		if (cpi==null)
			loc = inst.getLocal();
	}

	// Delete instruction from a stmt (but in undoable fashion)
	// chid = change ID to mark deletion with
	// Notes:
	//	Mech is to create a 1-inst piece whose size is 0 and which
	//	pts at a loan stmt that contains original inst
	public void delete(String chid)
	{
		Statement st = parent;

		// Handle case of deleted inst being only inst of its stmt
		if (st.insts.length==size)
		{
			st.delete(chid);		// Just delete whole stmt
			return;
		}

		// Locate chunk whose 1st inst needs to be deleted
		Code code = new Code(chid, st, this, Trek.ST_AT);
		Statement succ = st.split(code);
		if (succ.addr==-succ.insts.length)
			return;						// Already deleted

		// If deleting last inst of its stmt, split just done has
		// already isolated inst being deleted. Other do a split
		// right after inst being deleted in order to isolate it
		Statement stdel;
		Instruction instn = next();
		if (instn!=null)
		{
			code.offset = instn.start;
			stdel = st.split(code).prev;
		}
		else stdel = succ;

		// Do actual delete and save info to make delete undoable (and findable)
		stdel.loan = new Statement(st.parent, 0, 0, null);
		stdel.loan.insts = stdel.insts;
		stdel.addr = -stdel.insts.length;
		stdel.insts = new byte[0];
		stdel.string = code.changeId;
		st.parent.parent.modified = true;
	}

	// Return # of words taken up by method's arglist
	private int getArgListSize()
	{
		// Get pool entry if havent yet
		if (cpi==null)
		{
			ClassFile refcf = parent.parent.parent;
			cpi = refcf.constant(getCpIdx());
		}

		// Get desc of method's arglist
		String desc = cpi.toSignature();
		desc = Method.toArgsDesc(desc);
		if (desc==null) return 0;

		// Accum size of each argument in desc.
		int pos = 0;
		int size = 0;
		while (true) {
			int len = Datum.getDescLength(desc, pos);
			if (len<=0) return size;
			String argdesc = desc.substring(pos, pos+len);
			size += Datum.getDataSize(argdesc);
			pos += len;
		}
	}

	// If this is an array-element instruction, this method returns the
	// instruction that loaded the array reference.
	public Instruction getArrayLoad()
	{
		// At <clinit> code which cant be decomposed or not at array element
		Method meth = parent.parent;
		if (meth.parent.clinit==meth)
			return null;
		if (type!=Vmi.it_loadel && type!=Vmi.it_storel)
			return null;

		// At array element, do work. Arrload allows for nesting of array refs
		Instruction arrload[] = new Instruction[20];
		Instruction inst;
		int top = -1;

		// Look for latest array-load that is not "consumed" by an
		// array-element inst further down. Recall that
		// there is no permanent insts array, so cant chk inst=this to
		// tell when reach this. Instead chk starting offset
		for (inst=parent.firstInstruction(); inst.start!=start; inst=inst.next()) {
			if (top>=0 && (inst.type==Vmi.it_loadel || inst.type==Vmi.it_storel))
			{
				top -= 1;
				continue;
			}

			// If putting val on stack, setup desc to see if loading an array
			String desc;
			if (inst.type==Vmi.it_loadlv)
				desc = inst.getLocal().desc;
			else if (inst.type==Vmi.it_loadfld || inst.type==Vmi.it_call)
				desc = inst.getMemberCp().toDesc();

			// If dup-ing array load setup, capture this
			else if (inst.opcode==Code.dup2 && top>=0)
			{
				arrload[top+1] = arrload[top];
				top += 1;
				continue;
			}
			else continue;

			// If loading array, mark this. But mark it with empty entry
			// if it is retval rather than a load inst
			if (desc.charAt(0)=='[')
				arrload[++top] = inst.type==Vmi.it_call ? null : inst;
		}

		// If found the load at the array-element's "level", use it below.
		// Inst is null if retval rather than load is where array was from
		if (top<0) return null;
		inst = arrload[top];
		return inst;
	}

	// Returns object containing const data value of this inst if any
	// Or null if inst does not refer to constant.
	public Object getConstant()
	{
		switch (opcode) {

		// Specific constant implied by instruction
		case Code.aconst_null:
			return new Object();

		case Code.iconst_m1:
			return new Integer(-1);
		case Code.iconst_0:
			return new Integer(0);
		case Code.iconst_1:
			return new Integer(1);
		case Code.iconst_2:
			return new Integer(2);
		case Code.iconst_3:
			return new Integer(3);
		case Code.iconst_4:
			return new Integer(4);
		case Code.iconst_5:
			return new Integer(5);

		case Code.lconst_0:
			return new Long(0);
		case Code.lconst_1:
			return new Long(1);

		case Code.fconst_0:
			return new Float(0);
		case Code.fconst_1:
			return new Float(1);
		case Code.fconst_2:
			return new Float(2);

		case Code.dconst_0:
			return new Double(0);
		case Code.dconst_1:
			return new Double(1);

		// Get a value from instruction
		case Code.bipush:
			return new Integer(readByte(start+1));
		case Code.sipush:
			return new Integer(readShort(start+1));
		case Code.iinc:
			return new Integer(getIincDelta());

		case Code.invokeinterface:
		case Code.multianewarray:
			return new Integer(readUnsignedByte(start+3));
		case Code.newarray:
			return new Integer(readUnsignedByte(start+1));

		// Value is in cp
		case Code.ldc:
		case Code.ldc_w:
		case Code.ldc2_w:
			ClassFile cf = parent.parent.parent;
			ValueRef vref = (ValueRef)cf.constant(getCpIdx());
			return vref.getObject();
		}
		return null;
	}

	// If this instruction contains a pool reference, return it. Else ret -1
	int getCpIdx()
	{
		int arglist = arglists[opcode];
		switch (arglist) {
		case ia_cpbyte:
			return readUnsignedByte(start+1);			// 1-byte pool index
		case ia_cp:
		case ia_c1i1u:
			return readUnsignedShort(start+1);
		}
		return -1;
	}

	// For a member-using inst, calc # of words in data item it processes
	private int getDataSize()
	{
		// Get pool entry if havent yet
		if (cpi==null)
		{
			ClassFile refcf = parent.parent.parent;
			cpi = refcf.constant(getCpIdx());
		}

		// Return length based on item's type, incl support for void methods
		String desc = cpi.toDesc();
		return Datum.getDataSize(desc);
	}

	// Return iinc delta for this inst
	int getIincDelta()
	{
		if (size<4)
			return readByte(start+2);
		else return readShort(start+4);		// Wide iinc
	}

	// Returns local of this inst if any.
	// Returned local will be an array if this is an array elem load/store
	// of an array that is a local.
	public Local getLocal()
	{
		// Ret immed if already cached
		if (loc!=null)
			return loc;

		// Chk if raw ref there. If not, may need to find array element's
		// Otherwise just get the local for the returned index.
		int idx = getLocalIdx();
		if (idx < 0)
			checkArrayLoad();
		else getLocal(idx);
		return loc;
	}

	// Returns local's lvt entry or create free-standing Local for temp local
	// and sets up this.loc
	// idx = lvt idx of local to setup info for
	private Local getLocal(int idx)
	{
		loc = parent.parent.getLocal(idx, this);
		if (loc==null)
			loc = new Local(this, idx);
		return loc;
	}

	// If this instruction contains a lv reference, return it. Else ret -1
	int getLocalIdx()
	{
		switch (opcode) {
		case Code.aload_0:
		case Code.dload_0:
		case Code.fload_0:
		case Code.iload_0:
		case Code.lload_0:
		case Code.astore_0:
		case Code.dstore_0:
		case Code.fstore_0:
		case Code.istore_0:
		case Code.lstore_0:
			return 0;
		case Code.aload_1:
		case Code.dload_1:
		case Code.fload_1:
		case Code.iload_1:
		case Code.lload_1:
		case Code.astore_1:
		case Code.dstore_1:
		case Code.fstore_1:
		case Code.istore_1:
		case Code.lstore_1:
			return 1;
		case Code.aload_2:
		case Code.dload_2:
		case Code.fload_2:
		case Code.iload_2:
		case Code.lload_2:
		case Code.astore_2:
		case Code.dstore_2:
		case Code.fstore_2:
		case Code.istore_2:
		case Code.lstore_2:
			return 2;
		case Code.aload_3:
		case Code.dload_3:
		case Code.fload_3:
		case Code.iload_3:
		case Code.lload_3:
		case Code.astore_3:
		case Code.dstore_3:
		case Code.fstore_3:
		case Code.istore_3:
		case Code.lstore_3:
			return 3;
		case Code.ret:
		case Code.aload:
		case Code.dload:
		case Code.fload:
		case Code.iload:
		case Code.lload:
		case Code.astore:
		case Code.dstore:
		case Code.fstore:
		case Code.istore:
		case Code.lstore:
		case Code.iinc:
			if (size<4)
				return readUnsignedByte(start+1);
			else return readUnsignedShort(start+2);
		}
		return -1;
	}

	// Returns field or method object reffed by this inst. If refs mem,
	// but class inaccess, throw exception. If doest ref mem, return null.
	// Returned field will be the array if this is an array elem load/store
	// of an array that is a field.
	public Member getMember()
	{
		// If inst refs member, insure its cpi is setup.
		// Then access the member object, reading classfile if allowed
		if (getMemberCp()==null)
			return null;		// Doesnt ref member
		MemberRef mref = (MemberRef)cpi;
		return mref.getObject(0);
	}

	// Returns/insures inst's member-ref cpi is setup if inst refs member
	CpInfo getMemberCp()
	{
		if (cpi!=null) return cpi;

		// See if instruction has a pool index that is a member ref
		ClassFile refcf = parent.parent.parent;
		int idx = getCpIdx();
		if (idx >= 0)
		{							// Has pool idx at least
			CpInfo tcpi = refcf.constant(idx);
			if (!(tcpi instanceof MemberRef))
				return null;		// Array-loads dont have cpidx's, so done
			cpi = tcpi;

		}

		// If not, see if it refs an array that is a field
		else checkArrayLoad();
		return cpi;
	}

	// Get inst's opcode
	public int getOpcode()
	{
		return opcode;
	}

	// Returns amount of 0-padding switch has.
	// context = context to use in computing alignment. 3 cases:
	//		-1 = orig offset  0 = curr inst context  1 = use deltas
	//		For 0, orig stmt means dont use deltas & piece means use delta
	// Rule is that 1st addr past 1-byte opcode and n padding bytes must be
	// 32-bit aligned in its method
	int getPad(int context)
	{
		// True start of switch relative to its method depends on whether dealing
		// with orig method or modified method.
		int base = parent.start + start;		// Inst rel to its method
		switch (context) {
		case -1:
			return 3 - base%4;
		case 1:
			base += parent.delta + parent.preSize();
			return 3 - base%4;

		// Ret padding actual in inst, assuming switch's inst is last inst in
		// its stmt -- unless this is primordial stmt
		// (Note cant use this.size since it may not be setup yet).
		default:
			if (parent.type==Trek.ST_PRIMORD)
				return 3 - base%4;
			else return (parent.insts.length-start-1) % 4;
		}
	}

	// Returns method-relative start of this instruction
	public int getPc()
	{
		return start + parent.start;
	}

	// Returns what the instruction actually references, or null.
	public Object getReferenceArg()
	{
		// If already cached, use it
		if (loc!=null) return loc;
		if (cpi!=null) return cpi;

		// A pool entry?
		Method meth = parent.parent;
		int idx = getCpIdx();
		if (idx>0)
			return meth.parent.constant(idx);

		// A local?
		idx = getLocalIdx();
		if (idx>=0)
			return getLocal(idx);
		else return null;
	}

	// Returns # of bytes in this inst
	public int getSize()
	{
		return size;
	}

	// Returns type of instruction
	int getStackDelta()
	{
		int objref = 1;

		// Add field of some type to stack
		switch (opcode) {
		case Code.getfield:
			return -objref + getDataSize();
		case Code.getstatic:
			return getDataSize();

		// Remove field of some type from stack
		case Code.putfield:
			return -objref - getDataSize();
		case Code.putstatic:
			return -getDataSize();

		// Subtract args from stack
		case Code.invokestatic:
			objref = 0;
		case Code.invokeinterface:
		case Code.invokespecial:
		case Code.invokevirtual:
			return -objref - getArgListSize() + getDataSize();

		// Subtract dim sizes from stack
		case Code.multianewarray:
			return -readUnsignedByte(start+3) + 1;

		// For most insts, effect on stack is always the same
		default:
			return stackDelta[opcode];
		}
	}

	// Returns its statement
	public Statement getStatement()
	{
		return parent;
	}

	// Return branch target of this inst, or -1 if non-branching, or -2 if
	// list of branches (ie. *switch)
	public int getTarget()
	{
		// If branch instruction, get target and make it relative to start
		// of method (rather than curr instruction)
		switch (opcode) {
			case Code.ifeq:
			case Code.ifne:
			case Code.iflt:
			case Code.ifge:
			case Code.ifgt:
			case Code.ifle:
			case Code.if_icmpeq:
			case Code.if_icmpne:
			case Code.if_icmplt:
			case Code.if_icmpge:
			case Code.if_icmpgt:
			case Code.if_icmple:
			case Code.if_acmpeq:
			case Code.if_acmpne:
			case Code.goto_:
			case Code.jsr:
			case Code.ifnull:
			case Code.ifnonnull:
				return parent.start + start + readShort(start+1);
			case Code.goto_w:
			case Code.jsr_w:
				return parent.start + start + readInt(start+1);
			case Code.tableswitch:
			case Code.lookupswitch:
				return tar_multi;
		}
		return tar_none;
	}										// end getTarget

	// Returns type of instruction
	public int getType()
	{
		return type;
	}

	// Return true if instruction has arguments
	static boolean hasArgs(int opcode)
	{
		return arglists[opcode]!=ia_absent;
	}

	// Construct an instruction, which means storing a position in a
	// bytecode array & materializing the encoded info about the opcode
	// AND opcode with FF to elim sign extension int=byte does.
	// stmt = instruction's parent
	// stidx = offset of this inst within its statement
	// Notes:
	//	Must explicitly set all fields (except that managed by cache logic)
	//	since Instruction objects can be reused
	private void inits(Statement stmt, int stidx)
	{
		// Set basic properties
		cpi = null;
		loc = null;
		parent = stmt;
		byte insts[] = parent.insts;
		start = stidx;

		// Adjust opcode of wide to be opcode of what is performed
		boolean wide = false;
		opcode = insts[start] & 0xff;
		if (opcode==Code.wide)
		{
			wide = true;
			opcode = insts[start+1] & 0xff;
		}
		type = Vmi.types[opcode];

		// Store size of inst.
		int arglist = arglists[opcode];
		switch (arglist) {

		// Switches are var-length, and pad in particular is defined ugly:
		// Pad amt is set to 4-byte align the 32-bit quantities that follow
		case ia_tableswitch:
			int pad = getPad(0);
			int low = readInt(start + 1 + pad + 4);
			int high = readInt(start + 1 + pad + 8);
			size = 1 + pad + 4*3 + 4*(high-low+1);
			break;
		case ia_lookupswitch:
			pad = getPad(0);
			int npairs = readInt(start + 1 + pad + 4);
			size = 1 + pad + 4*2 + 4*npairs*2;
			break;

		// One of these has unused 5th byte at end
		case ia_c1i1u:
			if (opcode==Code.invokeinterface)
				size = 5;
			else size = 4;
			break;

		// Do rest by lookup (or wide special case)
		default:
			if (wide)
				size = opcode==Code.iinc ? 6 : 4;
			else size = sizes[arglist];
		}
	}

	// Instruction constructors
	private Instruction() {}
	Instruction(Statement stmt, int stidx) {inits(stmt, stidx);}

	// Returns true if instruction is load-local-integer inst
	boolean isIntegerLoad()
	{
		return opcode==Code.iload
			|| (opcode>=Code.iload_0 && opcode<=Code.iload_3);
	}

	// Returns true if this instruction load-obj-ref inst
	boolean isRefLoad()
	{
		return opcode==Code.aload
			|| (opcode>=Code.aload_0 && opcode<=Code.aload_3);
	}

	// Returns true if this instruction store-obj-ref inst
	boolean isRefStore()
	{
		return opcode==Code.astore
			|| (opcode>=Code.astore_0 && opcode<=Code.astore_3);
	}

	// Returns true if instruction modifies a variable
	boolean modifies()
	{
		if (opcode==Code.iinc)
			return true;
		if (type>=Vmi.it_storel && type<=Vmi.it_storelv)
			return true;
		return false;
	}

	// Navigate to next instruction of stmt. If at end, ret null instead
	public Instruction next()
	{
		// Move to next instruction in statement
		int nxtstart = start + size;
		if (nxtstart >= parent.insts.length)
			return null;				// Curr inst was last of stmt

		// Use cache if can
		if (icache!=null)
			return (Instruction)icache.elementAt(idx+1);
		else return new Instruction(parent, nxtstart);
	}

	// Navigate to prev instruction of stmt. If at 1st inst, ret null instead
	public Instruction prev()
	{
		// If at beginning, return no more.
		if (start <= 0)
			return null;

		// Use cache if can
		if (icache!=null)
			return (Instruction)icache.elementAt(idx-1);

		// Find prev by brute force forward scan
		Instruction curri = new Instruction(parent, 0);
		while (curri.start+curri.size < start) curri = curri.next();
		return curri;
	}

	// Instruction stream is high-endian. Assemble bytes
	// Be very careful to not let low-order bytes be sign-extended
	byte readByte(int idx)
	{
		return parent.insts[idx];
	}
	short readShort(int idx)
	{
		byte insts[] = parent.insts;
		return (short)((insts[idx]<<8) | (insts[idx+1]&0xff));
	}
	int readInt(int idx)
	{
		byte insts[] = parent.insts;
		return (insts[idx]<<24) | ((insts[idx+1]<<16)&0xff0000) |
			((insts[idx+2]<<8)&0x00ff00) | (insts[idx+3]&0xff);
	}

	// For unsigned results, must suppress sign extension 
	int readUnsignedByte(int idx)
	{
		return parent.insts[idx] & 0xff;
	}
	int readUnsignedShort(int idx)
	{
		return ((parent.insts[idx]<<8) | (parent.insts[idx+1]&0xff)) & 0xffff;
	}

	// Returns desc string this inst implies about the data it operates on
	// sdesc = specific source descriptor if known. "U" says apply
	//		default if inst is such that sdesc needs to be used.
	//		(sdesc redundant/ignored except for int and ref insts).
	// Notes:
	//	B, C, S, can only happen for array elements.
	//	the one odd inst (iinc) wants to return int by luck
	String toDesc(String sdesc)
	{
		String desc;
		int ch = sdesc.charAt(0);
		switch (names[opcode].charAt(0)) {
		case 'r':	// the "ret" instruction
		case 'a':
			desc = ch=='L' || ch=='[' ? sdesc : "Ljava/lang/Object;";
			break;

		// Simple firm numeric types
		case 'd':	desc = "D";		break;
		case 'f':	desc = "F";		break;
		case 'l':	desc = "J";		break;

		// Possible only for array-element insts
		case 'b':	desc = "B";		break;
		case 'c':	desc = "C";		break;
		case 's':	desc = "S";		break;

		// Integer is complicated in that it is also used for booleans.
		case 'i':
		default:
			desc = ch=='Z' ? sdesc : "I";
			break;
		}
		return desc;
	}

	// Infer local name from instruction and return it
	// idx = lvt index want name for
	// Notes:
	//	B, C, S, can only happen for array elements.
	//	the one odd inst (iinc) wants to return int by luck
	String toLocal(int idx)
	{
		String pfx;
		char ch = names[opcode].charAt(0);
		switch (ch) {
		case 'r':	// the "ret" instruction
		case 'a':	pfx = "obj";	break;
		case 'b':	pfx = "byte";	break;
		case 'c':	pfx = "char";	break;
		case 'd':	pfx = "dbl";	break;
		case 'f':	pfx = "fl";		break;
		case 'i':	pfx = "int";	break;
		case 'l':	pfx = "long";	break;
		case 's':	pfx = "sh";		break;
		default:	pfx = "lv";
		}
		return pfx + idx;
	}

	// Returns fully qualified external form of member's class, or null
	public String toMemberClass()
	{
		// If inst refs member, insure its cpi is setup
		if (getMemberCp()==null)
			return null;		// Doesnt ref member
		String qcname = cpi.toInternalClassName();
		return qcname.replace('/', '.');
	}
	
	// Returns name of member reffed in inst, or null
	public String toMemberName()
	{
		// If inst refs member, insure its cpi is setup
		if (getMemberCp()==null)
			return null;		// Doesnt ref member
		return cpi.toName();
	}

	// Return name of instruction
	static String toName(int opcode) {return names[opcode]; }
	String toName()	{return names[opcode]; }

	// Returns source code of instruction
	public String toString()
	{
		String args = null;
		String name = names[opcode];
		Method meth = parent.parent;
		ClassFile cf = meth.parent;

		// Get string repr of its arglist
		switch (arglists[opcode]) {
		case ia_local:
			if ((Trek.debug&Trek.Db_raw) != 0)
				args = String.valueOf(getLocalIdx());
			else args = meth.toLocal(this, getLocalIdx());
			break;
		case ia_br2:
			int pc = parent.start + start + readShort(start+1);
			args = String.valueOf(pc);
			break;
		case ia_br4:
			pc = parent.start + start + readInt(start+1);
			args = String.valueOf(pc);
			break;

		// Convert pool entry to string. Raw number or name of entry
		case ia_cp:
		case ia_cpbyte:
			int idx = getCpIdx();
			CpInfo cpi = cf.constant(idx);
			if ((Trek.debug&Trek.Db_raw) != 0)
				args = String.valueOf(idx);
			else if (opcode==Code.getstatic || opcode==Code.putstatic
				|| opcode==Code.invokestatic)
				args = cpi.toStaticRef();
			else if (type==Vmi.it_call)
				args = cpi.toName();
			else args = cpi.toRef();
			break;
		case ia_imm1:
			args = String.valueOf(readByte(start+1));
			break;
		case ia_imm2:
			args = String.valueOf(readShort(start+1));
			break;
		case ia_c1i1u:
			idx = readUnsignedShort(start+1);
			cpi = cf.constant(idx);
			if ((Trek.debug&Trek.Db_raw) != 0)
				args = String.valueOf(idx);
			else args = cpi.toRef();
			args += "," + readUnsignedByte(start+3);
			break;
		case ia_l1i1:
			String lvstr = meth.toLocal(this, getLocalIdx());
			args = lvstr + "," + getIincDelta();
			break;

		// Variable-width instructions
		case ia_tableswitch:
			int pad = getPad(0);
			int dfault = parent.start + start + readInt(start + 1 + pad + 0);
			int low = readInt(start + 1 + pad + 4);
			int high = readInt(start + 1 + pad + 8);
			args = dfault + "," + low + "," + high;
			break;
		case ia_lookupswitch:
			pad = getPad(0);
			dfault = parent.start + start + readInt(start + 1 + pad + 0);
			int npairs = readInt(start + 1 + pad + 4);
			args = dfault + "," + npairs;
			break;

		// For insts load xload_i, display var name even tho not explicit arg
		default:
			idx = getLocalIdx();
			if (idx>=0)
				args = meth.toLocal(this, idx);
		}

		// All done with args. Also distinguish wide from regular insts
		if (args!=null)
			name += " " + args;
		int rawop = parent.insts[start] & 0xff;
		return rawop==Code.wide ? name + " (wide)" : name;
	}

	// Write an int into instruction array
	// Returns updated endpt of array
	static int writeInt(int val, byte insts[], int start)
	{
		insts[start++] = (byte)((val & 0xff000000) >> 24);
		insts[start++] = (byte)((val & 0xff0000) >> 16);
		insts[start++] = (byte)((val & 0xff00) >> 8);
		insts[start++] = (byte)(val & 0xff);
		return start;
	}

	// Write a short into instruction array
	// Returns updated endpt of array
	static int writeShort(int val, byte insts[], int start)
	{
		insts[start++] = (byte)((val & 0xff00) >> 8);
		insts[start++] = (byte)(val & 0xff);
		return start;
	}

}

