//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

package dec.trek;
import java.io.*;

class IntegerConstant extends ValueRef
{
	private int intVal;

	// Add scalar value entry
	IntegerConstant(int ival, ClassFile cf)
	{
		super(cf);
		intVal = ival;
		cf.addConstant(this);
	}

	// Hack: so can next() for Constant
	IntegerConstant(ClassFile cf)
	{
		super(cf);
	}

	// Add scalar value entry if doesnt exist. Either way, return ptr
	static IntegerConstant add(int ival, ClassFile cf)
	{
		IntegerConstant ic = get(ival, cf);
		if (ic!=null) return ic;
		return new IntegerConstant(ival, cf);
	}

	// Returns true if specified object has same type/value(s) as "this".
/*
	boolean equals(CpInfo cpi)
	{
		return isSameType(cpi) && cpi.intVal==intVal;
	}
*/

	// Return matching scalar value entry, or null
	private static IntegerConstant get(int ival, ClassFile cf)
	{
		return (IntegerConstant)CpInfo.get(new Integer(ival), cf);
	}

	// Returns object containing value of cpi, or null if non-value cp entry
	public Object getObject()
	{
		return new Integer(intVal);
	}

	// Returns true if specified object has same type as "this".
	protected boolean isSameType(CpInfo cpi)
	{
		return cpi instanceof IntegerConstant;
	}

	// Returns unique key for specified object
	protected Object keyOf()
	{
		return new Integer(intVal);	
	}

	// Returns descriptor of pool entry
	String toDesc()
	{
		return "I";
	}

	// Return a source-oriented string repr of the specified cp entry
	public String toString()
	{
		if (intVal==0x80000000)
			return "0x80000000";
		else return "" + intVal;
	}

	// For dump report
	protected String toTag()
	{
		return "(Integer)  ";
	}

	// Write pool entry for this constant
	protected void write(DataOutput chan) throws IOException
	{
		// Write type of constant and dispatch to write rest
		chan.writeByte(Cp_integer);
		chan.writeInt(intVal);
	}

}

