//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

package dec.trek;
import java.io.*;
import java.util.Vector;

//
//
// Local. Use locals variables in inst/stmt toString if available

//
//
public final class Local 
{
	public Object userDefined;		// User-owned field

	// The fields in a classfile
	private int rawStart;		// Only used when orig info needed
	private int rawLength;		// Only used when orig info needed
	private int nameIdx;		// Index of its name (0 for non-lvt)
	private int descIdx;		// Index of its data type (0 for non-lvt)
	int idx;					// Its local-var slot in its stack frame
								// Note that mult entries can have same idx

	// Can change to get around bad lvt info put out by a java compiler
	int startPc;				// Where usage of this local starts
	int length;					// Start + this is where usage ends

	// Other info
	int arrayIdx;				// Its index in lvt array
	Statement defStmt;			// Its defining stmt
	String desc;				// Its data type
	String name;				// Name of the local 
	Method parent;				// For navigation
	int soft;					// Used by buildLvt. <=0 is firm. >0 is soft
	
	// Returns number of bytes in local-vars table
	static int attrLength(Method method)
	{
		int len = 2;
		for (int i=0; i<method.locals.size(); i++) {
			Local loc = (Local)method.locals.elementAt(i);
			if (loc!=null)
				len += 10;
		}
		return len;
	}

	// Set dcl info of local from info in 2nd local
	// loc = where source info
	// dtm = attr info to help set name
	void copyDclInfo(Local from, Datum dtm)
	{
		soft = from.soft<=Datum.firm ? from.soft : Math.max(soft, from.soft);
		desc = from.desc;
		name = toNewName(dtm);
	}

	// Returns stmt where local is defined
	public Statement getDefinition()
	{
		return defStmt;
	}

	// Returns parent's object
	public Method getMethod()
	{
		return parent;
	}

	// Returns index of this local var within its array, not its index
	// in its method's stack frame (mult locals can have same stack pos)
	public int indexOf()
	{
		return arrayIdx;
	}

	// Returns true if the local's definition is active at specified pc
	// pc = pos in its method's code
	// chkdesc = insure data type matches too if non null
	// Notes:
	//	Args always active, so this chk really should be redundant with
	//	range chk. But there is compiler that puts out a bad range for args.
	boolean isActive(int pc)
	{
		return isActive(pc, null);
	}
	boolean isActive(int pc, String chkdesc)
	{
		if (((pc>=startPc && pc<=startPc+length) || isArg())
			&& (chkdesc==null || desc.equals(chkdesc)))
			return true;
		else return false;
	}

	// Returns true if this local is a user-visible arg (ie. not "this")
	// Notes:
	//	this + args are supposed to be the 1st locals in an lvt.
	//	Must use loop because arglist can have gaps because of 64-bit
	//	args && (in some compilers) catch args have a startPc of 0.
	public boolean isArg()
	{
		// Non-static args start with 1 because "this" has index 0.
		int firstarg = parent.hasModifier(Trek.ACC_STATIC) ? 0 : 1;
		int argcnt = parent.getArgCount();
		int seen = 0;

		// Find last arg
		for (int i=firstarg; seen<argcnt && i<=parent.maxLocals; i++) {
			//Local loc = parent.getLocal(i, 0);
			Local loc = (Local)parent.locals.elementAt(i);
			if (loc==null) continue;		// Empty slot from long/double
			if (loc.startPc!=0) continue;	// Not an arg (shouldnt happen)
			if (idx==i) return true;
			//if (idx == loc.idx) return true;
			seen += 1;
		}
		return false;
	}

	// Returns true if this local is an array
	public boolean isArray()
	{
		return desc.charAt(0)=='[' ? true : false;
	}

	// Returns true if this local is "this"
	public boolean isThis()
	{
		if (idx==0 && !parent.hasModifier(Trek.ACC_STATIC))
			return true;
		else return false;
	}

	// Allow user of Code class to create a temp local
	// nam = name to assign local
	// obj = source obj or descriptor (to get data type)
	// opcode = opcode of store instruction that will define it
	public Local(String nam, Object obj, Method meth)
	{
		// Do easy stuff
		parent = meth;
		startPc = 0;			// Scope-less, represent as whole method
		length = 1000*1000*1000;
		name = nam;
		arrayIdx = -1;			// Not placed in lvt

		// Set data type stuff
		if (obj instanceof Member)
			desc = ((Member)obj).toDesc();
		else if (obj instanceof CpInfo)
			desc = ((CpInfo)obj).toDesc();

		// Obj is itself the descriptor. toDef used just to chk its syntax
		else if (obj instanceof String)
		{
			desc = (String)obj;
			meth.parent.toDef(null, desc, 0);
		}

		// Got a valid something
		if (desc==null)
			throw new TrekException("Could not determine data type of " + nam);
		if (desc.equals("V"))
			throw new TrekException("Data type of "
				+ nam + " cannot be set to 'void'");

		// Set lv idx and update max according to local's data type,
		// but dont collide with temp locals used for inserted calls
		int addidx = meth.addLocals;
		if (meth.callLocalsBase>=0)
			addidx = Math.max(meth.callLocalsBase + 4, addidx);
		if (desc.equals("D") || desc.equals("J"))
			meth.addLocals = addidx + 2;
		else meth.addLocals = addidx + 1;
		idx = meth.maxLocals + addidx;
	}

	// Construct local for temp local reffed in an inst
	// inst = where it is used
	// lvidx = its lvt index
	Local(Instruction inst, int lvidx)
	{
		// Get context
		Method meth = inst.parent.parent;
		Statement last = meth.lastStatement();

		// Setup free-standing local (not in lvt)
		startPc = 0;			// Scope-less, represent as whole method
		length = last.start + last.insts.length;
		name = inst.toName(idx);
		desc = inst.toDesc("U");
		idx = lvidx;
		parent = meth;
		arrayIdx = -1;			// Not placed in lvt
	}

	// Construct a local def from usage info (eg. from method signature)
	// startpc = where this local starts to exist in its method
	// len = range of its existence
	// source = datum of source code
	// nam/namidx = its name (null says create)
	// des/descidx = its data type
	// lvidx = its entry # in lvt. Note that mult local's can have same
	//			entry -- as when 2 loops both define the name "i"
	// meth = local's method
	Local(int startpc, int len, String nam, String des, int lvidx, Method meth)
	{
		this(startpc, len, null, nam, des, lvidx, meth);
	}
	Local(int startpc, int len, Datum source,
		String nam, String des, int lvidx, Method meth)
	{
		startPc = startpc;
		length = len;
		desc = des;
		idx = lvidx;
		store(meth);
		name = nam==null ? toNewName(source) : nam;
	}

	// Construct a local from read-in lvt entry
	private Local(int startpc, int len,
		int namidx, int descidx, int lvidx, Method meth)
	{
		rawStart = startPc = startpc;
		rawLength = length = len;
		nameIdx = namidx;
		descIdx = descidx;
		idx = lvidx;

		// Imbed strings for compat with constructed locs (they dont have *Idx)
		ClassFile cf = meth.parent;
		name = cf.constant(nameIdx).toInternalString();
		desc = cf.constant(descIdx).toInternalString();
		store(meth);
	}

	// Returns the next local in the lvt, or null if no more
	public Local next()
	{
		for (int i=arrayIdx+1; i<parent.locals.size(); i++) {
			Local loc = (Local)parent.locals.elementAt(i);
			if (loc!=null) return loc;
		}
		return null;
	}

	// Returns the prev local in the lvt, or null if no more
	public Local prev()
	{
		for (int i=arrayIdx-1; i>=0; i--) {
			Local loc = (Local)parent.locals.elementAt(i);
			if (loc!=null) return loc;
		}
		return null;
	}

	// Read the specified method's local-var table
	// chan = where to read from
	// method = obj whose locals are to be read
	static void readLocals(DataInput chan, Method method)
		throws IOException
	{
		int i = -1;

		try {

			// Setup starting table size. Needed size is # of simultaneous
			// local defs -- incl gaps for longs, doubles, and (apparently)
			// Exceptions. This is maxLocals. Note that total locals, which
			// is lvRead, can be larger than this because mult locals can be
			// assigned to same lv idx when their scopes are disjoint. (Oddly
			// enough lvRead can sometimes be smaller than maxLocals because
			// it does not include the gaps that are included in maxLocals).
			method.lvRead = chan.readUnsignedShort();
			method.addLvt();

			// Read each locals entry
			for (i=0; i<method.lvRead; i++) {
				Local loc = new Local(
					chan.readUnsignedShort(),		// Start pc
					chan.readUnsignedShort(),		// Len of pc range
					chan.readUnsignedShort(),		// CP Index of its name
					chan.readUnsignedShort(),		// Index of its data type
					chan.readUnsignedShort(),		// Its index in lv table
					method);
			}									// End loop

			// Suppress compiler's lvt if debugging that area of code
			if ((Trek.debug&Trek.Db_bldloc)!=0)
			{
				method.locals = null;			// Discard what read
				method.lvRead = -2;				// Pre buildLvt
			}
		}										// End try

		// Gen msg for error if any
		catch (Exception e) {
			method.parent.logError(toContext("reading", method, i), e);
		}
		return;
	}

	// Remove the specified local
	void remove()
	{
		// Physically remove only if in constructed lvt. Compress vector
		// tho only if it is last entry -- so that arrayIdx of other overflow
		// locals are not invalidated. Otherwise just leave empty slot
		if (parent.lvRead<0)
		{
			if (arrayIdx==parent.locals.size()-1)
				parent.locals.removeElementAt(arrayIdx);
			else parent.locals.setElementAt(null, arrayIdx);
		}

		// Never truly wipe out entries created by compiler
		else {
			startPc = 0;
			length = -1;
		}
	}

	// Set defStmt and reflect it into startPc
	void setDefStatement(Statement stmt)
	{
		defStmt = stmt;

		// See if this location implies larger scope
		int stend = stmt.start+stmt.insts.length;
		if (startPc>stend)
		{
			length += startPc - stend;
			startPc = stend;
		}
	}

	// Store new local in meth.locals...
	// Deal with case when same name used for diff vars in
	// diff parts of method. Code is optimized for no such
	// dups. When they occur, stick them at end of array
	private void store(Method meth)
	{
		parent = meth;

		// The easy case? Just fill the slot
		if (meth.locals.elementAt(idx)==null)
		{
			arrayIdx = idx;
			meth.locals.setElementAt(this, idx);
		}

		// Overloaded index. Extend table
		else {
			arrayIdx = meth.locals.size();
			meth.locals.addElement(this);
		}
	}

	// Create context string for error during processing of lvt
	// doing = action being performed at time of exception
	// method = method whose exceptions are being processed
	// i = index of exception when problem occurred
	static private String toContext(String doing, Method meth, int i)
	{
		// Setup content of error and combine with context
		String mname = "M=" + meth.toName() + ";";
		String elem = "E=local[" + i + "];";
		return "D="+doing+";" + mname + elem;
	}

	// Returns name of this local
	public String toName()
	{
		return name;
	}
	
	// Returns a name for the local variable being defined
	// source = source of assign (not used)
	String toNewName(Datum source)
	{
		String pfx;
		int ch = desc.charAt(0);

		// Use the data type of dest and data type and source expression to
		// choose a name for the dest of this defining assignment stmt
		switch (ch) {
		case 'L':
			pfx = Datum.getDescClass(desc);
			int start = pfx.lastIndexOf('/');
			if (start >= 0) pfx = pfx.substring(start+1);
			pfx = pfx.toLowerCase();
			pfx = pfx.replace('$', '_');
			break;
		case 'B':	pfx = "byte";	break;
		case 'C':	pfx = "char";	break;
		case 'D':	pfx = "dbl";	break;
		case 'F':	pfx = "fl";		break;
		case 'I':	pfx = "int";	break;
		case 'J':	pfx = "long";	break;
		case 'S':	pfx = "sh";		break;
		case 'Z':	pfx = "bool";	break;
		case '[':	pfx = "arr";	break;
		default:	pfx = "lv";
		}

		// Construct unique name
		return pfx + arrayIdx;
	}										// toNewName

	// Return string that uniquely identifies this local (see Trek.getObject)
	public String toObjectId()
	{
		// Construct and return qualclass.method(desc).local#stmtnum
		int stnum = defStmt!=null ? defStmt.getNumber() : 0;
		return parent.toObjectId() + "." + name + "#" + stnum;
	}

	// Returns source ref for this local
	// cf = classfile of local
	// pc = pc of reference
	String toRef(ClassFile cf, int pc)
	{
		// If ref to loc is also its def, return data-type info if
		// local is arg (has no defining stmt) or its 1st usage defines it
		if (startPc==pc
			&& (defStmt==null || defStmt.type!=Trek.ST_LOCAL))
			return cf.toDef(name, desc, 0);
		else return name;
	}

	// Returns presentation string for this local (for user)
	public String toString()
	{
		return parent.parent.toDef(name, desc, 0);
	}
	
	// Returns presentation string for local attribute (for dump)
	String toString(ClassFile cf)
	{
		// Setup name of field and its type info.
		// Normally combine with pc range, and return. However in special
		// case of unused local, the startPc is short -1 (65535 int)
		String dcl = cf.toDef(name, desc, 0);
		String pfx = "  " + dcl + " (" + idx + ": ";
		if (rawStart==65535)
			return pfx + "unreferenced)";
		else if ((Trek.debug&Trek.Db_currloc) == 0)
		{
			int endpc = rawStart + rawLength;
			return pfx + rawStart + "-" + endpc + ")";
		}

		// Do current info as modified or constructed by Lvt class
		else {
			if (length<0)
				return pfx + "removed)";
			int endpc = startPc + length;
			return pfx + startPc + "-" + endpc + ")";
		}
	}

	// Returns external-form type of the local
	public String toType()
	{
		return parent.parent.toDef(null, desc, 1);
	}

	// Write the current object's local-var table
	// chan = where to write to
	// method = obj whose locals are to be written
	static void writeLocals(DataOutput chan, Method method)
		throws IOException
	{
		int i = -1;

		try {
			int len = method.lvRead;
			chan.writeInt(2 + len*10);			// Size of this attr
			chan.writeShort(len);				// # of locals
			if (len <= 0) return;

			// Write each local entry after insuring PC's uptodate
			for (i=0; i<method.locals.size(); i++) {
				Local loc = (Local)method.locals.elementAt(i);
				if (loc==null) continue;

				// Args start at 0 even if code added before 1st stmt
				int startpc = loc.rawStart;
				int endpc = startpc + loc.rawLength;
				if (startpc!=0)
					startpc = Statement.adjustPc(method, startpc);

				// Write out each field
				chan.writeShort(startpc);
				endpc = Statement.adjustPc(method, endpc);
				chan.writeShort(endpc-startpc);
				chan.writeShort(loc.nameIdx);
				chan.writeShort(loc.descIdx);
				chan.writeShort(loc.idx);
			}									// End loop
		}										// End try

		// Gen msg for error if any
		catch (Exception e) {
			method.parent.logError(toContext("writing", method, i), e);
		}
		return;
	}

}
