//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

package dec.trek;
import java.io.*;

class LongConstant extends ValueRef
{
	private long longVal;

	// Add scalar value entry
	LongConstant(long lval, ClassFile cf)
	{
		super(cf);
		longVal = lval;
		cf.addConstant(this);
		cf.addConstant(null);			// Acct for wasted-entry after double entry
	}

	// Add scalar value entry if doesnt exist. Either way, return ptr
	static LongConstant add(long lval, ClassFile cf)
	{
		LongConstant lc = get(lval, cf);
		if (lc!=null) return lc;
		return new LongConstant(lval, cf);
	}

	// Returns true if specified object has same type/value(s) as "this".
/*
	boolean equals(CpInfo cpi)
	{
		return isSameType(cpi) && cpi.longVal==longVal;
	}
*/

	// Return matching scalar value entry, or null
	private static LongConstant get(long lval, ClassFile cf)
	{
		return (LongConstant)CpInfo.get(new Long(lval), cf);
	}

	// Returns object containing value of cpi, or null if non-value cp entry
	public Object getObject()
	{
		return new Long(longVal);
	}

	// Returns true if specified object has same type as "this".
	protected boolean isSameType(CpInfo cpi)
	{
		return cpi instanceof LongConstant;
	}

	// Returns unique key for specified object
	protected Object keyOf()
	{
		return new Long(longVal);	
	}

	// Returns descriptor of pool entry
	String toDesc()
	{
		return "J";
	}

	// Return a source-oriented string repr of the specified cp entry
	public String toString()
	{
		if (longVal==0x8000000000000000L)
			return "0x8000000000000000L";
		else return "" + longVal + "L";
	}

	// For dump report
	protected String toTag()
	{
		return "(Long)     ";
	}

	// Write pool entry for this constant
	protected void write(DataOutput chan) throws IOException
	{
		// Write type of constant and dispatch to write rest
		chan.writeByte(Cp_long);
		chan.writeLong(longVal);
	}

}

