//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

package dec.trek;
import java.util.Hashtable;
import java.util.Vector;

//
// Lvt (accessed via declareLocals)
// If compiler did not supply lvt, analyze instructions that process 
// local variables and build lvt (ie. set name, desc, and scope of each local)
// Also has methods to cleanup bugs in compiler-supplied lvt.
// Either way, then create code-less declarations of locals as needed.
// Notes:
//	If merely know data type is integer at jvm-level, could be B, C, S, or Z
//	at src level. That is, more info to make desc solid
//
final class Lvt extends Vmi 
{
	Hashtable cfhash = new Hashtable();
	boolean pass2;				// See buildLvt
	boolean repeat;				// Set if an analyze firms up a local

	private static final ClassFile cfObject = new ClassFile(null);
	private static final ClassFile cfFail = new ClassFile(null);
	private static final int softval[] = {
		Datum.byplus, Datum.chplus, Datum.shplus, Datum.intonly
	};

	// Analyze stmt's usage of specified local (see Finish.declareLocals)
	// stmt = statement to analyize
	// locidx = index of local var to to analyze
	// Notes:
	//	Loans arent processed to save processing time. This is viewed as
	//	safe because loans are obj refs -- not the thing being used.
	//	Iow, in a = ref.b, it is b not ref that can influence type of a.
	private void analyze(Statement stmt)
	{
		// Analyze only stmt types that use user-defined locals
		switch (stmt.type) {
		case Trek.ST_ASSIGN:
		case Trek.ST_CALL:
		case Trek.ST_CATCH:
		case Trek.ST_ENDDO:
		case Trek.ST_ENDFOR:
		case Trek.ST_ENDWHILE:
		case Trek.ST_IF:
		case Trek.ST_RETURN:
		case Trek.ST_SWITCH:
		case Trek.ST_SYNC:
		case Trek.ST_THROW:
			break;
		default:
			return;
		}

		// Display status if so requested
		if ((Trek.debug&Trek.Db_atsmall) != 0)
			Trek.println(System.out,
				"Analyzing statement in " + stmt.parent.toName());

		// Scan thru statement's instructions, analyzing local. If local
		// firmed up, repeat the analysis with the newer context in place
		while (true) {
			repeat = false;
			init(stmt);
			executeLoan(stmt.loan);
			for (inst=stmt.firstInstruction(); ; inst=inst.next()) {
				if (inst==null) return;
				if (!xinst()) return;
				if (repeat) break;
			}
		}
	}

	// Store top of stack in specified local variable & store l/ropd
	// lidx = identifies local to store value into
	protected void assop(int lidx)
	{
		// Setup source and lookup loc
		Datum dest;
		Datum source = pop();
		Local loc = meth.getLocal(lidx, inst);

		// If no lvt entry, build one with initial scope ending at defeob.
		// Reads of this local past defeob can extend the scope
		if (loc==null)
		{
			dest = buildDest(lidx, source);
			loc = dest.loc;
		}

		// Local exists.
		else {
			dest = new Datum(loc);

			// It is possible to have dcl-only blks. If so, a new local with
			// same index can exist. For ex: exactly {int a=1;} {bool b=true;}
			// would assign a & b the same local index. If this case applies,
			// shorten scope of 1st loc to end here, and create the 2nd loc.
			// Note that invis blks dont create real blk, so old & new locals
			// have the same eob even tho they have disjoint (start, len)'s.
			if (isInvisNewBlock(loc, source))
			{
				dest = buildDest(lidx, source);

				// End 1st local right before where 2nd local defined -- and
				// need +inst.start because when optimizer in-lines methods,
				// it will use the same loc idx for the obj ref and the return
				// value of the method. Thus, the scope of a local can end in
				// the middle of a statement!
				loc.length =  inst.parent.start + inst.start - loc.startPc;
				loc = dest.loc;			// "Replace" shrunken looked-up loc
			}

			// If dont have new dcl-only blk and loc's type is soft, chk if
			// can firm it up
			else if (loc.soft>=Datum.softy)
				checkSoftResult(dest, source);
		}

		// Do the assign, setting result soft if local's desc not firm.
		checkSoftStatement(dest);
		assop(dest, source);
	}

	// Define and return destination datum and local
	// lidx = index of local to define
	// source = what it's being assigned from
	private Datum buildDest(int lidx, Datum source)
	{
		// Finally & Sync blks define some src-invisible locals.
		// Else find bottom of blk loc def is withing
		Statement stmt = inst.parent;
		Statement defeob;
		if (stmt.type==Trek.ST_FINALLY || stmt.type==Trek.ST_SYNC)
			defeob = stmt.mate;
		else if (stmt.type==Trek.ST_ENDSYNC)
			defeob = stmt;
		else defeob = stmt.getEob();

		// Build the lvt entry of local, setting pos and descriptor.
		int startpc = getStartPc();
		int endpc = defeob.start + defeob.insts.length;
		String desc = inst.toDesc(source.desc);
		Local loc = new Local(startpc, endpc - startpc,
			source, null, desc, lidx, meth);
		loc.soft = Datum.softy;
		loc.defStmt = defeob;

		// Setup dest datum and softness of it and local
		Datum dest = new Datum(loc);
		setResultSoftness(dest, source);
		return dest;
	}

	// Build a local-vars table for the specified method
	// Notes:
	//	While building lvt, defStmt is end of local's blk. This helps set
	//	the stage for setting it to its real value in declareLocals
	private void buildLvt()
	{
		// Analyze method's stmts to create each local & set its data type
		meth.lvRead = -1;			// Reached buildLvt
		Statement first = meth.firstStatement();
		for (Statement st=first; st!=null; st=st.next()) {
			analyze(st);
		}

		// If 1st pass left a local with a soft dcl, later data gathered
		// in 1st pass usually will allow analyze to firm it up now. If not,
		// go with soft dcl.
		pass2 = true;
		for (Statement st=first; st!=null; st=st.next()) {
			if (st.delta>0)
				analyze(st);
		}

		// Combine all redundant defs by doing pairwise combines
		for (int i=meth.maxLocals; i<meth.locals.size(); i++) {
			Local loc = (Local)meth.locals.elementAt(i);
			if (loc==null) continue;

			// If no 1st or it starts after 2nd, then 2nd is part of indep blk
			while (true) {
				Local locp = meth.getPrevLocal(loc.idx, loc);
				if (locp==null || locp.startPc>loc.startPc)
					break;

				// If start of 2nd local in source code is before start
				// of 1st local, merge them. If its start is in middle of
				// 1st local's scope, merge them only if not clearly diff.
				// If cant merge them, un-overlap  their scopes.
				// get2ndStart will set start2 earlier than loc.startPc
				// when loc's scope extends into broader blk.
				int start2 = get2ndStart(locp, loc);
				if (start2<locp.startPc+locp.length)
				{
					if (start2>locp.startPc
						&& is2ndClearlyDifferent(locp, loc))
					{
						locp.length = start2 - locp.startPc;
						break;
					}
					else mergeLocalProps(locp, loc);
				}

				// 2nd starts later. If it also ends later, they are in 
				// disjoint blocks, so dont merge. (Blks are not
				// disjoint if blk2 is nested in blk1 -- which would
				// show up as 1st's start being > 2nd's start)
				else if (locp.defStmt.start<loc.defStmt.start)
					break;

				// In same blk. Originally seemed that dcl-only blks needed.
				// If types now same or soft-compatible, merge them.
				else if (is2ndSubsumable(loc, locp))
						locp.copyDclInfo(loc, null);
				else if (is2ndSubsumable(locp, loc))
					;							// Using locp's info
				else if (!locp.desc.equals(loc.desc))
					break;

				// Fell thru. Merge by extending locp and removing loc.
				// Use firmer soft-value as softness of merged local.
				new Datum(locp).refineSoftness(loc.soft);
				int endpc = loc.startPc + loc.length;
				locp.length = endpc - locp.startPc;
				locp.defStmt = locp.defStmt.getCommonEob(loc.defStmt);
				loc.remove();

				// Finally see if any other dcl is affected by this change
				reanalyze(locp, loc.startPc, endpc);
				loc = locp;
			}									// End inner loop
		}										// End outer loop

		// Finally if an integer has done only boolean-safe stuff, convert it
		for (int i=0; i<meth.locals.size(); i++) {
			Local loc = (Local)meth.locals.elementAt(i);
			if (loc==null) continue;
			if (loc.soft==Datum.softy && loc.desc.equals("I"))
			{
				loc.desc = "Z";
				loc.soft = Datum.firm;
				loc.name = loc.toNewName(null);
			}
		}
	}											// End buildLvt

	// Workaround scope-def bugs in java compilers
	private void checkAdjustScope()
	{
		int numloc = meth.locals.size();

		// Scan the locals of this method for scope problems.
outer:
		for (int i=0; i<numloc; i++) {
			Local loc = (Local)meth.locals.elementAt(i);
			if (loc==null) continue;
			if (loc.startPc==0 || loc.startPc==65535) continue;

			// Chk for and fix "for-loop-var's scope ends too far down"
			// If local ends in middle of stmt, extend scope to stmt boundary
			// (Jikes has 2nd problem, btw)
			Statement stmt1 = checkScopeTooFar(loc);
			int locend = loc.startPc + loc.length;
			Statement past = meth.statementOf(locend, stmt1);
			if (past!=null && past.start!=locend)
			{
				loc.length += past.start + past.insts.length - locend;
				locend = loc.startPc + loc.length;
			}

			// Is local defined within the specified scope (as it should be)?
			for (Statement st=stmt1; st!=null && st.start<locend; st=st.next()) {

				// 1) If used in 1st stmt of scope, probably no work to do.
				// 2) If def later, move start later
				// Dont treat temp var in finally/sync as ref to real var
				Instruction refinst = st.references(loc);
				if (st.type!=Trek.ST_ENDFINALLY
					&& st.type!=Trek.ST_ENDSYNC
					&& refinst!=null)
				{
					if (st!=stmt1)
					{
						int start = st.start;

						// If 1st stmt loc is used in is def, then scope
						// starts after this stmt
						//Instruction fetchinst = st.fetches(loc);
						//if (fetchinst==null || fetchinst.start>refinst.start)
							start += st.insts.length;
						loc.length -= start - loc.startPc;
						if (loc.length<=0)
							break;			// Effectively not reffed
						loc.startPc = start;
					}

					// Delete def if same name more broadly defined.
					// Unfort biffer scopes arent neces earlier in lvt, so
					// too-far chk for bigger may not have been done above yet
					Local bigger = meth.getLocal(loc.name, loc.startPc-1);
					if (bigger!=null)
					{
						checkScopeTooFar(bigger);
						if (bigger.startPc+bigger.length>loc.startPc)
						{
							loc.remove();
							continue outer;
						}
					}

					// If earlier local with same idx, may be more work
					Local locp = meth.getPrevLocal(loc.idx, loc);
					if (locp==null)
						continue outer;

					// 3) If fetching stmt (ie. 1st fetch precedes 1st
					// mod if any), merge with prev. 
					Instruction fetchinst = st.fetches(loc);
					if (fetchinst!=null && fetchinst.start<=refinst.start)
					{
						locp.length = locend - locp.startPc;
						loc.remove();
					}

					// For 1 & 2 insure end of prev doesnt hit start of curr.
					else {
						if (locp!=null && locp.startPc+locp.length==loc.startPc)
							locp.length -= stmt1.insts.length;
						if (locp.length<0)
							locp.remove();
					}
					continue outer;
				}
			}							// End inner loop

			// 4) If not ref at all, discard.
			loc.remove();
		}								// End outer loop
	}

	// Create a dcl-only blk if compiler-supplied lvt has a used blk that 
	// does not adequately match a code-structure blk. (Iow, blk exists
	// purely because developer wanted to define a scope for local vars).
	// loc = the local being checked
	// stmt1 = stmt where local is dcl-ed (by 1st usage or ST_LOCAL)
	private void checkMakeDclBlock(Local loc, Statement stmt1)
	{
		// Get stmt past local's scope, insuring dont end in middle of
		// something -- like between if and then
		Statement pastscope = meth.statementOf(loc.startPc+loc.length-1, stmt1);
		if (pastscope==null)
			return;				// End pc past end of method
		pastscope = pastscope.next();
		pastscope = pastscope.getEobInsertPoint(stmt1.level);

		// If scope is to end of method, no dcl blk needed of course
		if (pastscope.level<1 || pastscope.next().level<1)
			return;

		// Incr & loop eob logically same stmt.
		if (pastscope.addr==Statement.ad_incr)
			pastscope = pastscope.next();
		int endpc = pastscope.start;

		// A local's usable scope obviously ends at local's eob. So if
		// endpc is at (or past) that eob, there is no work to do.
		// Similarly if specified scope is smaller than real scope,
		// no work to do (ie. stmt before pastscope is in inner blk of
		// local's scope -- this can occur when dcl for local used in incr
		// clause had to be pulled outside loop).
		Statement eob = stmt1.getEob();
		if (endpc>=eob.start || pastscope.prev().level>stmt1.level)
			return;
			
		// Get top and bot of loc's blk for segregation test.
		// For this, consider a Case a blk (& calc its top/bot explic)
		Statement top  = eob.mate;
		if (eob.type==Trek.ST_ENDSWITCH)
		{
			eob = stmt1.getCodeEob();
			for (top=stmt1; top.level>=stmt1.level; top=top.prev()) ;
		}

		// Exit loop early if any other locals have dcl-scopes within
		// loc's blk that are entirely before or after loc's scope. That
		// is, if there are no scopes loc needs to be segregrated from,
		// it is safe to view whole blk as its scope -- unless it
		// is pseudo-blk (ie. a Case in a switch).
		int size = meth.locals.size();
		int i;
		for (i=0; i<size; i++) {
			if (i==loc.arrayIdx) continue;
			Local cloc = (Local)meth.locals.elementAt(i);
			if (cloc==null) continue;
			int cendpc = cloc.startPc + cloc.length;
			if (cloc.startPc>=top.start && cendpc<=loc.startPc)
				break;			// In stmt1's blk before loc's scope
			if (cloc.startPc>=endpc && cendpc<=eob.start)
				break;			// In stmt1's blk after loc's scope
		}

		// If didnt exit loop early, stmt1's blk suffices -- unless it
		// is pseudo-blk (ie. a Case in a switch).
		if (i==size)
		{
			if (top.type!=Trek.ST_CASE)
				return;
			stmt1 = top.next();
			endpc = eob.start;
		}

		// If missing blk is right after Then that ends with "exiting stmt",
		// make Else-blk whose end matches end of this local's scope
		Statement pretop = stmt1.prev();
		Statement prepretop;
		if (pretop!=null && pretop.type==Trek.ST_ENDTHEN
			&& ((prepretop=pretop.prev()).type==Trek.ST_THROW
				|| prepretop.type==Trek.ST_RETURN
				|| prepretop.type==Trek.ST_BREAK
				|| prepretop.type==Trek.ST_CONTINUE))
		{
			pretop.string = "}";			// Suppress End-then's comment
			top = new Statement(meth, stmt1.start,
				Trek.ST_ELSE, stmt1.level, "else {");
			top.insert(stmt1);
			Finish.bumpBlock(top, endpc, Trek.ST_ENDELSE, "} // End if");
		}

		// Otherwise build just a freestanding blk
		else {
			top = new Statement(meth, stmt1.start,
				Trek.ST_LABEL, stmt1.level, "{");
			top.insert(stmt1);
			Finish.bumpBlock(top, endpc, Trek.ST_ENDLABEL, "}");
		}
	}

	// Returns true if made object firmer
	// to = datum that should be checked
	// from = contains data type info that is context of To usage
	// result = true if datum is used as target, else false
	private boolean checkMakeObjectFirmer(Datum to, Datum from, boolean result)
	{
		if (to.soft<=Datum.firm) return false;
		int fsoft = from.soft;
		String fdesc = from.desc;

		// If to is a result, fdesc must be subclass of it.
		// So if fdesc is a super class of to.desc, then 
		// fdesc is better. If no possible ancestor of fdesc,
		// then we have definite type for To. (Conversely if to.desc
		// were discovered to be final, its type is firm -- but because
		// of equals test above, never get here in that case)
		if (result)
		{
			if (fsoft<=Datum.firm && fdesc.equals("Ljava/lang/Object;"))
				to.soft = Datum.firm;		// x = Object says x is Object
			else if (from.opr==Term.op_plus
				&& fdesc.equals("Ljava/lang/String;"))
				to.soft = Datum.atleast;	// String concatenation

			// Have solid info?
			else if (fsoft==Datum.atleast || fsoft<=Datum.firm)
			{
				if (to.soft<=Datum.softobj)
					to.soft = Datum.atleast;
				else if (to.soft==Datum.atmost)
				{
					if (fdesc.equals(to.desc))
						to.soft = Datum.firm;	// 2-way constraint
					else return false;			// Learned nothing
				}
				else {
					if (isAncestorOf(fdesc, to.desc))
						;					// Narrower 1-way constraint
					else return false;		// Learned nothing
				}
			}

			// No, use it only if currently have nothing
			else if (to.soft==Datum.softy)
				to.soft = Datum.softobj;
			else return false;				// Fdesc "useless"
		}

		// If to is a ref, fdesc must be superclass of it. So if
		// curr to.desc is superclass of fdesc, then fdesc is
		// better. Similarly if fdesc is final, it must be best.
		//		((refcf=getClassFile(fdesc))!=null
		//		&& refcf.hasModifier(Trek.ACC_FINAL))
		//		to.soft = Datum.firm;
		else {
			if (fsoft==Datum.atmost || fsoft<=Datum.firm)
			{
				if (to.soft<=Datum.softobj)
					to.soft = Datum.atmost;
				else if (to.soft==Datum.atleast)
				{
					if (fdesc.equals(to.desc))
						to.soft = Datum.firm;	// 2-way constraint
					else return false;			// Learned nothing
				}
				else {
					if (isAncestorOf(to.desc, fdesc))
						;					// Narrower 1-way constraint
					else return false;		// Learned nothing
				}
			}

			// From info not solid, use it only if curr have nothing
			else if (to.soft==Datum.softy)
				to.soft = Datum.softobj;
			else return false;				// Fdesc "useless"
		}

		// Made improvement, save desc in case it as well as softness changed
		to.desc = fdesc;
		return true;
	}

	// Workaround javac bug that loop-var scopes can go too far down
	// loc = the local being checked
	// Also returns stmt that loc defined in
	private Statement checkScopeTooFar(Local loc)
	{
		// Is stmt where local is defined the ad_init of for-loop
		Statement stmt1 = meth.statementOf(loc.startPc-1, null);
		if (stmt1.addr!=Statement.ad_init)
			return stmt1;

		// Chk if end of scope past end of loop
		int endpc = loc.startPc + loc.length;
		Statement endsc = meth.statementOf(endpc-1, null);
		Statement loopeob = stmt1.next().mate;
		if (endsc.start<=loopeob.start)
			return stmt1;

		// Scan up to last ref of loc (but no earlier than end of its loop)
		while (endsc.start>=loopeob.start) {
			if (endsc==loopeob)
			{
				loc.length = endsc.start + endsc.insts.length - loc.startPc;
				break;
			}
			else if (endsc.references(loc)!=null)
				break;
			endsc = endsc.prev();
		}
		return stmt1;
	}

	// Clarify dcl of to if from provides adequate clarif. If To assoc
	// with local (usual case), propagate clarifs to the local as well.
	// to = datum that should be checked
	// from = contains data type info that is context of To usage
	// result = true if datum is used as target, else false
	// Notes:
	//	A soft data type can only be an object or an integer value
	//	(ie. int, byte, char, short, or boolean) because other
	//	data types are always firm from the outset.
	private void checkSoftLocal(Datum to, Datum from, boolean result)
	{
		// Contin if To is soft non-const and there is data to firm it with
		if (!to.isSoft() || from.soft==Datum.softy)
			return;

		// If from's de-scriptor is not for an integer val, loc is object.
		int pos = "BCSIZ".indexOf(from.desc);
		if (pos<0)
		{
			if (!checkMakeObjectFirmer(to, from, result))
				return;
		}

		// If fdesc is boolean, then loc must be as well. Also at overall
		// end of analysis, if code below cannot clarify length of value
		// to 8+, then loc's type will be set to boolean.
		else if (pos==4)
		{
			to.soft = Datum.firm;
			to.desc = "Z";
		}

		// If trying to firm a result, cant if desc is "small" 32-bit int
		// For ex, code for int = int looks the same as int = short.
		// Conversely, int = short should insure softy is at least shplus.
		else if (result)
		{
			setResultSoftness(to, from);
			if (to.soft==Datum.softy)
				return;
		}

		// Processing a ref (ie. from = to). Unless both may still be boolean,
		// now know "to" is not boolean.
		else {
			if (to.soft==Datum.softy && from.soft!=Datum.softy)
				to.soft = Datum.byplus;

			// Cant firm int-length of a ref if desc is "big" int.
			// For ex, char = char looks the same as char = byte.
			else if (pos>="BCSI".indexOf(to.desc))
				return;

			// Else since shorter = longer (with no cast) is impos, desc of
			// target defines max int-len that loc (ie. the source) can be.
			// Thus have narrower max now, so update desc/name below. Also
			// if min now equals max, also set loc's desc firm
			to.desc = from.desc;
			if (to.soft==softval[pos])
				to.soft = Datum.firm;
		}

		// Chk if just doing data type analysis
		Local loc = to.loc;
		if (loc==null)
			return;

		// No, real local. Update it's info. If object or still soft, done.
		loc.soft = to.soft;
		loc.desc = to.desc;
		loc.name = loc.toNewName(null);
		if (loc.soft>=Datum.softy || pos < 0)
			return;

		// Force re-analysis of stmt if some other soft int loc was laid
		// by wayside to learn about this loc. Also re-analysis will
		// decide whether a pass2 is needed, so either way clear delta
		if (inst.parent.delta>1)
			repeat = true;
		inst.parent.delta = 0;
	}

	// See checkSoftLocal
	protected void checkSoftRef(Datum dtm, Datum from)
	{
		checkSoftLocal(dtm, from, false);
	}
	protected void checkSoftResult(Datum dtm, Datum from)
	{
		checkSoftLocal(dtm, from, true);
	}

	// Mark current stmt for reanalysis if it uses a soft local
	private void checkSoftStatement(Datum dtm)
	{
		if (dtm.isSoft() && !pass2)
			inst.parent.delta += 1;
	}

	// Workaround compiler bug where int i; if () i=1; else i=2; creates
	// 2 locals. That is, if processing of else's i cause ST_LOCAL to
	// be generated before then's i, then merge the 2 defs.
	// Either way, return the object that (now) represents this local
	// loc = local that just got a ST_LOCAL created
	// start = where the ST_LOCAL stmt was inserted
	private Local checkUnsplitScope(Local loc, int start)
	{
		// Find prev local. If none, then done. (Needs to be loop
		// because each nested clause of If can be a separate local).
		while (true) {
			Local locp = meth.getPrevLocal(loc.idx, loc);
			if (locp==null) return loc;

			// If locp starts bef inserted dcl, or is for later code in
			// method, or it is diff name or data type, keep defs separate
			if (locp.startPc<=start || locp.startPc>loc.startPc
				|| !locp.name.equals(loc.name)
				|| !locp.desc.equals(loc.desc))
				return loc;

			// Extend 1st and remove 2nd. If 1st was already had dcl-only
			// def, remove it since extension will supplant this
			locp.length = loc.startPc + loc.length - locp.startPc;
			loc.remove();
			Statement def = locp.defStmt;
			if (def!=null && def.type==Trek.ST_LOCAL)
			{
				def.remove();
				def = null;
			}
			loc = locp;
		}
	}

	// (Build method's LVT if not supplied by compiler, then:)
	// Create ST_LOCAL stmts for locals that must be defined prior to
	// their 1st usage (eg. AA {lv=1;} return lv; -- lv must be def at AA)
	// Notes:
	// 1) loc.defStmt influences stmt.toString (ie. is data type in source?),
	//		so "real" toString/vm.execute cant be done til after dclLocals
	// 2) Java does not allow defs imbedded in a stmt. For
	//		example, vec.setSize(int a=1) is syntax error)
	void declareLocals(Method method)
	{
		// If no compiler-supplied local variables table, build it now.
		// If is, clean up scopes defined for each local
		meth = method;
		if (meth.lvRead<0)
			buildLvt();
		else checkAdjustScope();

		// Scan locals to cleanup defs and put out dcl-only stmts when
		// end of def persists past end of block that def is in.
		for (int i=0; i<meth.locals.size(); i++) {
			Local loc = (Local)meth.locals.elementAt(i);
			if (loc==null) continue;

			// Chk if a local has same name as a field of this class
			// If so, some refs will need to start with explicit "this."
			if (Trek.lookup(meth.parent.fields, loc.name)!=null)
				meth.fieldLocal = true;

			// Args dont need to be chked, since their scope is entire method 
			// in 1st place. (Also dont chk unused locals)
			loc.defStmt = null;
			if (loc.startPc==0 || loc.startPc==65535)
				continue;

			// StartPc is the end of the defining inst (eg. istore).
			// So for assignment stmts, it really points to start of
			// next stmt -- thus the -1. If doing local for a Catch,
			// no work to do.
			Statement stmt1 = meth.statementOf(loc.startPc-1, null);
			if (stmt1.type==Trek.ST_CATCH)
			{
				loc.setDefStatement(stmt1);
				continue;
			}

			// Setup state that indicates start of local's scope.
			// View for-init as logically in loop.
			int deflevel = stmt1.level;
			Statement stmt = stmt1;
			if (stmt1.addr==Statement.ad_init)
			{
				deflevel += 1;
				stmt = stmt1.next().next();
			}

			// Setup last real stmt of scope. Except for do {} while, the
			// code in a blk-end stmt is in the blk, so dont want scope that
			// runs to end of blk to falsely promote to blk-end's level.
			// (Note that javac includes goto of else as part of Then blk).
			int endpc = loc.startPc + loc.length;
			Statement endst = meth.statementOf(endpc-1, stmt1);
			if (endst.type==Trek.ST_ELSE
				|| (endst.type>=Trek.ST_END && endst.type!=Trek.ST_ENDDO))
				endpc = endst.start;

			// Find end of loc's definition scope. If there is usage
			// of loc in incr clause of For, insure deflevel incl level
			// of For-blk unless the local was defined before incr clause
			// but within the loop -- ie. in the loop's init clause.
			int origlevel = deflevel;
			for (; stmt!=null; stmt=stmt.next()) {
				if (stmt.addr==Statement.ad_incr
					&& (stmt1.addr!=Statement.ad_init || stmt1.level!=stmt.level-1)
					&& deflevel>=stmt.level
					&& stmt.references(loc)!=null)
					deflevel = stmt.level - 1;	// Expand to loop's level

				// If reached outer stmt past loc's scope of definition, done.
				// Else (if not a case stmt), broaden level of loc's def.
				if (endpc<=stmt.start)
					break;
				if (stmt.level<deflevel && stmt.type!=Trek.ST_CASE)
					deflevel = stmt.level;
			}

			// Put out codeless def for local if neces:
			// 1) Scope of def went past end of stmt1's block
			// 2) Store inst is before end of its stmt (ie. 1st
			//    assign is imbedded -- eg. if ((xxx=yyy) > 0)
			// (Note tho: on-fly dcl cant be inside stmt (eg. xxx(int a=1))
			//
			// Not 3) Def is only stmt in a blk. This is subsumed by bogus-blk
			// check above because that will lead to 1-stmt blks being unsplit
			//  || (def.prev().level<deflevel && def.next().level<deflevel)
			if (deflevel < origlevel
				|| (stmt1.start+stmt1.insts.length > loc.startPc
					&& stmt1.type!=Trek.ST_SYNC
					&& stmt1.type!=Trek.ST_ENDSYNC))
			{
				// Put def out as late as possible so that it is not before
				// an earlier def of same name in imbedded blk. If have:
				//		AA {int x=1;} BB {x=2.0;} x++;
				// putting "float x" at AA is syntax error. It must be put at
				// BB. However dont put it at impos pos like before "else {".
				Statement inspt = stmt1.getInsertPoint(deflevel);
				stmt1 = new Statement(meth, inspt.start,
					Trek.ST_LOCAL, inspt.level, loc.toString() + ";");
				stmt1.insert(inspt);
				if (meth.lvRead>=0)
					loc = checkUnsplitScope(loc, inspt.start);
				loc.setDefStatement(stmt1);
			}

			// Note that init var's defining stmt (stmt1.mate) is not
			// yet in method's linked list (see finishLoops)
			else if (stmt1.addr==Statement.ad_init)
				loc.setDefStatement(stmt1.mate);
			else loc.setDefStatement(stmt1);

			// If there is compiler-supplied lvt, use info in it to
			// create dcl-only blk if one is needed
			if (meth.lvRead>=0)
				checkMakeDclBlock(loc, stmt1);
		}								// End loop
	}									// End declareLocals

	// Use function prototype info to firm up arg's data type
	// dtm = arg to process
	// argdesc = its descriptor
	// args = accumulated args as source string
	protected void doArg(Datum dtm, String argdesc, StringBuffer args)
	{
		checkSoftRef(dtm, new Datum("", argdesc));
	}

	// Build overall condit expr from pieces on stack and val
	// val = rightmost term of overall condit expr
	// Returns val updated to the overall expr
	protected Datum doCondExpr(Datum val)
	{
		condEnd[ifLevel+1] = 0;		// Clear condit-expr mode at this level

		// Build condit expr from its pieces, incl sub condit expr's
		// A particular triplet could be the left or right term in an
		// outer expr. Rsave non-0 says last built a left-term
		int rsidx = -1;
		Datum rsave[] = new Datum[30];
		Datum left = val;
		Datum right = null;
		while (true) {

			// If leftmost of 3 terms is test, then build outer right term
			// Else building outer left, so save the "unused" right terms
			Datum test = pop();
			while (test.condExpr==null) {
				if (right!=null)
					rsave[++rsidx] = right;
				right = left;
				left = test;
				test = pop();
			}

			// If either is boolean, then both must be. And vice versa
			constrainOperand1(left, Trek.DOIF_EQ, right);
			constrainOperand1(right, Trek.DOIF_EQ, left);

			// Build datum for the condit expr. Without ability to keep
			// left/right visible when cond expr used, cant entirely do both
			// of these approaches because softer doesnt have to be broader
			// 1) Pass soft datum in hope restrictive result can firm it
			// 2) Pass broader datum in hope it can firm result the most
			if (left.loc!=null
				&& (left.soft==Datum.softy || left.soft>=Datum.byplus))
				val = new Datum(left.loc);
			else if (right.loc!=null
				&& (right.soft==Datum.softy || right.soft>=Datum.byplus))
				val = new Datum(right.loc);
			else {
				String lpromo = Datum.promote(left.desc);
				val = new Datum("", lpromo, Datum.softy);
				checkSoftResult(val, left);
				checkSoftResult(val, right);
			}
			val.src = "a?b:c";
			val.opr = Term.op_cond;

			// If sub-expr, use it as left or right of outer expr
			// depending on whether there is unused right term
			if (rsidx>=0)
			{
				right = rsave[rsidx--];
				left = val;
			}
			else {
				right = val;
				if (top<=condMark[ifLevel+1]) break;
				left = pop();
			}
		}										// End loop
		return val;
	}											// End doCondExpr

	// Returns real start of loc's scope.
	// (In special case, may remove init-clause from loop loc is in)
	private int get2ndStart(Local locp, Local loc)
	{
		// For catch stmt, start of dcl is by def the catch stmt itself.
		Statement top2 = meth.statementOf(loc.startPc-1, null);
		if (top2.type==Trek.ST_CATCH)
			return loc.startPc;

		// A local starts at startPc unless a dcl-only local is
		// required at a broader level. Main dcl-only case is:
		//		if (xxx) i=1; else i=2; x(i);
		// 2nd i's scope runs to end, so eob2.level+1 = 1 and
		// getInsertPt(1) rets if-stmt.
		Statement ip2 = top2.getInsertPoint(loc.defStmt.level + 1);
		if (top2==ip2)
			return loc.startPc;

		// If locp's scope extends into init clause of loc's loop
		// "degrade" the init-clause back to regular assign stmt
		if (ip2.addr==Statement.ad_init && ip2.references(locp)!=null)
		{
			ip2.addr = 0;
			ip2.mate = null;
			ip2 = ip2.next();
		}
		return ip2.start;
	}

	// Returns classfile of specified object type, or cfFail for
	// could not find it, or cfObject for java/lang/Object.
	// Caches results so failures not repeated over and over
	private ClassFile getClassFile(String desc)
	{
		// Go from descriptor to class name. clname can be null if
		// desc is array of primitives
		String clname = Datum.getDescClass(desc);
		if (clname==null)
			return cfFail;

		// Do fast path
		ClassFile result = (ClassFile)cfhash.get(clname);
		if (result!=null)
			return result;

		// Do the lookup, mapping exception to failure and avoiding
		// actual read of Object.class
		if (clname.equals("java/lang/Object"))
			result = cfObject;
		else {
			try {			
				result = cf.parent.getClassFile(clname);
				if (result==null)
					result = cfFail;
			}
			catch (Throwable th) {
				result = cfFail;
			}
		}

		// Permanize result
		cfhash.put(clname, result);
		return result;
	}

	// Return datum for this local var
	// idx = index of local to get
	protected Datum getLocalRef(int idx)
	{
		Datum dtm = super.getLocalRef(idx);
		checkSoftStatement(dtm);
		return dtm;
	}

	// Return local startpc implied by current stmt/inst
	private int getStartPc()
	{
		// Deal with special case that local defined at bottom of
		// loop code-wise, but top of loop source-wise. For ex,
		//   while ((a=xxx.get())!=null). More specifically, the local
		// is defined WITHIN loop-start stmt (as opposed to at its end,
		// as in an assign stmt)
		Statement stmt = inst.parent;
		int startpc = stmt.start + inst.start + inst.size;
		if (stmt.type==Trek.ST_ENDFOR || stmt.type==Trek.ST_ENDWHILE)
			startpc = stmt.mate.start + 1;
		return startpc;
	}

	// Insure local name unique
/*
	private void insureNameUnique(Local loc, Local locp)
	{
		while ((locp=meth.getPrevLocal(loc.idx, locp))!=null) {
			if (loc.name.equals(locp.name))
			{
				loc.name += "a";
				return;
			}
		}
	}
*/

	// Returns true if loc2 has soft data type that is compat with loc1's
	// Notes:
	//	soft>=byplus implies loc is internally an integer
	//	If soft>=(x)plus and ch2 is (x), then loc has already been set firm
	private boolean is2ndSubsumable(Local loc1, Local loc2)
	{
		// Chk if 2nd loc might be mergable (ie. desc not firm)
		if (loc2.soft<=Datum.firm)
			return false;

		// To here means ch2 is integer or object (incl array)
		int ch1 = loc1.desc.charAt(0);
		int ch2 = loc2.desc.charAt(0);

		// Can subsume object only if it fully soft and loc1 is object
		// For object types, fully soft is required for merge
		if (ch2=='[')
			return false;			// Array cant be fully soft
		if (ch2=='L')
			return loc2.soft==Datum.softy && (ch1=='L' || ch1=='[');

		// Is 2nd integer's desc compat with 1st?
		switch (ch1) {
		default:
			return false;

		// For types internally repr as 32-bit integers, 2 issues:
		// 1) min length of soft one is compat with source type of firm one.
		//    In partic, if loc2.soft is 1, it is still compat with boolean.
		// 2) Soft one's "max" data type is >= firm one's type.
		case 'Z':
			if (loc2.soft>Datum.softy)
				return false;		// No, cant merge
			else return true;
		case 'B':
			if (loc2.soft>Datum.byplus)
				return false;
			else return true;
		case 'C':
			if (loc2.soft>Datum.chplus)
				return false;
			else return true;
		case 'S':
			if (loc2.soft>Datum.shplus || ch2=='C')
				return false;
			else return true;
		case 'I':
			if (ch2=='C' || ch2=='S')
				return false;
			else return true;
		}
	}

	// Returns true if no way 2 locals types are compatible
	private boolean is2ndClearlyDifferent(Local locp, Local loc)
	{
		// If both firm, just chk desc
		if (locp.soft<=Datum.firm && loc.soft<=Datum.firm)
			return !locp.desc.equals(loc.desc);

		// Elim most cases where non-integers involved
		int ch1 = locp.desc.charAt(0);
		int ch2 = loc.desc.charAt(0);
		switch (ch1) {
		case 'L':
		case '[':
			if (locp.desc.equals(loc.desc))
				return false;		// Same desc
			if (!(ch2=='L' || ch2=='['))
				return true;		// Obj and non obj
			if ((ch2=='[' && !locp.desc.equals("Ljava/lang/Object;"))
				|| (ch1=='[' && !loc.desc.equals("Ljava/lang/Object;")))
				return true;		// Array and not java.lang.Object
			else return false;		// Else may be same
			/*
			if ((locp.soft>=Datum.softy && locp.soft!=Datum.atmost)
				|| (loc.soft>=Datum.softy && loc.soft!=Datum.atmost))
				return false;		// Too little info to be clearly diff
			if (isAncestorOf(locp.desc, loc.desc)
					|| isAncestorOf(loc.desc, locp.desc))
				return false;		// Super-sub, so could be same
			else return true;
			*/

		// Since these always firm, failure of 1st chk implies other diff
		case 'D':
		case 'F':
		case 'J':
			return true;

		case 'Z':
			return !(ch2=='I' && loc.soft==Datum.softy);
		}

		// Elim rest of cases with non-integers involved
		switch (ch2) {
		case 'L':
		case '[':
		case 'D':
		case 'F':
		case 'J':
			return true;

		case 'Z':
			return !(ch1=='I' && locp.soft==Datum.softy);
		}

		// Both integers and either not firm, so not clearly diff
		return false;
	}

	// Returns true if ances is ancestor of descen. Note that false
	// is returned if cant determine their relationship
	// ances/descen = descriptors
	private boolean isAncestorOf(String ances, String descen)
	{
		// Do chks that can be done by just looking at descen or ances.
		// Note that Object is ancestor even of arrays. Except for
		// that, arrays have no ancestors
		if (descen.equals("Ljava/lang/Object;"))
			return false;
		if (ances.equals("Ljava/lang/Object;"))
			return true;
		if (descen.startsWith("[") || ances.startsWith("["))
			return false;

		// Now look at arg2's classfile. If cant, ret false thru lack of info
		ClassFile cfdes = getClassFile(descen);
		if (cfdes==cfFail)
			return false;

		// Chk immed ancestor by looking at superClass property
		CpInfo cpi = cfdes.constant(cfdes.superClass);
		if (cpi==null)
			return false;
		if (cpi.toInternalString().equals(ances))
			return true;

		// If ances accessible, do real chk
		ClassFile cfanc = getClassFile(ances);
		if (cfanc==cfFail)
			return false;
		else {
			try {
				return cfanc.isAncestor(cfdes);
			}
			catch (Throwable th) {
				return false;			// Again lack of info
			}
		}
	}

	// Returns true if definitely have 2 diff source data types
	// loc = existing loc that may need to end scope now
	// src = datum whose type will be given to 2nd local if it is needed
	private boolean isInvisNewBlock(Local loc, Datum src)
	{
		// Not invis new blk if pass2-analyze, or loc is arg, or
		// this loc already defined (because analysis being repeat-ed),
		// or loc has same data type as source datum. But if loc is soft,
		// treat them as not same because they may end up different.
		if (pass2 || loc.startPc==0 || loc.startPc==getStartPc()
			|| inst.parent.addr==Statement.ad_incr
			|| (loc.desc.equals(src.desc) && loc.soft<=Datum.firm))
			return false;

		// But if they're guaranteed diff even in presence of softness,
		// ret true. Deals with ugly case that optimizer may overload local
		// (in single stmt!) when it in-lines a method. Eg, a variable used
		// in computing method's obj ref (possibly just itself) and ret val
		// may be same local index.
		if (!Datum.promote(loc.desc).equals(Datum.promote(src.desc)))
			return true;

		// Otherwise treat reading and writing loc
		// in same stmt as meaning they are same variable. Do reads+writes
		// slow chk (ie. look at each inst) only if fast chk inconclusive.
		// (R+W chk needed because desc's can appear diff now if either is U)
		if (loc.soft>=Datum.softy
			&& (loc==src.loc || inst.parent.fetches(loc)!=null))
			return false;
		else return true;
	}

	// Dont need to process detailed source strings for lvt analysis
	// opd = source of operand (var or expr)
	// subopr = the top-level operator within the operand
	// opr = the opr being applied to the operand
	protected String isolate(String opd, int subopr, int opr)
	{
		if (Term.precs[subopr] < Term.precs[opr])
			return "()";
		else return opd;
	}

	// Merges dcl of two locals (with diff types) whose scope overlaps
	private void mergeLocalProps(Local locp, Local loc)
	{
		// If already same, no work to do.
		if (locp.desc.equals(loc.desc))
			return;

		// Use 1st local if it is firm or 2nd completely soft.
		// Since 2nd local has the broader scope, use its data type
		// otherwise -- except when there are conflicting objects.
		if (loc.soft==Datum.softy || locp.soft<=Datum.firm)
			return;			// Use 1st
		if (locp.soft==Datum.softy || loc.soft<=Datum.firm
			|| locp.desc.length()<=1)
		{					// Use 2nd
			locp.copyDclInfo(loc, null);
			return;
		}

		// Objects with tentative data types can have conflicts.
		// A conflict can happen when the usage situation is
		// something like "if () loc = cl1; else loc = cl2;"
		// Use a common superclass when this occurs. If one or the
		// other desc-guesses is super of the other, 3 cases exist:
		// 1) if descendant has atmost-desc, then ancestor
		// simply has a too-loose desc, so use descendant's info.
		// 2) if descendant is 2nd (ie. loc) and has atleast-desc,
		// then later usage never required the broader class, so if locp
		// usage also didnt require broader class, use descandant's info.
		// 3) Else use the ancestor's info.
		if (isAncestorOf(locp.desc, loc.desc))
		{
			if (loc.soft==Datum.atmost
				|| (loc.soft==Datum.atleast && locp.soft!=Datum.atleast))
				locp.copyDclInfo(loc, null);
			else ;						// Use dcl info of locp
		}
		else if (isAncestorOf(loc.desc, locp.desc))
		{
			if (locp.soft==Datum.atmost)
				;
			else locp.copyDclInfo(loc, null);
		}

		// Otherwise hope next ref-usage (eg. return loc;) 
		// identifies the unifying superclass of both. Scan
		// for this unifying ref stmt by resetting loc to soft and
		// calling reanalyze to process stmts past loc's defining stmt.
		else {
			loc.soft = Datum.softy;
			reanalyze(loc);
			locp.copyDclInfo(loc, null);
		}

		// It may be by fiat, but assume we now have firm class of locp
		locp.soft = Datum.firm;
	}

	// Reanalyze stmts from def up to loc's end of scope.
	private void reanalyze(Local loc)
	{
		// Init scan
		Statement def = meth.statementOf(loc.startPc-1, null);
		Statement first = def.next();

		// Reanalyze data type of this local until it is hard or its
		// end of scope is reached
		for (Statement stmt=first; stmt!=null; stmt=stmt.next()) {
			analyze(stmt);
			if (loc.soft<=Datum.firm || stmt.start>=loc.defStmt.start)
				return;
		}
	}

	// Reanalyze stmts that might have been affected by change to loc
	// loc = local whose def was changed
	// start = where change starts
	// end = where change ends
	private void reanalyze(Local loc, int start, int end)
	{
		// Process only stmts that have soft locals and in which loc is used
		Statement def = meth.statementOf(start-1, null);
		for (Statement stmt=def; stmt!=null; stmt=stmt.next()) {
			if (stmt.delta>0 && stmt.references(loc)!=null)
				analyze(stmt);
			if (stmt.start>=end)
				return;
		}
	}

	// Update softness of local that is being used as result(during buildLvt)
	// to reflect addit info gleaned from source.
	// result = thing being assigned to
	// source = what result is being set from
	private void setResultSoftness(Datum result, Datum source)
	{
		// If already firm, do nothing. Else use source softness to
		// set result softness. But note that a firm short
		// integer doesnt make firm result because i = s gens same
		// code as i = i. Similary obj1 = obj2, obj1 can be superclass
		if (result.soft<=Datum.firm)
			return;
		int soft = source.soft;

		// Learn what can for integers and objects (no B case since
		// B implies firm).
		switch (result.desc.charAt(0)) {
		case 'I':
		case 'C':
		case 'S':
			if (source.soft==Datum.firm)
			{
				int ch = source.desc.charAt(0);
				if (ch=='B')
					soft = Datum.byplus;
				else if (ch=='C')
					soft = Datum.chplus;
				else if (ch=='S')
					soft = Datum.shplus;
				else soft = Datum.firm;
			}
			break;

		// Result objects are same or superclass of source
		case 'L':
		case '[':
			if (source.soft==Datum.firm)
				soft = Datum.atleast;
			else if (source.soft==Datum.atmost)
				soft = Datum.softobj;
		}

		// Update result's softness if calc-ed value further constrains it
		result.refineSoftness(soft);
	}

}

