//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

package dec.trek;
import java.io.*;
import java.util.Vector;

//
//
// Member. Superior for fields and methods
//
//
public abstract class Member implements Sortable
{
	// Info needed to make its context self-contained
	ClassFile parent;				// Identifies field's class
	int idx;

	// Contents of member
	protected int accessFlags;		// Access prop of field (eg. public)
	protected int nameIdx;			// Index of its name in constant pool
	protected int descIdx;			// Index of its data type info
	Attr attrs[];					// Its non-code attributes
	public Object userDefined;		// User-owned field

	// Insure item index matches where stored in vector. Binding is delayed
	// until read done because indexes in flux while sorting going on
	static void bindIndexes(Vector mems)
	{
		int vecsize = mems.size();
		for (int i=0; i<vecsize; i++) {
			Member mem = (Member)mems.elementAt(i);
			mem.idx = i;
		}
	}

	// Returns std 1,0,-1 from comparing This to arg
	// this = the object that is in the vector
	// obj = the object/value outside the vector
	public int compareTo(Object obj)
	{
		String key, objkey;

		// Comparing to another classfile obj or just a string?
		if (obj instanceof String)
			objkey = (String)obj;
		else {
			Member mem = (Member)obj;
			objkey = mem.toName();
		}

		// Compare it to key from This and return result
		key = toName();
		return key.compareTo(objkey);
	}

	// Return ptr to its classfile
	public ClassFile getClassFile()
	{
		return parent;
	}

	// Returns true if obj's accessFlags has any of the modifiers in flags arg
	public boolean hasModifier(int flags)
	{
		return flags==-1 || (accessFlags&flags)!=0;
	}

	// Returns index of this member
	public int indexOf()
	{
		return idx;
	}

	protected Member() {}

	// Read the member from its classfile
	// chan = where to read from
	// Notes:
	//	For methods, stmts & excepts are filled in by readAttrs (via readCode)
	void read(DataInput chan) throws IOException
	{
		// Read its declaration and its attr.
		accessFlags = chan.readUnsignedShort();
		nameIdx = chan.readUnsignedShort();
		descIdx = chan.readUnsignedShort();
		attrs = Attr.readAttrs(chan, parent, this);
	}

	// Returns internal-fmt data type of member (overridden for method
	// to return desc of method's return value)
	String toDesc()
	{
		return toSignature();
	}

	// Returns presentation string for name of member dcl or member ref
	// classname = member ref's unqual class name (non-null if "this"
	//		was constructed from a MemberRef -- see notes)
	// Notes:
	//	Maps special case of <init> to class name (a "constructor method").
	//	When processing member ref, this.parent is irrelevant because it
	//	is the referencing class, not the class of the member being reffed
	public String toName() {return toRef(null);}

	protected String toRef(String classname)
	{
		String name = parent.constant(nameIdx).toInternalString();

		// Constructor special case?
		if (name.startsWith("<init>"))
		{
			if (classname==null)
				return parent.toName();
			else return classname;			// Constructor def
		}

		// Regular member.
		else if (classname==null)
			return name;					// Member dcl
		else return classname + "." + name;	// Member ref
	}

	// Returns full desc string of member
	String toSignature()
	{
		return parent.constant(descIdx).toInternalString();
	}

	// Let user do mem.xxx() where reasonable for method/field methods
	public abstract String toObjectId();
	public abstract String toString();
	public abstract String toType();

	// Write the member to its classfile
	void write(DataOutput chan) throws IOException
	{
		// Write member's declaration & attr (in partic code if method)
		chan.writeShort(accessFlags);
		chan.writeShort(nameIdx);
		chan.writeShort(descIdx);
		Attr.writeAttrs(chan, attrs, parent, this);
	}

}

