//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

package dec.trek;
import java.io.*;
import java.util.*;

//
//
// Method. The content of a method.
//
//
public class Method extends Member
{
	int argCount;						// # of args method has
	Attr cattrs[];						// Other code attrs
	int addLocals;						// Extra locals needed for added calls
	int addStack;						// Min stack size needed for added code
	int callLocalsBase = -1;			// Where extra locals for added calls are
	ExceptionHandler excepts[];			// Method's exception table
	boolean fieldLocal;					// A field of this class has same name 
										// as a local of this method
	Statement first, last;				// 1st and last stmts of method
	int lvRead = -2;					// # of local vars read from classfile,
										// or < 0 for not in classfile:
										// -2 pre-buildLvt and -1 after that
	Linum linums[];						// Line-number table
	Vector locals;						// Local variable table
	private int maxStack;				// Compiler value for max opd stack size
	int maxLocals;						// Compiler value # of locals in method
	int thrown[];						// Thrown-exceptions table

	// Creates entries for args in local-vars table if lvt not read
	// Notes:
	//	This is "touchy" because classfiles have lvt-idx gaps in args
	//	for long/double entries. Gaps are important because code refs to
	//	lv indexes must match lvt positions
	private void addArgLocals()
	{
		// Setup the storage. Codeless method may not have maxLocals.
		// If so, build it.
		if (first==null)					// Code-less method?
		{
			maxLocals = argCount;
			if (!hasModifier(Trek.ACC_STATIC))
				maxLocals += 1;
		}
		addLvt();

		// Define "this" if non-static method. If static, map arg1 to local0.
		// If non-static, skip "this" and map arg1 to local1.
		int adj;
		if (hasModifier(Trek.ACC_STATIC))
			adj = 1;
		else {
			adj = 0;
			String classdesc = "L" + parent.toInternalName() + ";";
			new Local(0, 65000, "this", classdesc , 0, this);
		}

		// Get descriptors of arg list
		String desc = toArgsDesc();
		if (desc==null) return;

		// For each arg, add an entry into lvt. Create gaps as needed too.
		int pos = 0;
		for (int i=1; ; i++) {
			int len = Datum.getDescLength(desc, pos);
			if (len<=0) return;

			// Add the entry
			String argdesc = desc.substring(pos, pos+len);
			String name = "a" + i;
			new Local(0, 65000, name, argdesc, i-adj, this);

			// Chk for long or double
			if (first!=null && Datum.getDataSize(argdesc)==2)
				i += 1;					// Create a gap
			pos += len;
		}
	}										// addArgLocals

	// Build empty local var table
	void addLvt()
	{
		locals = new Vector(maxLocals, 10);
		locals.setSize(maxLocals);
	}

	// Support for classfile dumping
	void dump(PrintWriter chan, int flags)
	{
		// Dump the method unless it is internal construct and only source
		// code is being dumped. <clinit> and class$ are internal constructs
		if (flags==Trek.DUMP_STMT
			&& (parent.clinit==this || toName().equals("class$")))
			return;
		boolean hasbody = !hasModifier(Trek.ACC_ABSTRACT|Trek.ACC_NATIVE);

		// Maximize likelihood get something useful out when crash occurs
		try {
			String hdr = toString() + toThrowsString();
			if (hasbody)
				Trek.printPretty(chan, "    ", hdr);
			else Trek.printPretty(chan, "    ", hdr + ";");
		}
		catch (Exception e) {
			logMethodError("dumping", e);
		}

		// Attributes
		try {
			dumpAttr(chan, flags);
		}
		catch (Exception e1) {
			logMethodError("dumping", e1);
		}

		// Code
		try {
			if (hasbody)
				dumpCode(chan, flags);
		}
		catch (Exception e) {
			logMethodError("dumping", e);
		}

		// Display method separator
		Trek.println(chan);
	}										// End of dump

	// Dump method attr
	void dumpAttr(PrintWriter chan, int flags)
	{
		// Do attributes if requested
		if ((flags & Trek.DUMP_ATTR)!=0)
		{
			Trek.println(chan, "    Max stack: " + maxStack);
			Trek.println(chan, "    Max locals: " + maxLocals);

			// Print summary of its line num table, if any
			if (linums!=null)
			{
				String entries = linums.length==1 ? " entry" : " entries ";
				Trek.println(chan, "    Line Number Table: "
					+ linums.length + entries);
			}

			if (first == null)
			{
				Trek.println(chan);
				return;					// No code for this method
			}

			// Print its extended attrs (Excluding instructions of course)
			for (int j=0; j<attrs.length; j++) {
				Attr attr = attrs[j];
				if (attr.value != null)
					attr.dump(chan, parent, "    ");
			}

			// Print its Code's attrs (Excluding locals)
			for (int j=0; j<cattrs.length; j++) {
				Attr attr = cattrs[j];
				if (attr.value != null)
					attr.dump(chan, parent, "    ");
			}

			// Print its exception table
			if (excepts.length > 0)
			{
				Trek.println(chan, "    Exceptions:");
				for (int i=0; i<excepts.length; i++) {
					Trek.println(chan, "    " + "  " + excepts[i].toString());
				}
			}

			// Print its local-vars table. If dumping built stuff, read
			// a stmt to insure that method has been exploded
			if (lvRead > 0 || (Trek.debug&Trek.Db_currloc) != 0)
			{
				if ((Trek.debug&Trek.Db_currloc) != 0)
					firstStatement();
				Trek.println(chan, "    Local variables:");
				for (int i=0; i<locals.size(); i++) {
					Local loc = (Local)locals.elementAt(i);
					if (loc != null)
						Trek.println(chan, "    " + loc.toString(parent));
				}
			}
		}								// End dump_attr
	}

	// Dump method code
	void dumpCode(PrintWriter chan, int flags)
	{
		// Print its source and/or instructions.
		// If both, sep stuff for each stmt with blank line
		boolean do_i = (flags & Trek.DUMP_INST) != 0;
		boolean do_s = (flags & Trek.DUMP_STMT) != 0;
		if (do_i || do_s)
		{
			Vmi vm = new Vmi();
			Statement stp = null;
			Statement stn;
			Trek.println(chan, "    {");

			// Scan the code.
			for (Statement stmt=firstStatement(); stmt!=null; stmt=stn) {
				stn = stmt.next();
				if (do_i) stmt.dump(chan);
				if (do_s)
				{
					// Output stmt unless it is the superf return at end
					// of void method (and doing just stmts)
					String src = stmt.toString(vm);
					if (do_i || stn!=null || !src.endsWith("return;"))
					{
						// Add blank line at end of a series of end-blks or
						// at the start of a "run-in" block
						if (stp!=null && !do_i &&
							stp.type>=Trek.ST_END && stmt.type<Trek.ST_END)
							Trek.println(chan);

						// Format and output it
						String ind = stmt.toIndent(8, 4, 80);
						if ((flags & Trek.DUMP_ATTR)!=0 & !do_i)
						{
							String pfx = "" + (10000 + stmt.start);
							ind = pfx.substring(1) + ind.substring(4);
						}
						Trek.printPretty(chan, ind, src);
					}
				}

				// If both stmts & insts, put blank line betw non-empty stmts
				if (do_s && do_i && stn!=null && stn.insts.length>0)
					Trek.println(chan);
				stp = stmt;
			}
			Trek.println(chan, "    } // End of " + toName());
		}
	}

	// Explode the primordial 1st statement into method's "actual" statements.
	// Note that source code may have fewer statements than are created here.
	// For ex: "if ((a=b)>0)" creates Statement1: a=b, and St2: if(a>0).
	// Notes:
	//	A trek has an instruction cache for explodes. Only one explode can
	//	use the cache at a time. Thus the sync is needed
	private void explode()
	{
		// Do outside to quickly handle normal case that explode already done
		// Note: a thread may not immediately see changes made by another
		// thread to non-volatile variables. But that is okay here because it
		// will only lead to a "false negative" -- which will be corrected
		// inside the synchronized block
		Statement prim = first;
		if (prim!=null && prim.level!=0)
			return;								// Initialized already

		// Test inside too to account for race between 2 calls to explode
		synchronized (parent.parent.icache) {
			prim = first;
			if (prim==null) return;				// Codeless method

			// Remove primordial stmt and replace it with actual statements, then
			// finish organizing source code (eg. make Else blks, create locals).
			// If divide crashes, it re-inserts prim. Dont Finish in that case.
			if (prim.type==Trek.ST_PRIMORD && prim.next==null)
			{
				prim.remove();
				new Divide(this, prim);
				if (first!=prim && (Trek.debug&Trek.Db_divonly) == 0)
					new Finish(this, prim);
			}
		}

		// Should policy be to throw error if explode failed???
		// (Which is indicated by 1st statement having level of 0)
	}									// End of explode

	// Returns 1st local of method
	public Local firstLocal()
	{
		if (locals.size()==0) return null;
		explode();
		return (Local)locals.firstElement();
	}

	// Ret 1st orig stmt, exploding primordial statement if not already done
	public Statement firstStatement()
	{
		if (first==null) return null;

		// Find 1st real stmt (after insuring method had been split int stmts)
		explode();
		for (Statement st=first; ; st=st.next) {
			if (st.level >= 0) return st;	// Skip inserted stmts
		}
	}

	// Return the number of args this method has
	public int getArgCount()
	{
		return argCount;
	}
	static int getArgCount(String desc, ClassFile cf)
	{
		// Get at classfile's internal-fmt ret-type/args info & chk
		// for empty arglist (not sure if empty arglist ever encoded
		// as void, but cant hurt to chk this)
		desc = toArgsDesc(desc);
		if (desc==null) return 0;

		// Count each argument in desc.
		int pos = 0;
		for (int num=0; ; num++) {
			int len = Datum.getDescLength(desc, pos);
			if (len<=0) return num;
			pos += len;
		}
	}

	// Return exception handlers of this method
	public ExceptionHandler[] getExceptionHandlers()
	{
		ExceptionHandler exarr[] = new ExceptionHandler[excepts.length];
		System.arraycopy(excepts, 0, exarr, 0, excepts.length);
		return exarr;
	}

	// Return local def corresponding to specified lvt index & access pc
	// idx = local-id in this method's local var table
	// inst = referencing instruction
	// pc = referencing pc
	// desc = data type local should have
	// Notes:
	//	scan needs to be low towards high to deal with compiler bugs in
	//	which scope of later var improperly overlaps scope of earlier var
	Local getLocal(int idx, Instruction inst)
	{
		return getLocal(idx, inst.parent.start+inst.start+inst.size, null);
	}
	Local getLocal(int idx, Instruction inst, String desc)
	{
		return getLocal(idx, inst.parent.start+inst.start+inst.size, desc);
	}
	Local getLocal(int idx, int pc)
	{
		return getLocal(idx, pc, null);
	}

	private Local getLocal(int idx, int pc, String desc)
	{
		// Temp locals dont have lvt entries.
		// Temp locals are locals not in source code. (Of course,
		// if building lvt, all locals are equally real). Compilers
		// create them in sync stmt, for ex. The Call class creates them
		// when insert pt context is used to access info
		Local loc = localAt(idx);
		if (loc==null) return null;

		// Chk if this matches main entry for the local. That is,
		// is this instruction in the pc-range of this def?
		if (loc.isActive(pc, desc))
			return loc;

		// 1st def of local not the right one. Scan overflow
		int max = locals.size();
		for (int i=maxLocals; i<max; i++) {
			loc = (Local)locals.elementAt(i);
			if (loc==null) continue;
			if (loc.idx==idx && loc.isActive(pc, desc))
				return loc;
		}
		return null;			// Shouldnt be reached
	}

	// Returns local corresponding to specified name & access pc
	// name = name of local to lookup
	// pc = referencing pc
	Local getLocal(String name, Instruction inst)
	{
		return getLocal(name, inst.parent.start+inst.start+inst.size);
	}
	Local getLocal(String name, int pc)
	{
		Local loc = localAt(0);		// Insure lvt setup
		if (loc==null) return null;

		// Lookup this name. Does it exist at specified pc?
		for (int i=0; i<locals.size(); i++) {
			loc = (Local)locals.elementAt(i);
			if (loc==null) continue;
			if (loc.isActive(pc) && name.equals(loc.name))
				return loc;
		}
		return null;				// Name doesnt exist at specified pc
	}

	
	// Returns nearest local with index of idx to specified array pos
	Local getPrevLocal(int locidx, Local last)
	{
		// Start or resume backwards scan
		int arridx;
		if (last == null)
			arridx = locals.size();
		else arridx = last.arrayIdx;

		// Look for earlier overflow def of this local
		for (int i=arridx-1; i>=maxLocals; i--) {
			Local loc = (Local)locals.elementAt(i);
			if (loc==null || loc.length<0) continue;
			if (loc.idx==locidx)
				return loc;					// Found one
		}

		// Return main entry if Last was not main entry already
		if (arridx <= locidx)
			return null;
		else return (Local)locals.elementAt(locidx);
	}

	// Returns true if method is a constructor
	public boolean isConstructor()
	{
		String name = parent.constant(nameIdx).toInternalString();
		return name.startsWith("<init>");
	}

	// Returns last local of method
	public Local lastLocal()
	{
		explode();
		if (locals.size()==0) return null;

		// Find last non-null entry
		for (int i=locals.size()-1; i>=0; i--) {
			Local loc = (Local)locals.elementAt(i);
			if (loc!=null) return loc;
		}
		return null;
	}

	// Ret last stmt, exploding primordial statement if not already done
	public Statement lastStatement()
	{
		if (first==null) return null;

		// Find 1st real stmt (after insuring method had been split int stmts)
		explode();
		for (Statement st=last; ; st=st.prev) {
			if (st.level >= 0) return st;
		}
	}

	// Returns local var corresponding to specified index
	// Return null if no entry in this slot (occurs for temp locals).
	public Local localAt(int idx)
	{
		if (locals==null || idx>=locals.size())
			return null;

		// Insure ask after lvt built
		Local loc = (Local)locals.elementAt(idx);
		if (loc!=null) return loc;
		explode();
		return (Local)locals.elementAt(idx);
	}

	// Logs error for problem during method processing
	void logMethodError(String doing, Exception e)
	{
		String msg = "C=" + parent.toQualifiedName()
			+ ";D=" + doing
			+ ";M="	+  toName() + ";";
		parent.logError(msg, e);
	}

	// Construct method object for method ref
	// (Used to convert method-ref cp entry to string repr of method)
	Method(ClassFile cf, int nameidx, int descidx)
	{
		parent = cf;
		nameIdx = nameidx;
		descIdx = descidx;
		idx = -2;			// Disting from real method
	}

	// Construct method object for method
	private Method(ClassFile cf)
	{
		parent = cf;
		idx = -1;
	}

	// Navigate to next method of classfile. If at end, ret null instead
	public Method next()
	{
		if (idx >= parent.methods.size()-1)
			return null;
		Method method = (Method)parent.methods.elementAt(idx + 1);
		return method;
	}

	// Navigate to prev method of classfile. If at start, ret null instead
	public Method prev()
	{
		if (idx <= 0)
			return null;
		Method method = (Method)parent.methods.elementAt(idx - 1);
		return method;
	}

	// Read the code attribute and create method's statements
	// chan = where to read from
	// len = size of code-and-stuff in the Code attribute
	void readCode(DataInput chan, int len) throws IOException
	{
		// Read hdr fields
		maxStack = chan.readUnsignedShort();
		maxLocals = chan.readUnsignedShort();

		// Read code and store in primordial Statement object
		int codelen = chan.readInt();
		Statement stmt = new Statement(this, codelen);
		if (stmt.insts.length <= 0)
			stmt.type = Trek.ST_EMPTY;			// No code, not primordial
		else chan.readFully(stmt.insts);

		// Read exceptions and attributes (such as local-vars table).
		ExceptionHandler.readExcepts(chan, this);
		cattrs = Attr.readAttrs(chan, parent, this);
	}

    // Read all the methods from the classfile
	static void readMethods(DataInput chan, ClassFile cf) throws IOException
	{
		// Get # of methods and alloc/read-in all of them
		int len = chan.readShort();
		cf.methods = new Vector(len);	// Get cnt & alloc methods array
		for (int i=0; i<len; i++) {
			Method meth = new Method(cf);

			// Read the method, get arg cnt, and add lvt if it wasnt
			// read (and put args into it)
			meth.read(chan);
			if ((Trek.debug&Trek.Db_at) != 0)
				Trek.println(System.out, "Reading method " + meth.toName());
			meth.argCount = getArgCount(meth.toSignature(), cf);
			if (meth.locals==null)
				meth.addArgLocals();
			if (meth.toName().equals("<clinit>"))
				cf.clinit = meth;

			// Normally methods are stored alphabetically. Allow
			// to keep orig order when debugging though
			if ((Trek.debug&Trek.Db_write)==0)
				Trek.insertSorted(cf.methods, meth);
			else cf.methods.addElement(meth);
		}
	}

	// Read the specified method's thrown-exceptions table
	// chan = where to write to
	void readThrown(DataInput chan)	throws IOException
	{
		int len = chan.readUnsignedShort();		// # of exceptions
		thrown = new int[len];

		// Read each entry
		for (int i=0; i<len; i++) {
			thrown[i] = chan.readUnsignedShort();
		}									// End loop
		return;
	}

	// Take an inner class constructor arglist and pull outer-this from it
	// Returns [0] = pulled out outer-this
	//		   [1] = shrunken arglist
	// Notes:
	//	Currently not used. Instead all inner classes are forced to static.
	//	This allows all outer vars, incl outer-this, to be handled uniformly
	static String[] refize(String arglist)
	{
		int past1, start2;

		// Isolate outer class's this
		past1 = arglist.indexOf(',');
		if (past1>0)
			start2 = past1 + 2;
		else past1 = start2 = arglist.indexOf(')');

		// Build retvals using this info
		String split[] = { arglist.substring(1, past1),
			"(" + arglist.substring(start2) };
		return split;
	}

	// Return statement this pc is in. This will not match a 0-length stmt.
	// pc = method-relative pc want stmt of
	// stmt = non-null says know that statement is at or after this
	Statement statementOf(int pc, Statement stmt)
	{
		// Scan from access pt to specified pc
		if (stmt==null) stmt = firstStatement();
		for (; stmt!=null; stmt=stmt.next()) {
			if (pc < stmt.start + stmt.insts.length)
				return stmt;
		}
		return null;			// PC past end of method
	}

	// Returns source of method's arglist
	public String toArgList()
	{
		StringBuffer args = new StringBuffer(64);

		// If "this" is method ref, just output data types
		if (idx==-2)
		{
			String desc = toArgsDesc();
			int pos = 0;
			while (true) {
				int len = Datum.getDescLength(desc, pos);
				if (len<=0) break;
				String argdesc = desc.substring(pos, pos+len);
				if (pos > 0) args.append(", ");
				pos += len;
				args.append(parent.toDef(null, argdesc, 0));
			}
		}

		// Non-static args start with 1 because "this" has index 0.
		else {
			int firstarg = hasModifier(Trek.ACC_STATIC) ? 0 : 1;
			int seen = 0;

			// Convert each arg desc to "dt name" & return whole arg list
			int argcnt = getArgCount();
			for (int i=firstarg; seen<argcnt; i++) {
				Local loc = getLocal(i, 0);
				if (loc==null) continue;	// Empty slot because of long/double
				if (seen > 0) args.append(", ");
				args.append(loc.toRef(parent, 0));
				seen += 1;
			}
		}
        return "(" + args.toString() + ")";
	}

	// Returns descriptor of method's arglist, or null if no args
	// desc = meth descriptor (ie. for a meth ref),
	//			or null to use this method's descr
	String toArgsDesc()
	{
		return Method.toArgsDesc(toSignature());
	}
	static String toArgsDesc(String desc)
	{
		desc = desc.substring(desc.indexOf('(') + 1, desc.indexOf(')'));
		if (desc.length()==0 || desc.charAt(0)=='V')
			return null;
		else return desc;
	}

	// Returns descriptor of method's return value (or "V")
	String toDesc()
	{
		String desc = toSignature();
		return desc.substring(desc.indexOf(')') + 1);
	}

	// Returns local's name
	// this = a method or a method object for a method ref
	// inst = inst in method where ref occurs
	// idx = lvt index of desired local var
	// Notes:
	//	Class-file definitional quirk: ref pts are defined with respect to
	//	end of deffing/reffing instruction -- rather than its beginning.
	//	A method reference does not have a locals table obviously
	String toLocal(Instruction inst, int idx)
	{
		Local loc = getLocal(idx, inst.parent.start+inst.start+inst.size);
		if (loc!=null)
			return loc.name;
		else return inst.toLocal(idx);
	}

	// Return string that uniquely identifies this method (see Trek.getObject)
	public String toObjectId()
	{
		// Construct and return qualclass.method(desc)#stmtnum
		String methname = toName();
		String desc = toSignature();
		return parent.toQualifiedName() + "." + methname + "(\"" + desc + "\")";
	}

	// In this context, constructor is handled as method.
	// Return string representation of method: access-flags return name(args),
	// but arg names themselves are not available, so just use a<n>.
	public String toString() {return toString(null);}

	String toString(String classname)
	{
		// Calc and combine pieces of method dcl:
		// access flags, ret type (unless constructor), name, arglist
		String access = Datum.toAccessString(accessFlags, this);
		String ret = isConstructor() ? "" : toType(0) + " ";
		String name = toRef(classname);
		String arglist = toArgList();
        return access + ret + name + arglist;
	}

	// Returns source for method's throws clause (or null string if none)
	String toThrowsString()
	{
		if (thrown==null) return "";

		// Concat each exception's name
		String str = null;
		for (int i=0; i<thrown.length; i++) {
			str = (i==0) ? " throws " : str + ", ";
			str += parent.constant(thrown[i]).toRef();
		}
		return str;
	}

	// Returns throws-clause exceptions as array of fully qualified names
	public String[] toThrowsTypes()
	{
		if (thrown==null) return new String[0];

		// Scan thru list, converting each const pool entry to qual nam
		String thr[] = new String[thrown.length];
		for (int i=0; i<thrown.length; i++) {
			TypeRef ref = (TypeRef)parent.constant(thrown[i]);
			thr[i] = ref.toQualifiedName();
		}
		return thr;
	}

	// Returns ext-form type of the method (ie. its return-value declaration)
	// mode = 1 says qualified type, 0 says unqualified
	public String toType() {return toType(1);}
	private String toType(int mode)
	{
		return parent.toDef(null, toDesc(), mode);
	}

    // Write the Code attribute
	// chan = where to write to
	// Notes:
	//	This is hardest piece, since this piece has a length field and itself
	//	contains variable-length pieces, each with a length field
	void writeCode(DataOutput chan) throws IOException
	{
		// Calc length of code attrs
		int attrlen = 0;
		for (int i=0; i<cattrs.length; i++) {
			attrlen += 2 + 4;			// Name-index and size of attr
			String attstr = cattrs[i].toName(parent);

			// Value of attr (either opaque bytes or the locals table)
			// Note: lvt can be sparse, so length is not just maxlocals*n
			if (attstr.equals("LocalVariableTable"))
				attrlen += Local.attrLength(this);
			else if (attstr.equals("LineNumberTable"))
				attrlen += Linum.attrLength(this);
			else attrlen += cattrs[i].value.length;
		}

		// Calc size of (modified) code and  set stmt.delta's from inserted code
		// Write header of code attr: attr len & max's
		int codelen = Statement.codeLength(this);
		chan.writeInt(2+2				// max stk/loc
			+ 4+codelen					// the code and its len-field
			+ 2+excepts.length*8		// the exceptions table & its cnt-field
			+ 2+attrlen);				// all the attributes & attr-cnt-field
		chan.writeShort(maxStack+addStack);
		chan.writeShort(maxLocals+addLocals);

		// Write code, exceptions, and the code attr's attributes
		chan.writeInt(codelen);
		Statement.writeStmts(chan, this);
		ExceptionHandler.writeExcepts(chan, this);
		Attr.writeAttrs(chan, cattrs, parent, this);
	}

	// Write all the methods to the classfile
	static void writeMethods(DataOutput chan, ClassFile cf) throws IOException
	{
		// Write # of methods and each method
		chan.writeShort(cf.methods.size());
		for (int i=0; i<cf.methods.size(); i++) {
			Method meth = (Method)cf.methods.elementAt(i);
			String name = meth.parent.constant(meth.nameIdx).toInternalString();

			// Conform to convention that classfiles have constructors
			// 1st & <clinit> last. Note that in between, the methods are
			// sorted alphabetically -- which may not be true in orig file.
			if (cf.clinit!=meth)
				meth.write(chan);
		}
		if (cf.clinit!=null) cf.clinit.write(chan);
	}

	// Write the current method's thown-exceptions table
	// chan = where to write to
	void writeThrown(DataOutput chan) throws IOException
	{
		int len = thrown.length;
		chan.writeInt(2 + len*2);			// Size of this attr
		chan.writeShort(len);				// # of thrown exceptions
		if (len <= 0) return;

		// Write each entry
		for (int i=0; i<len; i++) {
			chan.writeShort(thrown[i]);
		}									// End loop
		return;
	}

}

