//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

package dec.trek;
import java.io.*;
import java.util.Hashtable;

// Handles method and interface-method ref constants. There is no processing
// diffs between the two. The only diff is to capture it on read and
// preserve it on write.

public class MethodRef extends MemberRef
{
	private boolean fromInterface;		// True if method sig from an Interface

	// Add method ref
	MethodRef(boolean from, int clidx, int natidx, ClassFile cf)
	{
		super(clidx, natidx, cf);
		fromInterface = from;
	}

	// Construct a start-scan cp entry
	MethodRef(ClassFile cf)
	{
		super(cf);
	}

	// Add a method ref, if it doesnt already exist.
	// Either way, return a matching pool entry.
	// name = member's name
	// desc = member's descriptor
	// cf = classfile want to know if ref is defined within
	protected static MethodRef add(String name, String desc, ClassFile cf)
	{
		return (MethodRef)MemberRef.add(Cp_methodref, name, desc, cf);
	}

	// Scan for matching ref entry
	// name = its fully qualified internal name
	// desc = its internal data-type info
	// cf = classfile to search
	static MethodRef get(String name, String desc, ClassFile cf)
	{
		return (MethodRef)MemberRef.get(Cp_methodref, name, desc, cf);
	}

    // Return count of args of a method ref
	int getArgCount()
	{
		String desc = parent.constant(natIdx).toDesc();
		return Method.getArgCount(desc, parent);
	}

	// Return the method ref's method, or null
	public Method getObject()
	{
		return (Method)getObject(0);
	}

	// Returns true if specified object has same type as "this".
	protected boolean isSameType(CpInfo cpi)
	{
		return cpi instanceof MethodRef;
	}

	// Returns non-0 if reffed method has no return value,
	// but -1 for constructor and 1 otherwise
	int isVoidCall()
	{
		// By def, a constructor has a return val
		CpInfo cpi = parent.constant(natIdx);
		String name = cpi.toName();
		if (name.startsWith("<init>"))
			return -1;

		// Otherwise chk descriptor of method's return-property
		if (toDesc().startsWith("V"))
			return 1;
		else return 0;					// not-void
	}

	// Returns unique key for specified object
	protected Object keyOf()
	{
		return (char)Cp_methodref + toInternalString();
	}

	// Scan for next entry of specified type of constant
	// Returns matching constant or null
	public MethodRef next()
	{
		return (MethodRef)nextOf();
	}

	// Redirect a method reference to another class -- presumably
	// to optimize a method reference to a subclass
	// qcname = fully qualified external class name of class to redirect to
	// mname = name of method (in gcname) to redirect to, or null for same
	// desc = descriptor of new method, or null for same
	public void redirect(String qcname, String mname, String desc)
	{
		parent.cphash.remove(keyOf());

		// Get ptr to needed class ref
		qcname = qcname.replace('.', '/');
		TypeRef tref = TypeRef.add(qcname, parent);
		classIdx = tref.idx;

		// Build updated nat entry if either name or desc specified
		if (mname!=null || desc!=null)
		{
			NameAndType nat = (NameAndType)parent.constant(natIdx);
			if (mname==null) mname = nat.toName();
			if (desc==null) desc = nat.toDesc();
			nat = NameAndType.add(mname, desc, parent);
			natIdx = nat.idx;
		}

		// Modify entry and give it new hash entry
		parent.modified = true;
		parent.cphash.put(keyOf(), this);
	}

	// Get descriptor of method ref's entire arg list
	String toArgsDesc()
	{
		return Method.toArgsDesc(toSignature());
	}

	// Return a source-oriented string repr of the specified cp entry
	public String toString()
	{
		NameAndType nat = (NameAndType)parent.constant(natIdx);
		Method method = new Method(parent, nat.nameIdx,	nat.descIdx);
		return method.toString(getTypeRef().toString());
	}

	// For dump report
	protected String toTag()
	{
		return "(Method)   ";
	}

	// Write a method ref constant, preserving if from interface
	protected void write(DataOutput chan) throws IOException
	{
		chan.writeByte(fromInterface ? Cp_intmethodref : Cp_methodref);
		super.write(chan);
	}

}

