//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

package dec.trek;
import java.io.*;

// Handles name-and-type constants. These are only used by ref constants, btw.

class NameAndType extends CpInfo
{
	// Entities are classes, interfaces, & members (methods and fields)
	// Entities are identified by their name -- and if a member -- their data
	// type as well. This type info is represented as a "descriptor string".
	int nameIdx;			// Cp entry index of entry's name
	int descIdx;			// Cp entry index of member's descriptor string

	// Create nat constant
	NameAndType(int namidx, int descidx, ClassFile cf)
	{
		super(cf);
		nameIdx = namidx;
		descIdx = descidx;
		cf.addConstant(this);
	}

	// Add nat constant to pool if doesnt exist already. Either way, ret ptr
	static NameAndType add(String name, String desc, ClassFile cf)
	{
		NameAndType nat = get(name, desc, cf);
		if (nat!=null) return nat;

		// Add subsid elements as needed and then nat ref
		StringStorage ssn = StringStorage.add(name, cf);
		StringStorage ssd = StringStorage.add(desc, cf);
		return new NameAndType(ssn.idx, ssd.idx, cf);
	}

	// Return matching constant or null
	private static NameAndType get(String name, String desc, ClassFile cf)
	{
		return (NameAndType)CpInfo.get(name + ":" + desc, cf);
	}

	// Returns true if specified object has same type as "this".
	protected boolean isSameType(CpInfo cpi)
	{
		return cpi instanceof NameAndType;
	}

	// Returns unique key for specified object
	protected Object keyOf()
	{
		return toInternalString();
	}

	// Returns descriptor of pool entry
	// wantdef = unused
	String toDesc()
	{
		return parent.constant(descIdx).toInternalString();
	}

	// Return string(s) associated with specified constant pool entry.
	// Differs from toString in that member refs contain raw desc and /'s
	// rather than source-code signature and .'s.
	String toInternalString()
	{
		return toName() + ":" + toDesc();
	}

    // Returns name field of nat constant
	public String toName()
	{
		return parent.constant(nameIdx).toInternalString();
	}

	// Return a source-oriented string repr of the specified cp entry
	public String toString()
	{
		return toInternalString();
	}

	// For dump report
	protected String toTag()
	{
		return "(NameType) ";
	}

	// Write type of constant and its data
	protected void write(DataOutput chan) throws IOException
	{
		chan.writeByte(Cp_nameandtype);
		chan.writeShort(nameIdx);
		chan.writeShort(descIdx);
	}

}

