//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

package dec.trek;
import java.io.*;

class StringConstant extends ValueRef
{
	private int strIdx;			// Index of its string storage constant

	// Add scalar value entry
	StringConstant(int stridx, ClassFile cf)
	{
		super(cf);
		strIdx = stridx;
		cf.addConstant(this);
	}

	// Add scalar value entry if doesnt exist. Either way, return ptr
	static StringConstant add(String sval, ClassFile cf)
	{
		StringConstant sc = get(sval, cf);
		if (sc!=null) return sc;

		// Add subsid element as needed and then string ref
		StringStorage ss = StringStorage.add(sval, cf);
		return new StringConstant(ss.idx, cf);
	}

	// Return matching scalar value entry, or null
	private static StringConstant get(String sval, ClassFile cf)
	{
		return (StringConstant)CpInfo.get((char)Cp_string+sval, cf);
	}

	// Returns true if specified object has same type as "this".
	protected boolean isSameType(CpInfo cpi)
	{
		return cpi instanceof StringConstant;
	}

	// Returns unique key for specified object
	protected Object keyOf()
	{
		return (char)Cp_string + toInternalString();	
	}

	// Returns descriptor of pool entry
	String toDesc()
	{
		return "Ljava/lang/String;";
	}

	// Return internal string repr of the specified cp entry
	public String toInternalString()
	{
		return parent.constant(strIdx).toInternalString();
	}

	// Return a source-oriented string repr of the specified cp entry
	public String toString()
	{
		String sval = toInternalString();
		int len = sval.length();
		StringBuffer svbuf = new StringBuffer(len+6);
		for (int i=0; i<len; i++) {
			char ch = sval.charAt(i);
			if (ch == '"')
				svbuf.append("\\\"");
			else if (ch == '\\')
				svbuf.append("\\\\");
			else if (ch == '\b')
				svbuf.append("\\b");
			else if (ch == '\f')
				svbuf.append("\\f");
			else if (ch == '\n')
				svbuf.append("\\n");
			else if (ch == '\r')
				svbuf.append("\\r");
			else if (ch == '\t')
				svbuf.append("\\t");
			else if (Character.isISOControl(ch) || ch>=256)
			{
				String fourhex = Integer.toHexString(0x10000+ch).substring(1);
				svbuf.append("\\u" + fourhex);
			}
			else svbuf.append(ch);
		}
		return '"' + svbuf.toString() + '"';
	}

	// For dump report
	protected String toTag()
	{
		return "(String)   ";
	}

	// Write pool entry for this constant
	protected void write(DataOutput chan) throws IOException
	{
		chan.writeByte(Cp_string);
		chan.writeShort(strIdx);
	}

}

