//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

package dec.trek;
import java.io.*;

class StringStorage extends CpInfo
{
	private String string;

	// Add scalar value entry
	StringStorage(String sval, ClassFile cf)
	{
		super(cf);
		string = sval;
		cf.addConstant(this);
	}

	// Add scalar value entry if doesnt exist. Either way, return ptr
	static StringStorage add(String sval, ClassFile cf)
	{
		StringStorage ss = get(sval, cf);
		if (ss!=null) return ss;
		return new StringStorage(sval, cf);
	}

	// Returns true if specified object has same type/value(s) as "this".
/*
	boolean equals(CpInfo cpi)
	{
		return isSameType(cpi) && cpi.string.equals(string);
	}
*/

	// Return matching scalar value entry, or null
	private static StringStorage get(String sval, ClassFile cf)
	{
		return (StringStorage)CpInfo.get((char)Cp_rawstring+sval, cf);
	}

	// Returns true if specified object has same type as "this".
	protected boolean isSameType(CpInfo cpi)
	{
		return cpi instanceof StringStorage;
	}

	// Returns unique key for specified object
	protected Object keyOf()
	{
		return (char)Cp_rawstring + toInternalString();	
	}

	// Return internal string repr of the specified cp entry
	public String toInternalString()
	{
		return string;
	}

	// Return a source-oriented string repr of the specified cp entry
	public String toString()
	{
		return '"' + string + '"';
	}

	// For dump report
	protected String toTag()
	{
		return "(Storage)  ";
	}

	// Write pool entry for this constant
	protected void write(DataOutput chan) throws IOException
	{
		chan.writeByte(Cp_rawstring);
		chan.writeUTF(string);
	}

}

