//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

package dec.trek;
import java.io.*;

// For class and arrayrefs. Used in instancof/cast and new insts

public class TypeRef extends ThingRef
{
	ClassFile inner;				// Its classfile if TypeRef is inner class
	private int nameIdx;			// Cp index of class's name (or array's desc)
	private boolean temp;			// True if just a marker for import
	boolean used;					// True if class used in curr classfile

	// Add scalar value entry
	TypeRef(int namidx, ClassFile cf)
	{
		super(cf);
		nameIdx = namidx;
		cf.addConstant(this);
	}

	// Construct a start-scan cp entry
	TypeRef(ClassFile cf)
	{
		super(cf);
	}

	// Add a class entry, including any missing subsid entries
	static TypeRef add(String name, ClassFile cf)
	{
		// See if it already exists
		TypeRef tref = get(name, cf);
		if (tref!=null)
		{
			tref.temp = false;		// Make real if wasnt already
			return tref;
		}

		// Add subsid element as needed and then class ref
		StringStorage ss = StringStorage.add(name, cf);
		return new TypeRef(ss.idx, cf);
	}

	// Add a dummy reference (to mark an imported class)
	static TypeRef addTemp(String name, ClassFile cf)
	{
		// See if it already exists
		TypeRef tref = get(name, cf);
		if (tref!=null)
			return tref;

		// Add subsid element as needed and then class ref
		StringStorage ss = StringStorage.add(name, cf);
		tref = new TypeRef(ss.idx, cf);
		tref.temp = true;
		return tref;
	}
	
	// Scan for matching ref entry
	// name = its fully qualified internal name
	// cf = classfile to search
	static TypeRef get(String name, ClassFile cf)
	{
		return (TypeRef)CpInfo.get((char)Cp_typeref + name, cf);
	}

	// Return the typeref's info, or null
	public Object getObject()
	{
		String desc = toInternalString();
		if (desc.charAt(0)=='[')
			return null;		// Array ref has no name
		else return parent.parent.getClassFile(desc);
	}

	// Returns true if specified object has same type as "this".
	protected boolean isSameType(CpInfo cpi)
	{
		return cpi instanceof TypeRef;
	}

	// Returns true if class constant created just to mark imported class
	protected boolean isTempItem()
	{
		return temp;
	}
	
	// Returns unique key for specified object
	protected Object keyOf()
	{
		return (char)Cp_typeref + toInternalString();	
	}

	// Scan for next entry of specified type of constant
	// Returns matching constant or null
	public TypeRef next()
	{
		return (TypeRef)nextOf();
	}

	// Returns descriptor of pool entry
	String toDesc()
	{
		// Array ref already desc. For class, convert to desc fmt
		String desc = toInternalString();
		if (desc.charAt(0)!='[')
			desc = "L" + desc + ";";
		return desc;
	}

	// Return internal string repr of the specified cp entry
	String toInternalString()
	{
		return parent.constant(nameIdx).toInternalString();
	}

    // Return external [un]qualified class name
	public String toName() {return toNameRef(false);}
	public String toQualifiedName() {return toNameRef(true);}
	private String toNameRef(boolean qual)
	{
		String desc = toInternalString();
		if (desc.charAt(0)=='[')
			return null;		// Array ref has no name

		// Isolate & externalize the name, and include pkg if desired
		String pkgout[] = qual ? new String[1] : null;
		String name = ClassFile.toNamePieces(desc, true, pkgout);
		return qual ? pkgout[0] + name : name;
	}

    // Returns source-code string assoc with this entry.
	public String toString()
	{
		return parent.toDef(null, toDesc(), 0);
	}

	// For dump report
	protected String toTag()
	{
		return "(Type)     ";
	}

	// Write pool entry for this constant
	protected void write(DataOutput chan) throws IOException
	{
		chan.writeByte(Cp_typeref);
		chan.writeShort(nameIdx);
	}

}

