//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

package dec.trek;
import java.util.Vector;

////
// Vmi. Execute the instruction codes of the JVM
//
//
class Vmi 
{
	protected ClassFile cf;
	private boolean didAssign;			// True if stmt contains store inst
	protected Instruction inst;			// Temp for curr instruction
	protected Method meth;
	private Vector results;				// Statement (and intermediate) results
	private Vector stack = new Vector(10, 10);
										// The vm opd stack
	protected int top;					// Index of top item of stack's vector

	// ifexpr/condexr support. Arrays since an opd in a test can itself
	// be a condexpr. Conversely condend special because end of a?b:c is
	// after its ifLevel is decr (when end of "a" seen).
	int ifLevel;						// Level of test-expr nesting
	int condMark[] = new int[6];		// Stacktop idx before push topmost test
	int condEnd[];						// Addr past active a?b:c, if non-0
	int ifEnd[] = new int[6];			// Addr past end of curr if clause
	Vector ifExpr[] = new Vector[6];	// Where curr test expr accumulated

	// Info returned for largest overall expression in curr stmt.
	// So for assignment stmt, is dest, =, and source.
	String lopd;						// Left operand
	int opr;							// Operator (Trek.DOIF_* or op_*)
	String ropd;						// Right operand

	// Constant-mapping contexts
	private static final int mc_array = 1;
	private static final int mc_cast = 2;
	private static final int mc_all = -1;

	// Types of instructions. Used primarily for end-of-stmt analysis.
	// Must match IN_* in Trek class.
	static final byte it_binop = 0;
	static final byte it_dup = 1;
	static final byte it_loadel = 2;
	static final byte it_loadconst = 3;
	static final byte it_loadfld = 4;
	static final byte it_loadlv = 5;
	static final byte it_new = 6;
	static final byte it_storel = 7;
	static final byte it_storefld = 8;
	static final byte it_storelv = 9;
	static final byte it_unop = 10;
	static final byte it_branch = 11;
	static final byte it_call = 12;
	static final byte it_if = 13;
	static final byte it_return = 14;
	static final byte it_switch = 15;
	static final byte it_other = 16;

	// Constant pre-built datums
	private static final Datum nullDatum
		= new Datum("null", "Ljava/lang/Object;", Datum.softy, Term.op_pconst);
	private static final Datum booleanDatum = new Datum("", "Z");
	private static final Datum emptyDatum = new Datum("", "X");
	private static final Datum doubleDatum = new Datum("", "D");
	private static final Datum floatDatum = new Datum("", "F");
	private static final Datum longDatum = new Datum("", "J");
	private static final Datum missingDatum = new Datum("<missing>", "X");

	// Primitive data types (index as defined by classfile format)
	private static final String datatypes[] = {
		null, null, null, null,
		"boolean", "char", "float", "double",
		"byte", "short", "int", "long"
	};
	protected static final char descs[] = {
		0, 0, 0, 0,
		'Z', 'C', 'F', 'D', 'B', 'S', 'I', 'J'
	};

	// Datatype codes for newarray, with dt_ref added to list
	static final int dt_ref = 3;
	static final int dt_boolean = 4;
	static final int dt_char = 5;
	static final int dt_float = 6;
	static final int dt_double = 7;
	static final int dt_byte = 8;
	static final int dt_short = 9;
	static final int dt_int = 10;
	static final int dt_long = 11;

	// The sort of action each instruction performs
	static final byte types[] = {
		it_other, it_loadconst, it_loadconst, it_loadconst, it_loadconst,
		it_loadconst, it_loadconst, it_loadconst, it_loadconst, it_loadconst,	//9
		it_loadconst, it_loadconst, it_loadconst, it_loadconst, it_loadconst,
		it_loadconst, it_loadconst, it_loadconst, it_loadconst, it_loadconst,	//19
		it_loadconst, it_loadlv, it_loadlv, it_loadlv, it_loadlv, 
		it_loadlv, it_loadlv, it_loadlv, it_loadlv, it_loadlv,	//29
		it_loadlv, it_loadlv, it_loadlv, it_loadlv, it_loadlv, 
		it_loadlv, it_loadlv, it_loadlv, it_loadlv, it_loadlv,	//39
		it_loadlv, it_loadlv, it_loadlv, it_loadlv, it_loadlv, 
		it_loadlv, it_loadel, it_loadel, it_loadel, it_loadel,	//49
		it_loadel, it_loadel, it_loadel, it_loadel, it_storelv, 
		it_storelv, it_storelv, it_storelv, it_storelv, it_storelv,	//59
		it_storelv, it_storelv, it_storelv, it_storelv, it_storelv, 
		it_storelv, it_storelv, it_storelv, it_storelv, it_storelv,	//69
		it_storelv, it_storelv, it_storelv, it_storelv, it_storelv, 
		it_storelv, it_storelv, it_storelv, it_storelv, it_storel,	//79
		it_storel, it_storel, it_storel, it_storel, it_storel, 
		it_storel, it_storel, it_other, it_other, it_dup,//89
		it_dup, it_dup, it_dup, it_dup, it_dup, 
		it_other, it_binop, it_binop, it_binop, it_binop,	//99
		it_binop, it_binop, it_binop, it_binop, it_binop, 
		it_binop, it_binop, it_binop, it_binop, it_binop,	//109
		it_binop, it_binop, it_binop, it_binop, it_binop, 
		it_binop, it_unop, it_unop, it_unop, it_unop,	//119
		it_binop, it_binop, it_binop, it_binop, it_binop, 
		it_binop, it_binop, it_binop, it_binop, it_binop,	//129
		it_binop, it_binop, it_other, it_unop, it_unop, 
		it_unop, it_unop, it_unop, it_unop, it_unop,	//139
		it_unop, it_unop, it_unop, it_unop, it_unop, 
		it_unop, it_unop, it_unop, it_binop, it_binop,	//149
		it_binop, it_binop, it_binop, it_if, it_if, 
		it_if, it_if, it_if, it_if, it_if,				//159
		it_if, it_if, it_if, it_if, it_if, 
		it_if, it_if, it_branch, it_other, it_other,	//169
		it_switch, it_switch, it_return, it_return, it_return, 
		it_return, it_return, it_return, it_loadfld, it_storefld,	//179
		it_loadfld, it_storefld, it_call, it_call, it_call, 
		it_call, it_other, it_new, it_new, it_new,		//189
		it_unop, it_return, it_other, it_other, it_other, 
		it_other, it_other, it_new, it_if, it_if,		//199
		it_branch, it_other, it_other, it_other, it_other,
		it_other, it_other, it_other, it_other, it_other,	//209
		it_other, it_other, it_other, it_other, it_other,
		it_other, it_other, it_other, it_other, it_other,	//219 
		it_other, it_other, it_other, it_other, it_other, 
		it_other, it_other, it_other, it_other, it_other,	//229
		it_other, it_other, it_other, it_other, it_other, 
		it_other, it_other, it_other, it_other, it_other,	//239
		it_other, it_other, it_other, it_other, it_other, 
		it_other, it_other, it_other, it_other, it_other,	//249
		it_other, it_other, it_other, it_other, it_other, it_other
	};

	// Store top of stack in specified local variable & store l/ropd
	// lidx = identifies local to store value into
	protected void assop(int lidx)
	{
		String destname;
		Datum source = pop();

		// When constructed lvt, no-lvt-entry can only happen during explode
		// of method prior to buildLvt being called. If there is comp-supplied
		// lvt, no lvt entry means that dest is invis local var (eg. in Sync).
		// Either way, use a constructed name in assign. But if temp local
		// var, want the effect of checkReload to be to use the source
		// of the assign not the entire assign.
		Datum destdtm;
		Local loc = meth.getLocal(lidx, inst);
		if (loc==null)
		{
			destname = inst.toLocal(lidx);
			destdtm = new Datum(destname, inst.toDesc(source.desc));
			if (meth.lvRead>=0)
				destdtm.dest = "templocal";
		}

		// Otherwise use local to set up dest of assign
		else {
			destname = loc.name;
			destdtm = new Datum(
				loc.toRef(cf, inst.parent.start + inst.start + inst.size),
				loc.desc);
		}

		// Do the assign, but in case dest included data type, ovwrite lopd
		assop(destdtm, source);
		lopd = destname;
	}

	// Do a store operation, appending result to results vector unless
	// c-left-on-stack case applies (see #1 in Notes and pushAssign)
	// dest = left side of assignment
	// source = right side
	// Notes:
	//	Java allows short (etc) xxx++ or xxx opr=n, even tho it does not
	//	allow short xxx = xxx + 1 (cause right opd is actually an int
	//	expression). The gotcha is that it generates the "correct" code
	//	for xxx++ (ie. load, add, int2short, store). So assop does not
	//	"source optimize" this case: it seems better to give programmer
	//	feedback that he has written poorly performing code.
	//
	//	If context is something like: "(b=c) + d", either:
	//	1) c has been left on stack before the store into b. So c needs
	//		to be discarded since at source-level it is "b=c" that next
	//		expr wants to use (versus the dup-ed c during execution).
	//	2) there will be load of b after the store into b. So the load
	//		is redundant (ie. next expr wants b=c not b) & didAssign is set
	//		to cause load-arg to be discarded rather than pushed
	protected void assop(Datum dest, Datum source)
	{
		// Setup return info. If applic,
		// first gen source for cond expr and canon form of constant source.
		// Raw source is needed to compare with un-normalized item on stack
		String rawright = source.src;
		source = doCondExpr(source, dest.desc);
		checkSoftRef(source, dest);
		lopd = dest.src;
		opr = Term.op_assign;
		ropd = mapConstant(source, dest.desc);

		// If dest starts with "new ", this is start of array init stmt
		// Execute the array init so to speak, and leave init expr on stack
		if (lopd.startsWith("new "))
		{
			ArrayInit ai = new ArrayInit(inst.parent);
			Instruction in2 = ai.inst.next();
			for (Instruction in3=in2.next(); ; in3=in3.next()) {
				ai.xinst();
				if (in3==null) break;	// Dont process array store
				ai.inst = in2;
				in2 = in3;
			}

			// Put overall array init expr into main context and setup to
			// process array-store inst next time assop called
			push(ai.stringify(ai.outermost), peek().desc);
			inst = ai.inst;
			return;
		}

		// See if temp-local hack applies before propagating dest's
		// properties to the assign-expr. Also use dest's properties
		// to firm up a soft source if can
		boolean temp_local = dest.dest!=null;
		dest.dest = null;
		Datum assexpr = dest.copy();

		// Is dest already on stack because of load-incr expr?
		// Is source already on the stack and the store is not the
		// last inst of the stmt (the "c left on stack" case)?
		Datum tmp = peek();			// Get top of stack
		boolean loadincr = lopd.equals(tmp.src);
		boolean last = inst.parent.insts.length <= inst.start+inst.size;
		if (loadincr || (!last && rawright.equals(tmp.src)))
			pop();					// Yes, replace with assign expr
		else {						// No, save this result
			assexpr.dest = lopd;
			didAssign = true;
		}

		// Main case is vanilla assignment opr. But if assigning to temp local,
		// dont make it visible -- just use source of assign.
		String prefix = lopd + " ";
		if (!ropd.startsWith(prefix))
		{
			String assign = temp_local ? ropd : lopd + " = " + ropd;
			pushAssign(assign, assexpr);
			return;
		}

		// Updating dest in some way (eg. dest = dest + 1).
		// 1) Handle +/-1 special cases.
		String suffix = ropd.substring(prefix.length());
		if (suffix.endsWith("1"))
		{
			boolean plus1 = suffix.equals("+ 1");
			boolean minus1 = suffix.equals("- 1") || suffix.equals("+ -1");
			if (plus1 || minus1)
			{
				// If incr ends stmt, use the more common i++ notation.
				// If load&incr, must use i++/-- notation because only it
				// means use orig value and then change the dest var.
				if (inst.start+inst.size >= inst.parent.insts.length || loadincr)
				{
					if (plus1)
						pushAssign(lopd + "++", assexpr);
					else pushAssign(lopd + "--", assexpr);
				}

				// Otherwise must use ++i notation because
				// encompassing expression needs assexpr of incr, not orig val.
				else {
					if (plus1)
						pushAssign("++" + lopd, assexpr);
					else pushAssign("--" + lopd, assexpr);
				}
				return;
			}
		}						// End of If ends in 1

		// Output approp assign. Note that right term is implicitly 
		// enclosed in parentheses in opr= case.
		// 2) "dest OPR= rest_of_source"
		// 3) Regular assignment
		// Note: Object o = o + s is legal syntax, Object o += s isnt.
		int pfxlen = prefix.length();
		char chop = ropd.charAt(pfxlen);
		if (ropd.charAt(pfxlen+1)==' '
			&& "+-*/%&|^".indexOf(chop)>=0
			&& !dest.desc.equals("Ljava/lang/Object;"))
			pushAssign(prefix + chop + "=" + ropd.substring(pfxlen+1), assexpr);
		else pushAssign(lopd + " = " + ropd, assexpr);
	}

	// Perform an instruction of form: stack1 = stack1 opr stack2
	protected void binop(int opr)
	{
		Datum val2 = pop();
		Datum val1 = pop();

		// With respect to firming up data types of the opds,
		// about only thing can do is disting boolean and non-boolean.
		// In partic, a+b says nothing about length of other opd.
		val1 = constrainOperand1(val1, opr, val2);
		val2 = constrainOperand1(val2, opr, val1);

		// On the other hand, integer expressions are always 32-bit.
		// So except in maybe-boolean case, the expr says
		// nothing about the data types of soft opds. Similarly
		// desc is firm except in that case.
		if (val1.soft==Datum.softy && val2.soft==Datum.softy)
		{
			Datum expr = isolate(val1, opr, val2, val1.desc, Datum.softy);
			expr.loc = val1.loc!=null ? val1.loc : val2.loc;
			push(expr);
		}
		else push(isolate(val1, opr, val2, val1.desc, Datum.firm));
	}

	// Perform invoke instruction
	// inst = the invoke instruction to process
	private void callop(Instruction inst)
	{
		// Get number of args, via methodref pool entry
		CpInfo cpi = cf.constant(inst.readUnsignedShort(inst.start+1));
		StringBuffer args = new StringBuffer(64);
		int numargs = cpi.getArgCount();

		// Pop and re-store args to get them in right order
		if (numargs>0)
		{
			Datum argarr[] = new Datum[numargs];
			for (int i=numargs-1; i>=0; i--) argarr[i] = pop();

			// Build the arglist of the call, mapping boolean and char consts
			String desc = Method.toArgsDesc(cpi.toSignature());
			int pos = 0;
			for (int i=0; i<numargs; i++) {

				// Get data type of curr arg. Then
				// (map and) append it to arglist (or analyze local)
				int len = Datum.getDescLength(desc, pos);
				String argdesc = desc.substring(pos, pos+len);
				pos += len;
				doArg(argarr[i], argdesc, args);
			}
		}

		// Build ref expr for the call, and get type of call.
		Datum refdtm = getMemberRef(inst, "(" + args + ")");
		int vc = cpi.isVoidCall();

		// Mechanically a constructor is a void call. So store a result
		// unless this is "c (ie. new) left on stack" case. For that,
		// replace the "new" with the execution of the constructor.
		if (vc<0)							// Constructor
		{
			if (peek().src.equals("new"))
			{
				pop();
				push(refdtm);
			}
			else results.addElement(refdtm);
		}

		// Void call leaves nothing on stack, so setup result.
		// Non-void call leaves execution of call on stack (and is followed
		// by pop if the call itself is the result of its stmt).
		else if (vc>0)
			results.addElement(refdtm);		// Void call
		else push(refdtm);					// Non-void call
	}

	// Perform an instruction that equates to: (opr)src
	// opr = source form of cast-to
	// ddesc = cast-to as descriptor
	// sdesc = cast-from as descriptor
	protected void castop(String opr, String ddesc, Datum sdesc)
	{
		// sdesc non-null means a casting-inst (eg. int2char) was done. So
		// we can set src's desc to the type ident by the inst, namely sdesc.
		// Conversely chk if ddesc imples anything about sdtm.
		Datum sdtm = pop();
		sdtm = doCondExpr(sdtm, "X");
		checkSoftResult(sdtm, sdesc);
		checkNoopCast(sdtm, ddesc);

		// Combine opr and src, and push compound datum built from this
		// longer-int = shorter-int doesnt say alot about type of dest,
		// but xxx = (ddesc)yyy pretty much says xxx is a ddesc.
		String src = isolate(sdtm.src, sdtm.opr, Term.op_cast);
		push(new Datum(opr + src, ddesc, Datum.firm, Term.op_cast));
	}

	// If left and right are diff classes and not either is todesc,
	// then cast them in condit expr
	// todesc = target type of cond expr
	// left/right = terms after ?
	// Notes:
	//	Object = x ? Double : Long needs cast but ? Double : Number
	//	doesnt because Number/Double have super/sub relationship.
	//	Case is unusual enough, just do cast all the time.
	private void castPeerTypes(String todesc, Datum left, Datum right)
	{
		if (!todesc.startsWith("L"))
			return;
		if (!left.desc.equals(right.desc)
			&& !todesc.equals(right.desc)
			&& !todesc.equals(left.desc))
		{
			String casto = toCast(todesc);
			left.src = casto + isolate(left.src, left.opr, Term.op_cast);
			right.src = casto + isolate(right.src, right.opr, Term.op_cast);
		}
	}

	// If appear to have (Xxx)var-of-type-Xxx, then in reality we dont
	// yet know type of the variable. So reset it to full softness
	private void checkNoopCast(Datum dtm, String desc)
	{
		if (dtm.desc.equals(desc))
		{
			dtm.soft = Datum.softy;
			if (dtm.loc!=null)
				dtm.loc.soft = Datum.softy;
		}
	}

	// If dest of assign matches val's text. Effectively replace val with
	// the assign's datum.
	// Notes:
	//	It is possible to handcode an inst sequence that equates to
	//	k = (i = (i=a+b) + c) in which the inner store-to-i is not done.
	//	No compiler is known to do this, but it would equate to merging
	//	results(i-1).src with results(i).
	private Datum checkReload(Datum val)
	{
		int max = results.size() - 1;
		for (int i=max; i>=0; i--) {
			Datum item = (Datum)results.elementAt(i);
			if (item.dest!=null && item.dest.equals(val.src))
			{
				results.removeElementAt(i);
				return item;
			}
		}
		return val;
	}

	// Return (updated opd1), constraining it to boolean or integer if can,
	// and if opd1 is unexpanded cond expr, expand it
	// opd1 = datum to chk if completely soft
	// opr = opr to chk for boolean compat
	// opd2 = datum to chk for boolean compat
	protected Datum constrainOperand1(Datum opd1, int opr, Datum opd2)
	{
		// Return if opd1 already constrained/expanded. If opd1 is soft,
		// it must be object, integer, boolean, or unexpanded cond expr.
		if (opd1.soft<=Datum.firm)
			return opd1;

		// It's "hard" only if opd2 is an integer
		String desc = "X";		// Usually cant constrain opd1 data type
		switch (opd2.desc.charAt(0)) {
		case 'Z':
			desc = "Z";
			checkSoftRef(opd1, booleanDatum);
			break;

		// Boolean-compat oprs are only Eq, Ne, And, Or, Xor.
		// Also there must be chance opd2 is boolean for opd1 to be boolean
		case 'I':
			if (opd2.soft==Datum.softy && (opr<=Trek.DOIF_NE
				|| opr==Term.op_or || opr==Term.op_xor || opr==Term.op_and))
				return doCondExpr(opd1, desc);

		// Opd2 being integer says opd1 is integer, but nothing about its len.
		case 'B':
		case 'C':
		case 'S':
			opd1.refineSoftness(Datum.byplus);
		}

		// Even tho opd1 could be int and opd2 char, if using constant
		// with variable, might as well represent opd1 as char constant
		// (or boolean constant) if it is an integer constant.
		// Return what learned about opd1 and expand cond expr if applic
		mapConstant(opd1, opd2.desc);
		return doCondExpr(opd1, desc);
	}

	// Firm up desc of dtm's local if any (see Loci)
	protected void checkSoftRef(Datum dtm, Datum from) {}
	protected void checkSoftResult(Datum dtm, Datum from) {}

	// Map appends to a + b ... + z and return result
	// ref = call expression that is really string concatenations
	private Datum concatop(String ref)
	{
		int applen = "append(".length();
		String val = "";
		int pos = "new StringBuffer(".length();

		// Scan thru ref, mapping each StringBuffer/append arg to +xxx
		// Note that smartIndexOf skips nested parens and quoted strings
		while (true) {
			int endpos = Datum.smartIndexOf(ref, 0, ")", pos);
			if (endpos < 0)
				throw new TrekException("No matching right parenthesis");

			// Use arg of append as next term in series of string-plus's,
			// unless it is empty (in partic StringBuffer()). Also, code
			// can also pass non-null String arg to new StringBuffer. And
			// this is created by doing String.valueOf of atomic 1st term.
			endpos += 2;				// Hop over the ).
			if (endpos-2 > pos)
			{
				String valof = "String.valueOf(";
				String term = ref.substring(pos, endpos-2);
				if (val.length()!=0)
					val += " + ";
				else if (term.startsWith(valof) && term.indexOf(" ")<0)
					term = term.substring(valof.length(), term.length()-1);

				// If arg of append/valueOf is mixed-mode expr, need
				// to isolate it from the +'s. Since datums are long
				// done, do hack of looking for space not in quoted string.
				if (Datum.smartIndexOf(term, -1, " ", 0)>=0)
					val += "(" + term + ")";
				else val += term;
			}

			// Go past next append if any
			pos = ref.indexOf("append(", endpos);
			if (pos<0) break;
			pos += applen;
		}

		// Return transformed datum, including its opr
		return new Datum(val, "Ljava/lang/String;", Datum.firm, Term.op_plus);
	}

	// Process this arg
	// dtm = arg to process
	// argdesc = its descriptor
	// args = accumulated args as source string
	// Notes:
	//	Null args always cast since this may be only way to disambiguate
	//	overloaded methods
	void doArg(Datum dtm, String argdesc, StringBuffer args)
	{
		// Output this arg. Insure that, if it is a constant,
		// its normalized repr (eg. 0 to false for boolean) is output.
		dtm = doCondExpr(dtm, argdesc);
		if (dtm.src.equals("null"))
			dtm.src = toCast(argdesc) + "null";
		String val = mapConstant(dtm, argdesc, mc_all);
		if (args.length()>0) args.append(", ");
		args.append(val);
	}

	// Returns placeholder condit expr from pieces on stack and val
	// val = rightmost term of overall condit expr
	protected Datum doCondExpr(Datum val)
	{
		String pl = "#cond" + condEnd[ifLevel+1];
		condEnd[ifLevel+1] = 0;		// Clear condit-expr mode at this level

		// Save terms of cond expr
		push(val);					// So all on stack at same time
		int num = top - condMark[ifLevel + 1];
		Vector datums = new Vector(num);
		for (int i=0; i<num; i++) datums.addElement(pop());

		// Replace it with placeholder
		val = new Datum(pl, val.desc, Datum.softy, Term.op_cond);
		val.condExpr = datums;
		return val;
	}											// End doCondExpr

	// Returns overall condit expr from pieces pointed to by val
	// val = placeholder for overall condit expr
	// desc = descriptor of context (eg. a = b?c:d, the desc of a)
	private Datum doCondExpr(Datum val, String desc)
	{
		// Handle special case of !(x relop y) generating code of
		// (x relop y ? 1 : 0) ^ 1, which looks like ints until
		// "too late" (ie. when usage context shows boolean).
		// Map it back to boolean by replacing ^ 1 with ==0.
		String src = val.src;
		if (val.desc.equals("I") && desc.equals("Z")
			&& src.endsWith(" ^ 1"))
		{
			int len = src.length() - 4;
			val.src = src.substring(0, len) + " == 0";
			val.desc = "Z";
		}

		// If not a cond-expr placeholder, just ret input arg
		if (val.opr!=Term.op_cond || val.condExpr==null)
			return val;

		// Rebuild stack
		Vector datums = val.condExpr;
		int num = datums.size();
		int marker = top;
		for (int i=num-1; i>=0; i--) push((Datum)(datums.elementAt(i)));

		// Build condit expr from its pieces, incl sub condit expr's
		// A particular triplet could be the left or right term in an
		// outer expr. Rsave non-0 says last built a left-term
		int rsidx = -1;
		Datum rsave[] = new Datum[30];
		Datum left = pop();
		Datum right = null;
		while (true) {

			// If leftmost of 3 terms is test, then build outer right term
			// Else building outer left, so save the "unused" right terms
			Datum test = pop();
			while (test.condExpr==null) {
				if (right!=null)
					rsave[++rsidx] = right;
				right = left;
				left = test;
				test = pop();
			}

			// Build src for conditional expr. If it is boolean expr
			// and left & right are 0 & 1, gen inverted ifexpr since
			// and simple boolean assign. The inversion is done since
			// compilers generate inverted tests for boolean cond expr.
			String expr;
			Vector condvec = test.condExpr;
			Term term = (Term)condvec.firstElement();
			if (desc.equals("Z") && right.opr==Term.op_iconst
				&& left.opr==Term.op_iconst && left.intVal==0)
			{
				expr = term.toIfResult(this, condvec, test.intVal, false);
				val = new Datum(expr, "Z", Datum.firm, term.opr);
			}

			// Gen normal case unless is class literal.
			// Isolate right because as of 1.2 assigns must be in parens.
			else {
				String vdesc = val.desc;
				expr = term.toIfResult(this, condvec, test.intVal, true);
				val = getClassLiteral(expr, left, right);
				if (val==null)
				{
					right.src = isolate(right.src, right.opr, Term.op_cond);
					castPeerTypes(desc, left, right);
					expr += " ? " + mapConstant(left, desc, mc_all)
						+ " : " + mapConstant(right, desc, mc_all);
					val = new Datum(expr, vdesc, Datum.softy, Term.op_cond);
				}
			}

			// If sub-expr, use it as left or right of outer expr
			// depending on whether there is unused right term
			if (rsidx>=0)
			{
				right = rsave[rsidx--];
				left = val;
			}
			else {
				right = val;
				if (top<=marker) break;
				left = pop();
			}
		}										// End loop
		return val;
	}											// End doCondExpr

	// Returns true if explicit ptr needed for this member ref
	// ptr = ptr's datum
	// memname = thing being reffed
	// cpmem = info about member it pts to
	// opc = opcode of call inst
	private String doRefExpr(Datum ptr, String memname, CpInfo cpmem, int opc)
	{
		String ptrname = ptr.src;

		// Chk for direct call to method in ptr's superclass. If ptrname is
		// "this", ref is simply super.xxx. Else code is specifically
		// invoking a method on a super class or opposed to a virtual invoke.
		if (opc==Code.invokespecial
			&& !cf.toInternalName().equals(cpmem.toInternalClassName()))
		{
			if (ptrname.equals("this"))
				ptrname = "super";
			else ptrname =
				"(" + toCast(ptr.desc) + ptrname + ")";
		}

		// If have ptr other than "this", fall thru and merge ptr and refexpr.
		// For "this", gen x rather than this.x unless this.x is a field
		// and there is a local named x as well.
		else {
			if (ptrname.length()==0)  return memname;
			if (ptrname.equals("this"))
			{
				if (!(cpmem instanceof FieldRef) || !meth.fieldLocal
					|| meth.getLocal(memname, inst)==null)
					return memname;
			}
/*
			// Can also have ref to "this" of an outer class. Ref it as
			// mem if no conflicts with locals or inner members, else
			// ref it as outer-class.this.mem
			if (ptrname.endsWith(".this"))
			{
				if (Trek.lookup(meth.parent.fields, memname)==null
					&& Trek.lookup(meth.parent.methods, memname)==null
					&& meth.getLocal(memname, inst)==null)
					return memname;
			}
*/
		}
		return isolate(ptrname, ptr.opr, Term.op_atomic) + "." + memname;
	}

	// Perform some kind of dup of top of stack
	private void dupop(int opcode)
	{
		Datum val1, val2, val3;
		switch (opcode) {
		case Code.dup:
		case Code.dup2:
			val1 = pop();
			push(val1);
			push(val1.copy());
			break;
		case Code.dup_x1:
		case Code.dup2_x1:
			val1 = pop();
			val2 = pop();
			push(val1);
			push(val2);
			push(val1.copy());
			break;
		case Code.dup_x2:
		case Code.dup2_x2:
			val1 = pop();
			val2 = pop();
			val3 = pop();
			push(val1);
			push(val3);
			push(val2);
			push(val1.copy());
			break;
		}
	}

	// Perform some kind of dup2 where top 2 words are 2 opds, not a 64-bit opd
	private void dup2op(int opcode)
	{
		Datum val1, val2, val3, val4;
		switch (opcode) {
		case Code.dup2:
			val2 = pop();
			val1 = pop();
			push(val1);
			push(val2);
			push(val1.copy());
			push(val2.copy());
			break;
		case Code.dup2_x1:
			val2 = pop();
			val1 = pop();
			val3 = pop();
			push(val1);
			push(val2);
			push(val3);
			push(val1.copy());
			push(val2.copy());
			break;
		case Code.dup2_x2:
			val2 = pop();
			val1 = pop();
			val3 = pop();
			val4 = pop();
			push(val1);
			push(val2);
			push(val4);
			push(val3);
			push(val1.copy());
			push(val2.copy());
			break;
		}
	}

	// Convert Statement to source-code string by "executing it"
	// stmt = statement to convert
	String execute(Statement stmt)
	{
		if ((Trek.debug&Trek.Db_atsmall) != 0)
			Trek.println(System.out, "Executing a " +
				stmt.type + " at level " + stmt.level + " in " + stmt.parent.toName());

		// If special code, such as inserted code, return boilerplate string
		init(stmt);
		if (stmt.type<=Trek.ST_PRIMORD)
			return "Source-less code (in " + meth.toName() + ")";

		// Scan thru statement's instructions, building source inst-by-inst.
		// If start of this stmt loaded by earlier stmt, put this stuff back
		// on stack first.
		executeLoan(stmt.loan);
		for (inst=stmt.firstInstruction(); inst!=null; inst=inst.next()) {
			//if (meth.toName().equals("keep")) showStack();
			if (!xinst())
				return "EXCEPTION;;";
		}

		// In-lined if special case. Do the phantom assign to the temp
		// local, or return dcl-only local's source if havent yet. Setup
		// inst for stmt context since doing "freestanding" assop.
		if (top>=1 && results.size()==0)
		{
			inst = stmt.firstInstruction();
			Datum from = pop();
			Datum to = pop();
			if (stmt.loan.string==null)
				return cf.toDef(to.src, from.desc, 0) + ";";
			assop(to, from);
		}

		// Return computed source string placed in results.
		// Also handle special case of whole stmt of: string+string. javac
		// considers this illegal, but ms compiler can generate it.
		Datum dtm;
		if (results.size()==0)
			dtm = emptyDatum;
		else dtm = (Datum)results.lastElement();
		String str = dtm.opr!=Term.op_plus ? dtm.src : "";

		// Put proper boilerplate around computed string
		if (stmt.type <= Trek.ST_SEMI)
			return str + ";";
		else if (stmt.type == Trek.ST_IF)
			return "if (" + str + ")";
		else return str;
	}

	// Process loan-stmt
	protected void executeLoan(Statement stmt)
	{
		// If no loan, done. Conversely process any recursive loan 1st
		if (stmt==null)
			return;			
		if (stmt.loan!=null)
			executeLoan(stmt.loan);

		// Process this loan. If from code-less stmt, just create variable
		inst = stmt.firstInstruction();
		if (inst==null)
		{
			int spidx;
			String str = stmt.string;
			String desc = "LObject;";
			if (str!=null && (spidx=str.indexOf(' '))>0)
				desc = "L" + str.substring(0, spidx) + ";";
			push("if" + stmt.start, desc);
		}
		else {
			for (; inst!=null; inst=inst.next()) xinst();
		}
		initLoan();
	}

	// Returns source data type referred to by inst's const pool entry
	// inst = type-referencing instruction (eg. instanceof)
	private String getDataType(Instruction inst)
	{
		// Get class referred to by this inst
		CpInfo cpi = cf.constant(inst.readUnsignedShort(inst.start+1));
		return cpi.toString();
	}

	// Returns descriptor for this class ref
	// inst = instruction to process
	private String getClassDesc(Instruction inst)
	{
		// Get class referred to by this inst
		CpInfo cpi = cf.constant(inst.readUnsignedShort(inst.start+1));
		String dt = cpi.toInternalString();

		// If array, is already in encoded fmt
		if (dt.charAt(0)!='[')
			dt = "L" + dt + ";";
		return dt;
	}

	// Returns desc for class literal if cond expr is that, else null.
	// test = test expr of cond expr
	// left = expr to left of colon
	// right = expr to right of colon
	// Notes:
	//	Internally clname.class is class$name==null ? class$("name") : cl$n.
	//	The test polarity and l/r terms can be either order.
	private Datum getClassLiteral(String test, Datum left, Datum right)
	{
		// Is right or left of form class$("name")?
		String calc = right.src;
		int c$idx = calc.indexOf("class$(\"");
		if (c$idx<0)
		{
			calc = left.src;
			c$idx = calc.indexOf("class$(\"");
		}
		if (c$idx<0 || !calc.endsWith("\")"))
			return null;

		// Right's form is ok, chk if constructed name matches start
		// of cond expr: [outer-class.]class$name but with $'s not .'s.
		String name = calc.substring(c$idx+8, calc.length()-2);
		if (test.indexOf("class$" + name.replace('.', '$'))<0)
			return null;

		// Return datum containing class literal
		if (name.startsWith("java.lang."))
			name = name.substring(10);
		return new Datum(name + ".class", 
			"Ljava/lang/Class;", Datum.firm, Term.op_atomic);
	}

	// Return default value for specified data type
	protected String getDefaultValue(String desc)
	{
		int ch = desc.charAt(0);
		return ch=='V' ? ""
			: ch=='Z' ? "false"
			: ch=='L' ? "null" : "0";
	}

	// Return reference expression for the specified instance ref,
	// handling string concat and null-ref special cases
	// cpi = pool entry of member used by curr inst
	// suffix = stuff after ref (eg. (args) for a a method) if any
	// desc = its ref descriptor
	// opc = opcode of invoke inst
	// Notes:
	//	Context is that curr inst's ref object is top of stack.
	private Datum getInstanceRef(CpInfo cpi, String suffix, String desc, int opc)
	{
		// Get member used and pop/normalize the ref object in context of
		// that the member belongs to
		String refexpr = cpi.toRef();	// Rets mem name or new classname
		String cldesc = "L" + cpi.toInternalClassName() + ";";
		Datum ptr = doCondExpr(pop(), cldesc);

		// Chk consis of ptr and refine its type if it is soft local
		if (ptr.src.length()==0)
			ptr = new Datum("NO_OBJECT_REF", "Lnone;");
										// Bad stack. Force compile err
		checkSoftRef(ptr, new Datum("", cldesc));

		// If local was set to null, referenced field could be used
		// to set its type. However it seems useful to let compiler
		// report error rather than have user code get null ptr 
		// exception at runtime.
		//checkSoftLocal(ptr, cldesc);

		// If constructor, set the data type to the class name (ie.
		// the string after "new ") and imbed in L;
		if (refexpr.startsWith("new "))
		{
			TypeRef tref = ((MethodRef)cpi).getTypeRef();
			desc = cldesc;

			// this.new xxx() is call to constructor from within constructor.
			// If xxx is meth's class , gen this(), else gen super()
			if (ptr.src.equals("this"))
			{
				if (cpi.toClassName().equals(cf.toName()))
					refexpr = "this";
				else refexpr = "super";
			}

			// If inner class constructor, transform name to src syntax.
			// Note: could special case outer-this via Method.refize(suffix)
			// to cause: new a$b(outer[, args]) => outer.new b([args])
			else if (tref.inner!=null)
				refexpr = "new " + tref.inner.toName();
			else ;						// Normal new xxx()
		}

		// Start member ref with "ptr." unless just name is ok
		// If ptr object is "this", dont need to make it explicit
		// unless field name hidden by local name
		else refexpr = doRefExpr(ptr, refexpr, cpi, opc);

		// Build obj ref, and combine with suffix (eg. method args).
		// Handle special case where objref was null (can happen with
		// compiler that sloppily in-lines method passed an arg = null)
		// by supplying a constant val of ret type instead of obj ref.
		Datum refdtm = new Datum(refexpr + suffix, desc);
		if (ptr.src.equals("null"))
			refdtm.src = getDefaultValue(desc) + " /*" + refdtm.src + "*/";

		// If have string building expr (eg. a = b + "1"), which
		// looks like: new StringBuffer.append(b).append("1").toString()
		// just push(b + "1").
		else if (refexpr.startsWith("new StringBuffer")
				&& refexpr.endsWith("toString"))
			return concatop(refexpr);
		return refdtm;
	}

	// Return datum for this local var
	// idx = index of local to get
	// Notes:
	//	Since defs are created as seen, 1st ref should precede
	//	1st def in only 2 cases:
	//	1) For an arg. But arg-locals exist by now because they are
	//	created when method read (because position in args-desc is the
	//	only sure-fire way to determine the desc for a partic arg).
	//	2) Local defined in loop-bottom of for/while. This occurs because
	//	at source-level this def is at top of loop
	//
	//  A local-idx that is 1st reffed after it is def-ed can cause
	//	meth.getLocal to return null in 2 cases also:
	//	1) Called here for the ref-stmt before buildLvt called (which
	//	processes stmts sequentially in order to build the lvt)
	//  2) This 1st ref is past the end of the local's scope.
	//	This usually means there is def of this local-idx at outer level,
	//	and that it and the inner def need to be merged.
	protected Datum getLocalRef(int idx)
	{
		Local loc = meth.getLocal(idx, inst);

		// If no local & building lvt, create one or match another local.
		// If not building lvt, create one -- that is, this is temp local.
		Statement stin = inst.parent;
		if (loc==null)
		{
			if (meth.lvRead==-1)
			{						// Creating lvt and not before buildLvt
				while ((loc=meth.getPrevLocal(idx, loc)) != null) {
					if (loc.defStmt==null)
						continue;			// Chks moot -- other not inited
					int istart = stin.start;
					int locend = loc.startPc + loc.length;

					// Chk if "extend inner scope case" applies
					if (istart>=locend)
					{
						Statement eob = loc.defStmt.getCommonEob(stin.getEob());
						loc.defStmt = eob;
						loc.length = eob.start + eob.insts.length - loc.startPc;
						break;
					}

					// Chk if "case #2 of ref before def" applies
					if (loc.defStmt.start<=istart && locend>=istart)
						break;
				}
			}
			if (loc==null)
				return new Datum(meth.toLocal(inst, idx), inst.toDesc("U"));
		}

		// Create datum for this ref and indic it is assoc with a local
		Datum dtm = new Datum(loc);
		return dtm;
	}

	// Return ref to a member out of this instruction. If it is instance
	// ref, prepend the object instance to the member name.
	private Datum getMemberRef(Instruction inst, String suffix)
	{
		// Get to pool entry being referenced.
		CpInfo cpi = cf.constant(inst.readUnsignedShort(inst.start+1));
		String desc = cpi.toDesc();

		// Way to process member depends on inst that reffed it
		// 1) Static name
		switch (inst.opcode) {
		case Code.invokestatic:
		case Code.getstatic:
		case Code.putstatic:
			return new Datum(cpi.toStaticRef() + suffix, desc);

		// 2) Instance ref or constructor
		case Code.invokespecial:
		case Code.invokeinterface:
		case Code.invokevirtual:
		case Code.getfield:
		case Code.putfield:
			return getInstanceRef(cpi, suffix, desc, inst.opcode);
		}
		throw new TrekException("Instruction does not contain a member ref");
	}

	// Perform an instruction of form: if (stack1 opr stack2).
	// relop = relational operator to perform
	private void ifbinop(int relop)
	{
		Datum val2 = pop();
		Datum val1 = pop();
		ifop(val1, relop, val2);
	}

	// Perform an instruction of form: if (stack1 opr 0).
	// relop = relational operator to perform
	private void if0op(int relop)
	{
		Datum val = pop();

		// If prev instruction was a -1,0,1 comparitor, split opd apart to
		// put relop within it. Else just compare with 0.
		if (val.opr==Term.op_comp)
		{
			String opd = val.src;
			int residue = opd.indexOf(" @ ");
			Datum val1 = new Datum(opd.substring(0, residue), val.desc);
			Datum val2 = new Datum(opd.substring(residue+3), val.desc);
			ifop(val1, relop, val2);
		}
		else {
			Datum opd2 = new Datum("0", "I", Datum.softy, Term.op_iconst);
			ifop(val, relop, opd2);
		}
	}

	// Merge specified boolean expr with if-clause being built
	// opd1/relop/opd2 = operands and operator of If.
	// inst is if-inst that ends the current term
	private void ifop(Datum opd1, int relop, Datum opd2)
	{
		int condinfo[] = new int[2];

		// If's dont constrain data types except that term cant mix
		// booleans and non-booleans, and booleans can only be == or !=
		opd1 = constrainOperand1(opd1, relop, opd2);
		opd2 = constrainOperand1(opd2, relop, opd1);
		Term term = new Term(opd1, relop, opd2);

		// Init (nested) test-clause if new one
		int ifend = term.findIfEnd(this, top, condinfo);
		if (ifLevel<0 || ifend!=ifEnd[ifLevel])
		{
			ifEnd[++ifLevel] = ifend;
			ifExpr[ifLevel] = new Vector(10, 10);

			// Also setup end-cond-expr info if starting cond expr
			if (condEnd[ifLevel]==0 && condinfo[0]!=0)
			{
				condEnd[ifLevel] = condinfo[0];
				condMark[ifLevel] = condinfo[1];
			}
		}
		Vector ifexpr = ifExpr[ifLevel];

		// Store term in curr test clause, and ret if not at last term
		term.store(inst, ifexpr, ifend);
		if (inst.parent.start+inst.start+inst.size < ifend)
			return;					// Still in middle of if-clause

		// If at end of if-clause, gen source for if-clause and setup opd ret
		// info. (If mult terms, either _or or _and would do)
		Datum ifdtm = booleanDatum.copy();
		ifdtm.opr = ifexpr.size()>1 ? Term.op_or : term.opr;
		if (condEnd[ifLevel]==0)
		{
			ifdtm.src = term.toIfResult(this, ifexpr, ifend, true);
			lopd = term.opd1.src;
			opr = ifdtm.opr;
			ropd = term.opd2.src;
			results.addElement(ifdtm);
		}

		// If condit expr, push until have all 3 outermost terms.
		// a = (b relop c) "ought" to generate b relop c ? 1 : 0. Most
		// compilers generate the logically equiv c !relop b ? 0 : 1 tho.
		// So delay src code gen and do heuristic in doCondExpr...
		// when left and right are 0 and 1, use inverted form.
		else {
			ifdtm.condExpr = ifexpr;
			ifdtm.intVal = ifend;
			push(ifdtm);
		}

		// End test or sub-test
		ifExpr[ifLevel] = null;
		ifEnd[ifLevel--] = 0;
	}

	// Convert iinc to assignment. Note that local must be int because
	// it is possible the store part of the insert would overflow a
	// non-int if it was at its limit (eg. 2**16).
	// lv = local var to modify
	// delta = amt to modify it
	private void iincop(Datum lv, int delta)
	{
		checkSoftResult(lv, new Datum("", "I"));

		// Dont output "name += -d" if delta is negative.
		String opr;
		if (delta >= 0)
			opr = " + ";
		else {
			opr = " - ";
			delta = -delta;
		}

		// Do the assign
		Datum dest = new Datum(lv.src, "I");
		lv.src += opr + delta;
		assop(dest, lv);
	}

	// Init stmt execution context
	protected void init(Statement st)
	{
		// Init execution context
		top = -1;					// Stack empty
		meth = st.parent;
		cf = meth.parent;
		initLoan();
	}

	// Re-init stmt after loan done
	private void initLoan()
	{
		results = new Vector(5);	// No results yet
		didAssign = false;
		ifLevel = -1;
		condEnd = new int[6];		// Alloc and clear condend
	}

	// Returns a datum for the expression, enclosed in parens if approp
	// opd1 = left operand
	// opr = operator
	// opd2 = right operand
	// desc = data type of result value
	// soft = softness of result value
	protected Datum isolate(Datum opd1, int opr, Datum opd2,
		String desc, int soft)
	{
		// Build new expr
		String src1 = isolate(opd1.src, opd1.opr, opr);
		String src2 = isolate(opd2.src, opd2.opr, opr);
		String srcopr = Term.oprs[opr];
		return new Datum(src1 + srcopr + src2, desc, soft, opr);
	}
	private Datum isolate(int opr, Datum opd1)
	{
		String src1 = isolate(opd1.src, opd1.opr, opr);
		String srcopr = Term.oprs[opr];
		Datum expr = new Datum(srcopr + src1, opd1.desc, opd1.soft, opr);
		expr.loc = opd1.loc;		// Local if any still assoc with result
		return expr;
	}

	// Parthesize opd if operator within it has lower precedence than
	// the opr being applied to the opd
	// opd = source of operand (var or expr)
	// subopr = the top-level operator within the operand
	// opr = the opr being applied to the operand
	String isolate(String opd, int subopr, int opr)
	{
		if (Term.precs[subopr] < Term.precs[opr])
			return "(" + opd + ")";	// Parenthesize compound expr
		else return opd;
	}

	// Returns normalized form of dtm.src, modifying dtm as necessary to
	// achieve the normalization of a constant
	// dtm = datum containing value to chk
	// desc = descriptor to use in chking dtm.src
	// context = diff contexts have diff requirements (mc_*)
	protected String mapConstant(Datum dtm, String desc)
	{
		return mapConstant(dtm, desc, 0);
	}
	private String mapConstant(Datum dtm, String desc, int context)
	{
		// Array constants allowed as of 1.1. But dont confuse with regular
		// array init (ie. Xxx xxx[] = {...}), in which mc_array off.
		if ((context&mc_array)!=0
			&& dtm.src.length()>0 && dtm.src.charAt(0)=='{')
		{
			dtm.src = "new " + cf.toDef(null, desc, 0) + dtm.src;
			dtm.soft = Datum.firm;
			return dtm.src;
		}

		// Only things to directly map are soft integer constants
		if (dtm.opr!=Term.op_iconst || dtm.soft<=Datum.firm)
			return dtm.src;

		// For bytes and shorts, need cast in certain contexts (eg. args)
		boolean alltypes = (context & mc_cast) != 0;
		switch (desc.charAt(0)) {
		case 'B':
			if (!alltypes) return dtm.src;
			dtm.src = "(byte)" + dtm.src;
			break;
		case 'S':
			if (!alltypes) return dtm.src;
			dtm.src = "(short)" + dtm.src;
			break;

		// Map to boolean fmt
		case 'Z':
			if (dtm.intVal==0)
				dtm.src = "false";
			else dtm.src = "true";
			break;

		// Map to char constant if ascii char
		case 'C':
			char ch = (char)dtm.intVal;
			if (ch<' ' || ch>126)
				return dtm.src;
			String quote = ch=='\'' || ch=='\\' ? "\\" : "";
			dtm.src = "'" + quote + new Character(ch).toString() + "'";
			break;

		default:
			return dtm.src;
		}

		// Fall thru if did a transform, and set firm so dont do again
		dtm.soft = Datum.firm;
		return dtm.src;
	}

	// Push expression to create new array (curr inst is new array inst)
	// dims = # of dims to alloc now (or 0 for newarray inst)
	// multi = true says multianewarry
	// Top of stack is size of each dimension being alloc now
	protected void newArray(int dims, boolean multi)
	{
		// Do simple newarray case
		if (dims==0)
		{
			int dt = inst.readUnsignedByte(inst.start+1);
			String desc = "[" + descs[dt];
			push("new " + datatypes[dt]
				+ "[" + popUse("I", Term.op_absent) + "]", desc);
			return;
		}

		// Output new and name of array.
		String datatype = getDataType(inst);
		int dtlen = datatype.length();
		int brkidx = datatype.indexOf('[');
		if (brkidx<0) brkidx = dtlen;
		String newarr = "new " + datatype.substring(0, brkidx);

		// Build dimensions last to first, replacing leading []'s with
		// size of each allocated dims.
		int unallocidx = multi ? brkidx + dims*2 : brkidx;
		String dimstr = datatype.substring(unallocidx);
		for (int j=0; j<dims; j++) {
			dimstr = "[" + popUse("I", Term.op_absent) + "]" + dimstr;
		}
		String desc = getClassDesc(inst);
		if (!multi) desc = '[' + desc;
		push(newarr + dimstr, desc);
	}

	// Return top of stack
	private Datum peek()
	{
		Datum dtm;
		if (top < 0)
			dtm = missingDatum;
		else dtm = (Datum)stack.elementAt(top);
		return dtm;
	}

	// Remove and return top of stack. If none, return null string.
	// Notes:
	//	Because of dup* insts, a datum must not be "transformed" into another
	//	object after it has been popped.
	protected Datum pop()
	{
		Datum dtm;
		if (top < 0)
			dtm = missingDatum;
		else dtm = (Datum)stack.elementAt(top--);
		return dtm;
	}

	// Return source of stack item, insuring it's expanded
	// desc = context to expand cond expr in
	// opr = operator context in which the item is used
	private String popUse(String desc, int opr)
	{
		Datum dtm = doCondExpr(pop(), desc);
		return isolate(dtm.src, dtm.opr, opr);
	}

	// Reflect fact that datum used in integer context
	// dtm = datum used in integer context
	private void preventBoolean(Datum dtm)
	{
		dtm.refineSoftness(Datum.byplus);
	}

	// Merge result string with datum and save appropriately
	private void pushAssign(String str, Datum val)
	{
		val.src = str;
		val.opr = Term.op_assign;
		if (val.dest==null)
			push(val);					// Replace popped-off c-left-on-stack
		else results.addElement(val);	// Save "stackless" result
	}

	// Push an integer constant
	// val = its val as integer
	// strval = its val as source string (so could be symbolic)
	private void push(int val) { push(val, "" + val); }
	private void push(int val, String strval)
	{
		// Decide how constant's value constrains data type?
		int soft;
		if (val==0 || val==1)
			soft = Datum.softy;
		else if (val>=-127 && val<=127)
			soft = Datum.byplus;
		else if (val>=0 && val<=1<<16-1)
			soft = Datum.chplus;
		else if (val>=-(1<<15-1) && val<=1<<15-1)
			soft = Datum.shplus;
		else soft = Datum.firm;

		// Build and push the datum form this constant
		Datum dtm = new Datum(strval, "I", soft, Term.op_iconst);
		dtm.intVal = val;
		push(dtm);
	}

	// Push a value on operand stack. In context of building statement string,
	// this means concatenating pieces of operation and just storing whole
	// string on the stack.
	private void push(String str, String desc, int sof)
	{
		push(new Datum(str, desc, sof));
	}
	private void push(String str, String desc)
	{
		push(new Datum(str, desc));
	}
	protected void push(Datum val)
	{
		// Processed pushed pieces of a?b:c if about to push the c
		// Inst null for push of loaned temp local for inlined if
		if (inst!=null
			&& condEnd[ifLevel+1]==inst.parent.start+inst.start+inst.size)
			val = doCondExpr(val);

		// Chk for re-load special case. If so, dont push val.
		// Instead prep to put orig assign expr on top of stack
		else if (didAssign)
			val = checkReload(val);

		// Insert the value at top of stack
		top += 1;
		stack.setSize(top+1);
		stack.setElementAt(val, top);
	}

	// Replace 1 char in src (at pos) with specified string
	private String replace(String src, int pos, String with)
	{
		return src.substring(0, pos) + with + src.substring(pos+1);
	}

	// Debugging routine to dump operand stack
	protected void showStack()
	{
		for (int i=0; i<=top; i++) {
			Datum dtm = (Datum)stack.elementAt(i);
			int start = inst.parent.start + inst.start;
			System.out.println(start + "(" + i + "): " + dtm.src);
		}
		System.out.println(inst.toString());
		System.out.println("-----------");
	}


	// Process xastore inst. That is, do an assign to an array element
	// val = datum of val to store into specified array
	// idx = index to store it al
	// arr = datum of array to store it into
	protected void storeElement(Datum val, Datum idx, Datum arr)
	{
		// When analyzing data types, it is possible that array info
		// is not avail eg. because stmt loaning is not active
		arr = doCondExpr(arr, "X");
		idx = doCondExpr(idx, "I");
		String dest = isolate(arr.src, arr.opr, Term.op_atomic)
			+ "[" + idx.src + "]";
		String desc = arr.desc.length()>1 ? arr.desc.substring(1) : "X";
		arr = new Datum(dest, desc);
		assop(arr, val);
	}

	// Create src-level cast using this descriptor
	private String toCast(String desc)
	{
		return "(" + cf.toDef(null, desc, 0) + ")";
	}

	// Process current instruction
	boolean xinst()
	{
		Statement stmt = inst.parent;
		if ((Trek.debug&Trek.Db_atsmall) != 0)
			Trek.println(System.out,
				stmt.start + "(" + inst.start + "): " + inst.toString());
		try {
			Datum val1, val2, val3;
			CpInfo cpi;
			switch (inst.opcode) {
			case Code.nop:
				break;

			// Loading a constant onto operand stack
			case Code.aconst_null:
				push(nullDatum.copy());
				break;
			case Code.iconst_m1:
				push(-1, "-1");
				break;
			case Code.iconst_0:
				push(0, "0");
				break;
			case Code.iconst_1:
				push(1, "1");
				break;
			case Code.iconst_2:
				push(2, "2");
				break;
			case Code.iconst_3:
				push(3, "3");
				break;
			case Code.iconst_4:
				push(4, "4");
				break;
			case Code.iconst_5:
				push(5, "5");
				break;
			case Code.lconst_0:
				push("0", "J");
				break;
			case Code.lconst_1:
				push("1", "J");
				break;
			case Code.fconst_0:
				push("0", "F");
				break;
			case Code.fconst_1:
				push("1", "F");
				break;
			case Code.fconst_2:
				push("2", "F");
				break;
			case Code.dconst_0:
				push("0", "D");
				break;
			case Code.dconst_1:
				push("1", "D");
				break;

			// Loading a non-hard wired constant. First 2 are integers,
			// next two may be integers.
			case Code.bipush:
				push(inst.readByte(inst.start+1));
				break;
			case Code.sipush:
				push(inst.readShort(inst.start+1));
				break;
			case Code.ldc:
			case Code.ldc_w:
			case Code.ldc2_w:
				cpi = cf.constant(inst.getCpIdx());

				// Pool constants can have special case src form,
				// so let cpi.toString() provide that. Setup extra
				// stuff for integer constants if applic
				String valstr = cpi.toString();
				if (cpi instanceof IntegerConstant)
				{
					IntegerConstant ic = (IntegerConstant)cpi;
					Integer iobj = (Integer)ic.getObject();
					push(iobj.intValue(), valstr);
				}
				else push(valstr, cpi.toDesc());
				break;

			// Loading a variable
			case Code.iload:
			case Code.lload:
			case Code.fload:
			case Code.dload:
			case Code.aload:
				push(getLocalRef(inst.getLocalIdx()));
				break;
			case Code.iload_0:
			case Code.lload_0:
			case Code.fload_0:
			case Code.dload_0:
			case Code.aload_0:
				push(getLocalRef(0));
				break;
			case Code.iload_1:
			case Code.lload_1:
			case Code.fload_1:
			case Code.dload_1:
			case Code.aload_1:
				push(getLocalRef(1));
				break;
			case Code.iload_2:
			case Code.lload_2:
			case Code.fload_2:
			case Code.dload_2:
			case Code.aload_2:
				push(getLocalRef(2));
				break;
			case Code.iload_3:
			case Code.lload_3:
			case Code.fload_3:
			case Code.dload_3:
			case Code.aload_3:
				push(getLocalRef(3));
				break;

			// Loading array element.
			// Sub a [ from desc because this is elem oper, but dont
			// fail on bad array desc. This can happen when opds are
			// missing because of heavily instrumented code.
			case Code.iaload:
			case Code.laload:
			case Code.faload:
			case Code.daload:
			case Code.aaload:
			case Code.baload:
			case Code.caload:
			case Code.saload:
				val2 = doCondExpr(pop(), "I");
				val1 = doCondExpr(pop(), "X");
				String desc = val1.desc.length()>1 ? val1.desc.substring(1) : "X";
				val3 = new Datum(val1.src + "[" + val2.src + "]", desc);
				push(val3);
				break;

			// Storing a result in a local variable
			case Code.istore:
			case Code.lstore:
			case Code.fstore:
			case Code.dstore:
			case Code.astore:
				assop(inst.getLocalIdx());
				break;
			case Code.istore_0:
			case Code.lstore_0:
			case Code.fstore_0:
			case Code.dstore_0:
			case Code.astore_0:
				assop(0);
				break;
			case Code.istore_1:
			case Code.lstore_1:
			case Code.fstore_1:
			case Code.dstore_1:
			case Code.astore_1:
				assop(1);
				break;
			case Code.istore_2:
			case Code.lstore_2:
			case Code.fstore_2:
			case Code.dstore_2:
			case Code.astore_2:
				assop(2);
				break;
			case Code.istore_3:
			case Code.lstore_3:
			case Code.fstore_3:
			case Code.dstore_3:
			case Code.astore_3:
				assop(3);
				break;

			// Store array element. Since arrays are fully typed, set
			// src dt to dest's, but minus a [ because this is elem oper
			// Pops are respectively value, index, and array
			case Code.iastore:
			case Code.lastore:
			case Code.fastore:
			case Code.dastore:
			case Code.aastore:
			case Code.bastore:
			case Code.castore:
			case Code.sastore:
				storeElement(pop(), pop(), pop());
				break;

			// Miscellaneous stack bookkeeping oprs
			case Code.pop:
			case Code.pop2:			// A double or long
				results.addElement(pop());
				break;

			// Dup2's usually single dup of double/long, but can be dup of 2 opds
			// as in a[b] += 1, where a/b is dup2-ed to do iaload & iastore.
			case Code.dup2:
			case Code.dup2_x1:
			case Code.dup2_x2:
				if (Datum.getDataSize(peek().desc)!=2)
					dup2op(inst.opcode);
				else dupop(inst.opcode);
				break;
			case Code.dup:
			case Code.dup_x1:
			case Code.dup_x2:
				dupop(inst.opcode);
				break;

			// Swap top two stack slots
			case Code.swap:
				val1 = pop();
				val2 = pop();
				push(val1);
				push(val2);
				break;

			// Arithmetic operations
			case Code.iadd:
			case Code.ladd:
			case Code.fadd:
			case Code.dadd:
				binop(Term.op_plus);
				break;
			case Code.isub:
			case Code.lsub:
			case Code.fsub:
			case Code.dsub:
				binop(Term.op_minus);
				break;
			case Code.imul:
			case Code.lmul:
			case Code.fmul:
			case Code.dmul:
				binop(Term.op_mult);
				break;
			case Code.idiv:
			case Code.ldiv:
			case Code.fdiv:
			case Code.ddiv:
				binop(Term.op_div);
				break;
			case Code.irem:
			case Code.lrem:
			case Code.frem:
			case Code.drem:
				binop(Term.op_rem);
				break;
			case Code.ineg:
			case Code.lneg:
			case Code.fneg:
			case Code.dneg:
				val1 = doCondExpr(pop(), "I");
				Datum expr = isolate(Term.op_negate, val1);
				preventBoolean(expr);
				push(expr);
				break;

			// Shifting operations
			case Code.ishl:
			case Code.lshl:
				binop(Term.op_lshift);
				break;
			case Code.ishr:
			case Code.lshr:
				binop(Term.op_rshift);
				break;
			case Code.iushr:
			case Code.lushr:
				binop(Term.op_rushift);
				break;

			// Bitwise boolean oprs
			case Code.iand:
			case Code.land:
				binop(Term.op_and);
				break;
			case Code.ior:
			case Code.lor:
				binop(Term.op_or);
				break;
			case Code.ixor:
			case Code.lxor:
				binop(Term.op_xor);
				break;

			case Code.iinc:
				val1 = getLocalRef(inst.getLocalIdx());
				int delta = inst.getIincDelta();
				iincop(val1, delta);
				break;

			// Casting. Note that source of integer can still
			// mean source is byte, char, etc.
			case Code.i2l:
				push(popUse("X", Term.op_cast), "J");
				break;
			case Code.i2f:
				castop("(float)", "F", new Datum("", "I", Datum.byplus));
				break;
			case Code.i2d:
				castop("(double)", "D", new Datum("", "I", Datum.byplus));
				break;
			case Code.l2i:
				castop("(int)", "I", longDatum);
				break;
			case Code.l2f:
				castop("(float)", "F", longDatum);
				break;
			case Code.l2d:
				castop("(double)", "D", longDatum);
				break;
			case Code.f2i:
				castop("(int)", "I", floatDatum);
				break;
			case Code.f2l:
				castop("(long)", "J", floatDatum);
				break;
			case Code.f2d:
				push(popUse("X", Term.op_cast), "D");
				break;
			case Code.d2i:
				castop("(int)", "I", doubleDatum);
				break;
			case Code.d2l:
				castop("(long)", "J", doubleDatum);
				break;
			case Code.d2f:
				castop("(float)", "F", doubleDatum);
				break;
			case Code.int2byte:
				castop("(byte)", "B", new Datum("", "I", Datum.chplus));
				break;
			case Code.int2char:
				castop("(char)", "C", new Datum("", "I", Datum.shplus));
				break;
			case Code.int2short:
				castop("(short)", "S", new Datum("", "I"));
				break;

			// Bunch of if-related instructions. 
			case Code.lcmp:
			case Code.fcmpl:
			case Code.fcmpg:
			case Code.dcmpl:
			case Code.dcmpg:
				binop(Term.op_comp);
				break;

			// 0-based ifs
			case Code.ifeq:
				if0op(Trek.DOIF_EQ);
				break;
			case Code.ifne:
				if0op(Trek.DOIF_NE);
				break;
			case Code.iflt:
				if0op(Trek.DOIF_LT);
				break;
			case Code.ifge:
				if0op(Trek.DOIF_GE);
				break;
			case Code.ifgt:
				if0op(Trek.DOIF_GT);
				break;
			case Code.ifle:
				if0op(Trek.DOIF_LE);
				break;

			// Full a op b if's
			case Code.if_icmpeq:
				ifbinop(Trek.DOIF_EQ);
				break;
			case Code.if_icmpne:
				ifbinop(Trek.DOIF_NE);
				break;
			case Code.if_icmplt:
				ifbinop(Trek.DOIF_LT);
				break;
			case Code.if_icmpge:
				ifbinop(Trek.DOIF_GE);
				break;
			case Code.if_icmpgt:
				ifbinop(Trek.DOIF_GT);
				break;
			case Code.if_icmple:
				ifbinop(Trek.DOIF_LE);
				break;
			case Code.if_acmpeq:
				ifbinop(Trek.DOIF_EQ);
				break;
			case Code.if_acmpne:
				ifbinop(Trek.DOIF_NE);
				break;

			// Various control actions
			case Code.goto_:
			case Code.jsr:
			case Code.ret:
				break;
			case Code.tableswitch:
			case Code.lookupswitch:
				val1 = doCondExpr(pop(), "I");
				preventBoolean(val1);
				results.addElement(new Datum("switch (" + val1.src + ") {", ""));
				break;

			// Return a value. Integer constant at jvm-level can be 
			// true/false or char constant at source level
			case Code.ireturn:
			case Code.lreturn:
			case Code.freturn:
			case Code.dreturn:
			case Code.areturn:
				String retdesc = meth.toDesc();
				val1 = doCondExpr(pop(), retdesc);
				checkSoftRef(val1, new Datum("", retdesc));
				results.addElement(
					new Datum("return " + mapConstant(val1, retdesc, mc_array), ""));
				break;
			case Code.return_:
				results.addElement(new Datum("return", ""));
				break;

			// Data members. If integer source is specifically int,
			// then that does clarify type of dest.
			case Code.getstatic:
			case Code.getfield:
				push(getMemberRef(inst, ""));
				break;

			// Store into a field
			case Code.putstatic:
			case Code.putfield:
				val1 = pop();
				val2 = getMemberRef(inst, "");
				assop(val2, val1);
				break;

			// Calls
			case Code.invokestatic:
			case Code.invokespecial:
			case Code.invokeinterface:
			case Code.invokevirtual:
				callop(inst);
				break;
			case Code.Unused1:
				break;

			// New objects
			case Code.new_:
				push("new", "X");	// Really just surrogate of <init> call
				break;				// Will be followed by invokespecial
			case Code.newarray:
				newArray(0, false);
				break; 
			case Code.anewarray:
				newArray(1, false);
				break; 

			// Miscellaneous. Athrows for end-sync can be tucked in
			// after goto's and return's
			case Code.arraylength:
				push(popUse("X", Term.op_atomic) + ".length", "I");
				break;
			case Code.athrow:
				if (results.size()==0)
					results.addElement
						(new Datum("throw " + popUse("X", Term.op_absent), ""));
				break;

			// checkcast is source-level cast (which throws exception
			// if cast fails at runtime). Instanceof is test that rets
			// 1 or 0. So it is followed by an if0op.
			// If casting/chking array, then pool entry is descriptor
			// for array rather than a simple class name
			case Code.checkcast:
				String datatype = getDataType(inst);
				castop("(" + datatype + ")",
					getClassDesc(inst),
					new Datum("", "X", Datum.softy));
				break;
			case Code.instanceof_:
				val1 = new Datum(getDataType(inst), "X");
				val2 = doCondExpr(pop(), "X");
				push(isolate(val2, Term.op_instance, val1, "Z", Datum.firm));
				checkNoopCast(val2, "L" + val1.src + ";");
				break;

			// Get data type and number of dimensions
			case Code.multianewarray:
				int dims = inst.readUnsignedByte(inst.start+3);
				newArray(dims, true);
				break;

			// If-ptr insts
			case Code.ifnull:
				ifop(doCondExpr(pop(), "X"), Trek.DOIF_EQ, nullDatum.copy());
				break;
			case Code.ifnonnull:
				ifop(doCondExpr(pop(), "X"), Trek.DOIF_NE, nullDatum.copy());
				break;

			// Wide jumping
			case Code.goto_w:
			case Code.jsr_w:
				break;

			// Output sync stmt, eliminating dummy local used in sync
			case Code.monitorenter:
				String sync = popUse("X", Term.op_absent);
				int eqidx = sync.indexOf(" = ");
				if (eqidx > 0)
					sync = sync.substring(eqidx+3);
				results.addElement(new Datum("synchronized (" + sync + ") {", ""));
				break;

			// Miscellaneous.
			case Code.monitorexit:
				break;
			default:
				break;
			}
			return true;
		}

		// Isolate error to one stmt's execution
		catch (Exception e) {
			int stend = stmt.start + stmt.insts.length;
			String action = this instanceof Lvt ? "analyzing " : "executing ";
			String msg = "C=" + cf.toQualifiedName()
				+ ";D=" + action + "stmt at;M=" +  meth.toName()
				+ ";E=PC" + stmt.start + "-" + stend + ";";
			cf.logError(msg, e);
			return false;
		}
	}								// End xinst

	public static boolean z(int a, int b, int c) {
		System.out.println(a + ":" + b + ":" + c); return true;}
	public static boolean z(String a, int b, int c) {
		System.out.println(a + ":" + b + ":" + c); return true;}
	public static boolean z(int a, int b) {
		System.out.println(a + ":" + b); return true;}
	public static boolean z(String a, int b) {
		System.out.println(a + ":" + b); return true;}

}

