//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

import dec.trek.*;
import java.io.PrintStream;
import java.util.Vector;

class find extends Trek
{
	public int matches;				// # of matches
	protected boolean byDef;		// True to order report sorted by name defs
									// False to order report by classfile scan
	private ClassFile cFile;		// Curr classfile for findAdjacent
	private Vector classVec;		// Classfiles to consider
	private String classString;		// Classfiles to consider as ;-sep string
	private boolean datalog;
	private boolean didLine;		// True says 1+ top-level things output
	private boolean exact;			// True says do exact match
	private Vector findIt;			// Search terms
	private String findString;		// Search terms as semi-separated string
	private Object object;			// Last matched object for findAdjacent
	private FindLog log;			// Identies report file
	private int preWin;				// # of lines before a found ref to print
	private int postWin;			// # of lines after, incl ref line itself
	private int winSize=1;

	// Prepare to mark windows around refs
	private final static String defPrefix = "DEF= ";
//	private final static String defPrefix = "++++ ";
	private final static Integer near = new Integer(0);
	private final static Integer ref = new Integer(1);
	private final static Integer localDef = new Integer(2);

	// Decide if processing this class
	protected void atStartOf(ClassFile cf)
	{
		// cf not in classVec, skip this class
		if (classVec!=null && Jtu.isMatchFor(classVec, false, cf)<0)
			endTrek(1);
	}

	// At a class that refs mem.
	// Scan methods of class, printing each ref to mem
	protected void atReference(ClassFile cf, Member mem)
	{
		// Scan code of this class for refs & print ref lines (and friends)
		for (Method meth=cf.firstMethod(); meth!=null; meth=meth.next()) {
			printRefsInMethod(meth, mem, null);
		}
	}								// At end of atReference

	// Process each ref seen as go thru classfiles
	protected void at(Instruction inst)
	{
		if (byDef) return;
		if (Jtu.isMatchFor(findIt, exact, inst) < 0)
			return;						// No match

		// A match, mark it (and chk if local def)
		markRef(inst.getStatement(), inst.getLocal());
	}

	// Process each field and method seen in main trek
	protected void at(Field field)
	{
		processMemberDef(field);
	}
	protected void atStartOf(Method meth)
	{
		processMemberDef(meth);
		if (!byDef) return;
		if (Jtu.hasMatch(findIt, exact, meth, null) < 0) return;

		// Since sorting by defs, need context with def since it is
		// not visually within a class and method.
		// Display defs of matching locals -- and refs if user asked for them.
		String classpfx = meth.getClassFile().toName() + ": ";
		String inmeth = " [in " + meth.toName() + meth.toArgList() + "]";
		boolean didline = false;
		for (Local loc=meth.firstLocal(); loc!=null; loc=loc.next()) {
			if (Jtu.hasMatch(findIt, exact, meth, loc) < 0) continue;

			// Output def and possibly usage lines
			if (didline && postWin>0) log.writeEntry();
			if (datalog)
				log.writeLocal(loc.toObjectId());
			else printDef(classpfx + loc.toString() + inmeth);
			didline = true;
			if (postWin > 0)
				printRefsInMethod(meth, null, loc);
		}
	}

	// Process dcl-only locals
	protected void atStartOf(Statement st)
	{
		Local loc = matchesLocalDef(st);
		if (loc!=null)
			markRef(st, loc);
	}

	// Process each ref seen as go thru classfiles
	protected void atEndOf(ClassFile cf)
	{
		// If any defs or refs, put out hdr & scan the classfile
		// Separate refs in each classfile with blank line
		if (byDef || cf.userDefined==null)
			return;
		if (didLine)
			log.writeEntry();
		didLine = true;
		if (datalog)
			log.writeClass(cf.toObjectId());
		else log.writeEntry(cf.toQualifiedName() + ":");

		// If any fields have defs, output them
		for (Field fld=cf.firstField(); fld!=null; fld=fld.next()) {
			if (fld.userDefined==null) continue;
			if (datalog)
				log.writeField(fld.toObjectId());
			else printDef(defPrefix + fld.toString());
			fld.userDefined = null;
		}

		// If any methods have refs, put out hdr & scan the classfile
		for (Method meth=cf.firstMethod(); meth!=null; meth=meth.next()) {
			if (meth.userDefined==null) continue;
			printWindows(meth, 0);
		}
		cf.userDefined = null;
	}

	// Do the scan to find the desired stuff
	// findit = string to look for
	// classes = classes to look in
	// outspec = file to store results
	// exact = true if doing exact match
	// winsize = lines to display around each usage (0 says defs-only)
	public void doFind(String findit, String classes, String outspec,
		boolean datalog, boolean exact, int winsize)
	{
		if (findit==null || findit.length()==0)
			throw new TrekException("No search string specified");

		// Set general state vars
		matches = 0;
		findIt = separate(findit);
		classVec = separate(classes);
		this.exact = exact;
		this.datalog = datalog;

		// Set window/scan stuff
		// If only doing defs (-win 0), processing effectively by def
		byDef = false;
		preWin = winsize/2;
		postWin = winsize - preWin;	// So postWin>preWin when winsize odd
		if (postWin==0)
			byDef = true;

		// Open find log and do the work
		try {
			if (datalog)
				log = new FindLog(outspec, RunLog.LOG_DATA);
			else log = new FindLog(outspec, RunLog.LOG_TEXT);
		}
		catch (RuntimeException e) {
			throw new RuntimeException("Error opening "
				+ outspec + ": " + e.toString());
		}
		doTrek();
		log.closeLog();
	}

	// Returns adjacent source-level match, or null
	// findit = string to look for
	// classes = classes to look in
	// exact = true says do case-sensitive match
	// dir = true is forward, false is backward
	private Object doFindAdjacent(String findit, String classes,
		boolean exact, boolean dir)
	{
		boolean resuming;
		Method obj_meth = null;
		Statement obj_stmt = null;

		// Set class/etc to resume with if same search
		if (cFile!=null
			&& findit.equals(findString)
			&& classes.equals(classString)
			&& exact==this.exact)
		{
			resuming = true;
			if (object instanceof Method)
				obj_meth = (Method)object;
			else if (object instanceof Statement)
			{
				obj_stmt = (Statement)object;
				obj_meth = obj_stmt.getMethod();
			}
			else if (object instanceof Instruction)
			{
				obj_stmt = ((Instruction)object).getStatement();
				obj_meth = obj_stmt.getMethod();
			}
		}

		// Not resuming. Init search info and 1st class
		else {
			matches = 0;
			resuming = false;
			findString = findit;
			findIt = separate(findit);
			classString = classes;
			classVec = separate(classes);
			this.exact = exact;
			cFile = dir ? firstClassFile() : lastClassFile();
		}

		// Do or resume scan til next match
		for (; cFile!=null; cFile=dir?cFile.next():cFile.prev()) {
			if (classVec!=null && Jtu.isMatchFor(classVec, false, cFile)<0)
				continue;

			// Chk its fields
			Field fld =  dir ? cFile.firstField() : cFile.lastField();
			for (; fld!=null; fld=dir?fld.next():fld.prev()) {
				if (resuming)
				{
					if (!(object instanceof Field)) break;
					resuming = false;
					fld = (Field)object;
					continue;
				}
				if (Jtu.isMatchFor(findIt, exact, fld)>=0)
				{
					object = fld;
					matches += 1;
					return object;
				}
			}

			// Scan current classfile's methods
			Method meth = dir ? cFile.firstMethod() : cFile.lastMethod();
			for (; meth!=null; meth=dir?meth.next():meth.prev()) {
				if (resuming)
				{
					meth = obj_meth;
					if (object instanceof Method) resuming = false;
				}
				if (meth!=obj_meth && Jtu.isMatchFor(findIt, exact, meth)>=0)
				{
					object = meth;
					matches += 1;
					return object;
				}

				// Trek thru method's statements
				Statement stmt = dir ? meth.firstStatement() : meth.lastStatement();
				for (; stmt!=null; stmt=dir?stmt.next():stmt.prev()) {
					if (resuming)
					{
						stmt = obj_stmt;
						if (object instanceof Statement) resuming = false;
					}
					if (stmt!=obj_stmt && matchesLocalDef(stmt)!=null)
					{
						object = stmt;
						matches += 1;
						return object;
					}

					// Look for inst that refs a name that matches a term
					// in findIt. But avoid matching 
					// same stmt twice by using Resuming and object
					Instruction inst = stmt.firstInstruction();
					for (; inst!=null; inst=inst.next()) {
						if (resuming)
						{
							inst = (Instruction)object;
							resuming = false;
							break;
						}
						if (Jtu.isMatchFor(findIt, exact, inst)>=0)
						{
							object = inst;
							matches += 1;
							return inst.getStatement();
						}
					}					// End insts
				}						// End statements
			}							// End methods
		}								// End classfiles
		return null;
	}

	// Returns adjacent match in backward direction, or null
	public Object doFindNext(String findit, String classes, boolean exact)
	{
		return doFindAdjacent(findit, classes, exact, true);
	}
	public Object doFindPrev(String findit, String classes, boolean exact)
	{
		return doFindAdjacent(findit, classes, exact, false);
	}

	// Pick up the (possibly wildcarded) name to find
	public int getCmdLineArg(String argv[], int idx)
	{
		if (argv[idx].startsWith("-"))
		{
			// Process one of the common args if present
			int newi = Jtu.getCmdLineArg(argv, idx);
			if (newi > 0) return newi;		// Ate this arg

			if (argv[idx].equals("-cs"))
			{
				exact = true;
				return idx + 1;
			}
			else if (argv[idx].equals("-bydef"))
			{
				byDef = true;
				return idx + 1;
			}
			else if (argv[idx].equals("-win"))
			{
				if (idx>argv.length-2)
					throw new TrekException("No value specified for -win");
				if (argv.length<idx+2) return -1;
				winSize = Integer.parseInt(argv[idx+1]);
				return idx + 2;
			}
			else return 0;					// Some other arg
		}
		findString = argv[idx];
		return idx + 1;
	}

	public static void main(String argv[])
	{
		find ftrek = new find();
		try {
			ftrek.getCmdLine(argv);
			ftrek.initTrek();
			ftrek.doFind(ftrek.findString, null, Jtu.outSpec,
				false, ftrek.exact, ftrek.winSize);
			ftrek.endTrek();
		}
		catch (Throwable e) {
			System.out.println(ftrek.toErrorMessage(e));
			System.exit(1);
		}
	}

	// Mark the ref and its window for future display
	// st = stmt to mark
	// loc = if specified, chk if this st is def of this local
	private void markRef(Statement st, Local loc)
	{
		// Mark ref. Canon stmt to loop-top if ref in end stmt of for/while
		int type = st.getType();
		if (type==ST_ENDFOR || type==ST_ENDWHILE)
		{
			int looplev = st.getLevel();
			for (st=st.prev(); st.getLevel()!=looplev; st=st.prev());
		}

		// Mark as plain ref or local def (and dont clobber local def with ref)
		if (st.userDefined!=localDef)
			st.userDefined = loc!=null && st==loc.getDefinition() ? localDef : ref;

		// Mark context as having ref (or local def if loc is arg)
		Method meth = st.getMethod();
		if (meth.userDefined==null)
		{
			meth.userDefined = near;	// Something in method has ref
			meth.getClassFile().userDefined = near;
		}
		if (loc!=null && loc.isArg())
			meth.userDefined = ref;

		// Mark window above this ref
		Statement stp, stn;
		int pre = preWin;
		for (stp=st.prev(); stp!=null; stp=stp.prev()) {
			if (pre <= 0) break;
			pre -= 1;
			if (stp.userDefined==null) stp.userDefined = near;
		}

		// Mark window below this ref (incl unused part of preWin)
		int post = postWin - 1 + pre;
		for (stn=st.next(); stn!=null; stn=stn.next()) {
			if (post <= 0) break;
			post -= 1;
			if (stn.userDefined==null) stn.userDefined = near;
		}

		// Use up any unused part of postWin before match
		for (; stp!=null; stp=stp.prev()) {
			if (post-- <= 0) break;
			if (stp.userDefined==null) stp.userDefined = near;
		}
	}

	// Returns the local if st contains a matching local-var def
	// st = stmt to chk
	private Local matchesLocalDef(Statement st)
	{
		if (st.getType()!=ST_LOCAL) return null;

		// Brute force see if this local dcl counts
		Method meth = st.getMethod();
		for (Local loc=meth.firstLocal(); loc!=null; loc=loc.next()) {
			if (st==loc.getDefinition()
				&& Jtu.hasMatch(findIt, exact, meth, loc)>=0)
				return loc;
		}
		return null;
	}

	// Output name def
	// str = text to describe this definition
	private void printDef(String str)
	{
		log.writeEntry(str);
		matches += 1;
	}

	// Print windows around the refs of the specified variable
	// meth = method to process
	// mem/loc = 1 or other identifies variable to look for
	private void printRefsInMethod(Method meth, Member mem, Local loc)
	{
		// Scan code of this method for refs & print ref lines (and friends)
		Statement st;
		Statement st1 = meth.firstStatement();

		// Pass 1: find refs and mark ref and in-window lines
		for (st=st1; st!=null; st=st.next()) {
			if (mem!=null)
			{
				if (st.references(mem)==null)
					continue;
			}

			// Process locals. For code-less dcl, just use brute force
			// Otherwise chk if stmt's insts refs the local
			else if (st.getType()==ST_LOCAL)
			{
				if (st!=loc.getDefinition())
					continue;
			}
			else if (st.references(loc)==null)
					continue;
			markRef(st, loc);
		}

		// Pass 2: print each window, handling overlapping windows
		if (meth.userDefined!=null)
			printWindows(meth, loc!=null ? -1 : 1);
	}								// At end of printRefsInMethod

	// Print windows around each marked reference
	// meth = method to process
	// context = what to display for method context
	private void printWindows(Method meth, int context)
	{
		String mname = meth.toName();

		// If method itself reffed, show full source line.
		if (meth.userDefined==ref)
		{
			if (datalog)
			{
				log.writeMethod(meth.toObjectId());
				matches += 1;
			}
			else printDef(defPrefix + meth.toString());
		}

		// For locals byDef, context is displayed externally.
		// For not byDef, classfile is displayed externally.
		// For members byDef, nothing is known externally
		else if (context>=0)
		{
			String methinfo;
			if (datalog)
				log.writeInMethod(meth.toObjectId());
			else {
				methinfo = mname + meth.toArgList();
				if (context==1)		// By def member?
					methinfo = meth.getClassFile().toName() + "." + methinfo;
				log.writeEntry("**** " + methinfo);
			}
		}

		// Scan the method for its refs
		int inwin = 0;
		boolean didwin = false;
		for (Statement st=meth.firstStatement(); st!=null; st=st.next()) {

			// If ending multi-line window, clear count of lines in window
			if (st.userDefined==null)
			{
				inwin = 0;
				continue;
			}

			// If starting subsequent multi-line window, output separator line
			if (didwin && inwin==0 && preWin>0)
			{
				if (datalog)
					log.writeSeparator();
				else log.writeEntry("----");
			}
			inwin += 1;
			didwin = true;

			// Print number & stmt. Mark ref lines with * if multi-line windows
			String numstr = String.valueOf(st.getNumber());
			if (datalog)
			{
				if (st.userDefined==ref)
					log.writeRef(numstr);
				else if (st.userDefined==localDef)
					log.writeRefLocal(numstr);
				else log.writeRefNear(numstr);
			}
			else {
				char refmark = (st.userDefined!=near && preWin>0) ? '+' : ' ';
				numstr = "#" + numstr + refmark;
				String label = Jtu.toPaddedLabel(mname, numstr, 0, 13);
				log.writeEntry(label + st.toIndent(2, 2, 50) + st.toString());
			}
			matches += 1;
			st.userDefined = null;
		}									// End loop
		meth.userDefined = null;
	}

	// Print definition line for specified member if matches. Then
	// scan appropriate classes for references to this member,
	// printing a window around each reference that is found.
	private void processMemberDef(Member mem)
	{
		// If it doesnt match a name in the list, ret immed.
		if (Jtu.isMatchFor(findIt, exact, mem)<0)
			return;

		// If not displaying by def, mark for atEndOf(ClassFile).
		ClassFile cf = mem.getClassFile();
		if (!byDef)
		{
			mem.userDefined = ref;
			cf.userDefined = near;
			return;
		}

		// Separate defs by blank line if showing refs with each def
		if (didLine && postWin>0)
			log.writeEntry();
		didLine = true;

		// Print the location and name of the matching member.
		// Display refs to it if user asked for them.
		if (datalog)
		{
			if (mem instanceof Field)
				log.writeField(mem.toObjectId());
			else log.writeMethod(mem.toObjectId());
			matches += 1;
		}
		else printDef(cf.toName() + ": " + mem.toString());
		if (postWin <= 0) return;
		doReferences(mem);
	}									// End of processMemberDef

	// Split ;-sep terms into vector, returning vector (or null of terms empty)
	private Vector separate(String terms)
	{
		Vector vec = null;
		int cp[] = new int[2];

		// Add each search term to findIt vector
		while (true) {
			String name = Trek.nextItem(terms, cp);
			if (name==null) break;
			if (vec==null) vec = new Vector();
			vec.addElement(name);
		}
		return vec;
	}

}
