//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

import dec.trek.*;
import java.io.*;
import java.util.*;

//
//
// members. Display all members (imbedded & inherited) of each target class.
//
//
public class members extends Trek
{
	private PrintWriter outChan;
	private Hashtable hash;		// Keep track of new member names
	private Vector fld;			// Info for each field
	private Vector meth;		// Infor for each method
	private int incFlags = -1;	// Process member if it has any of these flags
	private int excFlags = 0;	// Exclude member if it has any of these flags

	// Process members-utility command-line stuff
	protected int getCmdLineArg(String argv[], int i)
	{
		// Process one of the common args if present
		int newi = Jtu.getCmdLineArg(argv, i);
		if (newi > 0) return newi;		// Ate this arg

		// Chk for include flags
		if (argv[i].equals("-inc"))
		{
			if (i>argv.length-2)
				throw new TrekException("No value specified for -inc");
			incFlags = getModifiers(argv[i+1]);
			return i + 2;
		}

		// Chk for exclude flags
		if (argv[i].equals("-exc"))
		{
			if (i>argv.length-2)
				throw new TrekException("No value specified for -exc");
			excFlags = getModifiers(argv[i+1]);
			return i + 2;
		}
		return 0;					// Not a custom arg
	}

	// Recursively scan class (and its supers) for unseen-as-yet members
	// class = class to scan
	// from = the context string to use in describing members of this class
	public void getMembers(ClassFile cf)
	{
		hash = new Hashtable(20);
		fld = new Vector(10, 10);
		meth = new Vector(10, 10);
		getMembers(cf, ":");
	}

	// Do the work for a class and recurse if needed
	// Support an [implements int-name] clause on method-output line???
	private void getMembers(ClassFile cf, String from)
	{
		// Scan current classfile's fields
		for (Field field=cf.firstField(); field!=null; field=field.next()) {
			if (field.hasModifier(excFlags)) continue;
			if (isNewName(field.toName()) && field.hasModifier(incFlags))
				fld.addElement(from + field.toString());
		}

		// Scan current classfile's methods
		// Feed full dcl string to isNewName because mult methods can have
		// same name (ie. names can be overloaded via diff arg lists)
		for (Method method=cf.firstMethod(); method!=null; method=method.next()) {
			if (method.hasModifier(excFlags)) continue;
			String str = method.toString();
			if (isNewName(str) && method.hasModifier(incFlags))
				meth.addElement(from + str);
		}							// End methods

		// Recurse if more levels to go (unless to java.lang.Object which
		// would be too redundant to include in each class's interface)
		String supname = cf.toSuperClass();
		if (supname==null || supname.equals("java.lang.Object"))
			return;

		// Access the super class's classfile
		ClassFile sup;
		try {
			sup = getClassFile(supname);
		}
		catch (Throwable th) {
			System.out.println("Could not get super class of " + cf.toName()
				+ " because:");
			System.out.println(toErrorMessage(th));
			return;
		}
		getMembers(sup, "From " + supname + ":");
	}

	// If new name, add to table and return true. Else return false
	private boolean isNewName(String name)
	{
		if (hash.containsKey(name))
			return false;
		hash.put(name, name);
		return true;
	}

	public static void main(String argv[])
	{
		members mtrek = new members();
		try {
			mtrek.getCmdLine(argv);
			mtrek.printMembers();
		}
		catch (Throwable e) {
			System.out.println(mtrek.toErrorMessage(e));
			System.exit(1);
		}
	}

	// Do the work of creating exploded list of each class's members
	private void printMembers()
	{
		outChan = Jtu.openReportFile();
		initTrek();

		// Scan each class, building info to describe all its target
		// members, including those from its super classes.
		// Write out class & all the exploded fields & members
		for (ClassFile cf=firstClassFile(); cf!=null; cf=cf.next()) {
			getMembers(cf);
			writeMembers(cf);
		}								// End outer loop
		endTrek();
		outChan.close();
	}									// End main

	// Write all gathered info for specified class
	private void writeMembers(ClassFile cf)
	{
		outChan.println();
		outChan.println(cf.toString());
		writeMembers(fld);
		if (fld.size()>0) outChan.println();
		writeMembers(meth);
	}

	
	// Write all the (exploded) members stored in this vector
	private void writeMembers(Vector vec)
	{
		String leader;
		String from;
		String currfrom = "";

		// Scan thru members, outputting each with proper indenting
		for (int i=0; i<vec.size(); i++) {
			String str = (String)vec.elementAt(i);
			int sepidx = str.indexOf(':');

			// If it comes from a super class, indent more. If
			// this is start of new super class, put out hdr line.
			if (sepidx > 0)
			{
				leader = "    ";
				from = str.substring(0, sepidx+1);
				if (!currfrom.equals(from))
				{
					outChan.println("  " + from);
					currfrom = from;
				}
			}

			// Imbedded member
			else {
				leader = "  ";
				from = "";
			}

			// Output the member line
			outChan.println(leader + str.substring(sepidx+1));
		}
	}
}

