//
// THIS SOFTWARE IS FURNISHED UNDER A LICENSE AND  MAY
// BE USED OR COPIED ONLY IN ACCORDANCE WITH THE TERMS
// OF SUCH LICENSE.
//
//
// COPYRIGHT (c) 1997 BY DIGITAL EQUIPMENT CORPORATION.
// ALL RIGHTS RESERVED.
//

import dec.trek.*;
import java.io.*;
import java.util.Vector;

//
//
// watch. Display each change to the specified fields
//
//
public class watch extends Trek
{
	// Method entry/exit watching can be side-effect of other watching.
	// These values (of userDefined) control processing in watchMethod.
	// ==touched means watching something in method
	// ==tracing means tracing of method already setup
	private final static Integer touched = new Integer(0);
	private final static Integer tracing = new Integer(1);

	// State avail to caller
	public String firstError;			// 1st exception inserting a call
	public int showFails;
	public int watchPts;
	public int watchFails;

	// Watch request/state
	private Call call;					// Curr call being built
	private boolean datalog;			// Gen live data in watch args
	private boolean exact;				// True says do exact match
	private int myIndex;				// Index into counters for curr loop
	private int context;				// Context options to gen info for
	private boolean quiet;				// True says dont display msg when 
										//  cant insert Condition
	private boolean real;				// True says makec chk, false says just chk
	private String trekId;				// For restoring trek of data log
	private String watchArg;			// The full watch syntax
	private Vector fields;				// Match terms for fields
	private Vector methods;				// Match terms for methods/locals

	// Stuff to support var-display and conditions, indexed by watch term
	private String show[];
	private String ifVar[];
	private int ifOprs[];
	private String ifVal2[];
	private int trace[];

	// Context options (Note: entexit implied by all the others)
	public static final int ctx_entexit = 1;	// Method entry/exit
	public static final int ctx_argret = 2;	// Arg values & ret val
	public static final int ctx_stack = 4;	// Stack trace
	public static final int ctx_thread = 8;	// Name/priority of thread
	public static final int ctx_this = 16;	// this obj's fields

	// Trace-method options (for argret/this, ctx_ or trm_ on does it)
	public static final int trm_argret = ctx_argret;
	public static final int trm_assign = 4;	// Assignments (incl for init/incr)
	public static final int trm_control = 8;	// Control stmts (eg. if, switch)
	public static final int trm_this = ctx_this;

	// If there are fields to look for, scan this classfile.
	// For perf, do pool 1st in hopes fields are not reffed at all.
	protected void atStartOf(ClassFile cf)
	{
		if (fields.size()<=0)
			return;

		// Look for pool refs in classfile to fields. If have any,
		// look for the insts in each method of this classfile that mod fields
		for (FieldRef ref=cf.firstFieldRef(); ref!=null; ref=ref.next()) {
			TypeRef tref = ref.getTypeRef();
			if (Jtu.isMatchFor(fields, exact,
					tref.toQualifiedName(), ref.toName(), true)>=0)
			{
				for (Method meth=cf.firstMethod(); meth!=null; meth=meth.next()) {
					scanMethod(meth, fields);
				}
				break;
			}
		}
	}

	// Insert calls for watching flow of whole method if it is watchlist
	// Notes:
	//	Done at end so possible to do touching watchMethod only here
	protected void atEndOf(Method meth)
	{
		// Trace method if it is in watchlist. Do changes of locals if any.
		if (methods.size()>0)
		{
			int idx = Jtu.isMatchFor(methods, exact, meth);
			if (idx >= 0)
				traceMethod(meth, trace[idx]);

			// If one or more local terms match this method, scan the method.
			// Must be 2nd since var-level watching subsumes any assignment stmt
			// watching done as part of trace method.
			if (Jtu.hasMatch(methods, exact, meth, null) >= 0)
				scanMethod(meth, methods);
		}

		// Do method entry/exit if method touched (and tracing not done)
		if (meth.userDefined==touched)
			watchMethod(meth, 0);
		meth.userDefined = null;
	}

	// Look for change of matching variable in this instruction
	// inst = instruction to chk
	// watchit = the match terms to chk
	private void check(Instruction inst, Vector watchit)
	{
		// Must be store-type inst or iinc for change to be possible
		int type = inst.getType();
		if (type!=IN_STOREFIELD && type!=IN_STORELOCAL && type!=IN_STOREELEMENT
			&& inst.getOpcode()!=Code.iinc)
			return;

		// If this inst matches a watch term, instrument it. Note: no risk 
		// of matching method in watchit cause stores only ref locs or flds
		int idx = Jtu.isMatchFor(watchit, exact, inst);
		if (idx >= 0)
			watchInst(inst, idx);
	}

	// Ditto, for specific local
	// loc = thing looking for
	// idx = support info for this loc
	private void check(Instruction inst, Local loc, int idx)
	{
		// Must be store-type inst or be iinc instruction,
		int type = inst.getType();
		if (type!=IN_STORELOCAL && type!=IN_STOREELEMENT
			&& inst.getOpcode()!=Code.iinc)
			return;

		// Must match local be looked for
		if (loc==inst.getLocal())
			watchInst(inst, idx);
	}

	// Support doing dry run just to validate user's input
	private void done()
	{
		if (real)
			call.done();
		else call.discard();
	}

	// Do the scan to insert the watch-calls
	// watcharg = what to watch
	// context = what context info to gather across the board
	// datalog = true says live data (ie. object IDs rather than readable text)
	// real = true sets set watchpts, false sets just verify settable
	// exact = true says use exact case matching names to watch
	// quiet = true says dont display errors (caller user *Fails & firstError)
	public void doWatch(String watcharg, int context,
		boolean datalog, boolean real, boolean exact, boolean quiet)
	{
		if (watcharg==null || watcharg.length()==0)
			throw new TrekException("No watchpoints specified");

		// Init activity ctrs
		firstError = null;
		watchPts = 0;
		watchFails = 0;
		showFails = 0;

		// Set general state vars
		this.watchArg = watcharg;
		this.context = context;
		this.exact = exact;
		this.quiet = quiet;
		this.real = real;
		this.datalog = datalog;

		// Show/condition state
		show = new String[20];
		trace = new int[20];
		ifVar = new String[20];
		ifOprs = new int[20];
		ifVal2 = new String[20];

		// Auto-scan the classfiles and methods. Note that scan is done
		// even if fields/methods empty because specific-object watchpts
		// dont do the side-effects, only scan does that
		parseWatch();
		trekId = datalog ? toObjectId() : null;
		//if (fields.size()>0 || methods.size()>0)
			doTrek(2);
	}

	// Pick up the (possibly wildcarded) name to watch
	public int getCmdLineArg(String argv[], int idx)
	{
		if (argv[idx].startsWith("-"))
		{
			// Chk for args specifically for this app
			if (argv[idx].equals("-cs"))
			{
				exact = true;
				return idx + 1;
			}
			else if (argv[idx].equals("-quiet"))
			{
				quiet = true;
				return idx + 1;
			}

			// -include
			if (argv[idx].equals("-inc"))
			{
				if (idx>argv.length-2)
					throw new TrekException("No value specified for -inc");
				String arg = argv[idx+1];

				// Convert arg i+1 to binary flags
				context = 0;
				if (arg.indexOf('e') >= 0) context |= ctx_entexit;
				if (arg.indexOf('a') >= 0) context |= ctx_argret;
				if (arg.indexOf('s') >= 0) context |= ctx_stack;
				if (arg.indexOf('h') >= 0) context |= ctx_thread;
				if (arg.indexOf('t') >= 0) context |= ctx_this;
				return idx + 2;			// Ate 2 args overall
			}
			else return 0;					// Some other arg
		}

		// Save watch terms for parsing
		watchArg = argv[idx];
		return idx + 1;
	}

	// Insert an if-test before call if so requested
	// meth = method of ins pt
	// idx = index of matched watch term
	// call = call object for call being inserted at ins pt
	// if* = describes condition to insert
	private void insertIfTest(Method meth, int idx)
	{
		String var1 = ifVar[idx];
		if (var1==null) return;
		call.doIf(var1, ifOprs[idx], ifVal2[idx]);
	}


	// Do the scan to insert the watch-calls
	public static void main(String argv[])
	{
		Call.systemMethodsSafe = true;
		watch wtrek = new watch();
		try {
			wtrek.getCmdLine(argv);
			wtrek.initTrek();
			wtrek.doWatch(wtrek.watchArg, wtrek.context,
				false, true, wtrek.exact, wtrek.quiet);
			wtrek.endTrek();
			System.out.println(wtrek.toWatchSummary());
		}

		// Catch any error thrown by Trek package, and output it to console.
		catch (Throwable e) {
			System.out.println(wtrek.toErrorMessage(e));
			System.exit(1);
		}
	}

	// Return method-name based prefix to use if applic
	private String methodName(Statement st)
	{
		String mname;
		if (context==0 || context==ctx_entexit)
			mname = Jtu.toPaddedLabel(st.getMethod().toName(), "", 0, 10);
		else mname = "";
		return mname;
	}

	// Parse a condition
	// term = the watch term
	// ifidx = index of colon preceding condition
	// idx = index of watch term in watch list
	// Returns where next piece of term starts
	private int parseIf(String term, int ifidx, int idx)
	{
		String ifstr;

		// Isolate When clause
		int optidx = term.indexOf(':', ifidx+1);
		if (optidx<0)
			ifstr = term.substring(ifidx + 1);
		else ifstr = term.substring(ifidx + 1, optidx);

		// Process the expr
		String opds[] = new String[2];
		int ifopr = Jtu.parseIf(ifstr, opds);
		if (ifopr==-2)				// Empty expression
			return optidx;
		if (ifopr==-1)				// Bad expression
			throw new TrekException(
				"Invalid comparison operator in watch term " + ++idx);

		// Return calculated stuff
		ifOprs[idx] = ifopr;
		ifVar[idx] = opds[0];
		ifVal2[idx] = opds[1];
		return optidx;
	}

	// Extract options field
	private void parseOptions(String term, int optidx, int idx)
	{
		int opts;
		String traw = term.substring(optidx+1);
		try {
			opts = Integer.parseInt(traw);
		}
		catch (Exception e) {
			opts = 0;
			if (traw.indexOf('a')>=0) opts |= trm_argret;
			if (traw.indexOf('n')>=0) opts |= trm_assign;
			if (traw.indexOf('t')>=0) opts |= trm_this;
			if (traw.indexOf('c')>=0) opts |= trm_control;
		}
		trace[idx] = opts;
	}

	// Extract show list into show[idx] and return index of Condition, if any
	// term = the watch term
	// showidx = index of colon preceding show-list
	// idx = index of watch term in watch list
	private int parseShow(String term, int showidx, int idx)
	{
		String vars;

		// Isolate the show list
		int ifidx = term.indexOf(':', showidx+1);
		if (ifidx<0)
			vars = term.substring(showidx + 1);
		else vars = term.substring(showidx + 1, ifidx);

		// Return idx of : and setup show[idx] if not empty
		if (vars.length()!=0)
			show[idx] = vars;
		return ifidx;
	}

	// Setup fields/methods and related if* stuff.
	// Syntax is target;target; ..., where each target is:
	//		where:show-list:condition:options
	// Where is: possibly wildcarded name or object name of a statement
	//	form of a field name is just name, form of a method name is
	//	name(), and form of a local var name is name().name
	// Show is: comma-separated list of variable refs
	// Condition is: variable-ref op value
	//		value is a number  or  'string'  or  variable-ref.
	//		variable-ref is any legal Java var ref. But if Trek pkg doesnt
	//		support it or it is bad syntax, pass and doIf throw exceptions.
	// Options is: a coded letter sequence (see parseOptions)
	private void parseWatch()
	{
		// Break apart each watch term
		fields = new Vector();
		methods = new Vector();
		int cp[] = new int[2];
		int idx = 0;
		while (true) {
			String term = Trek.nextItem(watchArg, cp);
			if (term==null) break;
			int termlen = term.length();
			trace[idx] = trm_control;
			ifVal2[idx] = show[idx] = null;

			// If no show clause, whole term is thing to watch.
			// Else append up to :, and isolate other clauses
			String watch;
			int showidx = term.indexOf(':');
			if (showidx < 0)
				watch = term;
			else {
				watch = term.substring(0, showidx);
				int ifidx = parseShow(term, showidx, idx);
				if (ifidx>=0)
				{
					int optidx = parseIf(term, ifidx, idx);
					if (optidx>=0) parseOptions(term, optidx, idx);
				}
			}

			// If term is for a specific obj, just process it with no scan.
			// (Scan avoided and 2 types of scan purely to get max perf)
			Object obj = getObject(watch);
			if (obj!=null)
			{
				if (obj instanceof Method)
				{
					traceMethod((Method)obj, trace[idx]);
					continue;
				}
				if (obj instanceof Statement)
				{
					watchStmt((Statement)obj, idx);
					continue;
				}
				if (obj instanceof Local)
				{
					watchLocal((Local)obj, idx);
					continue;
				}
			}

			// 2 types of scan:
			// 1) field usage in any classfile that refs the fld
			// 2) method defs for wild methods or locals
			// Have null items so can use common support arrays
			if (watch.indexOf('(')>0)
			{
				methods.addElement(watch);
				fields.addElement(null);
			}
			else {
				fields.addElement(watch);
				methods.addElement(null);
			}
			idx += 1;
		}							// End loop
	}								// End parseWatch

	// Pass method-context plus specified stmt[s] source code or stmt-num
	// Indent +1 for each level
	private void passSource(Statement st) {passSource(st, null);}
	private void passSource(Statement st, Statement stn)
	{
		if (datalog)
			call.passString(String.valueOf(st.getNumber()));
		else call.passString(methodName(st)
			+ st.toIndent(1, 1, 40) + st.toString()
			+ (stn!=null ? stn.toString() : ""));
	}

	// Scan a method for insts that matches the specified watch terms
	// meth = method to scan
	// watchit = list of code object name patterns to look for
	private void scanMethod(Method meth, Vector watchit)
	{
		for (Statement st=meth.firstStatement(); st!=null; st=st.next()) {
			for (Instruction in=st.firstInstruction(); in!=null; in=in.next()) {
				check(in, watchit);
			}
		}
	}

	// Return string describing watchpts set and failures
	public String toWatchSummary()
	{
		String wp = watchPts==1 ? " watchpoint" : " watchpoints";
		String summ = "Inserted " + watchPts + wp;
		if (watchFails==1)
			return summ + " (with 1 failure)";
		else return summ + " (with " + watchFails + " failures)";
	}

	// Insert tracing for all the control points in specified method
	// opts = level of tracing to do
	// Notes:
	//	Return-stmts are handled by watchMethod
	private void traceMethod(Method meth, int opts)
	{
		// Dont setup tracing for a method more than once
		if (meth.userDefined==tracing)
			return;

		// Scan prog to setup watching of control logic
		for (Statement st=meth.firstStatement(); st!=null; st=st.next()) {
			int type = st.getType();

			// Assignments
			if (type==ST_ASSIGN)
			{
				if ((opts&trm_assign)==0) continue;
				Instruction inst = st.lastInstruction();
				if (inst!=null)	watchInst(inst, -1);
				continue;
			}

			// Control actions
			if ((opts&trm_control)==0) continue;
			switch (type) {
			case ST_FOR:
			case ST_WHILE:
			case ST_DO:
				Statement bot;

				// If already have loop var watchpt (which would actually be
				// in loop-end stmt), dont redundantly setup printWhile wpt
				int looplev = st.getLevel();
				for (bot=st.next(); bot.getLevel()!=looplev; bot=bot.next());
				if (bot.hasChange("val-*"))
				{
					watchFirstIter(st);
					break;
				}

				// Setup to watch this loop
				call = Call.addBefore("while:WatchLog.printWhile", st.next());
				if (call==null) continue;
				call.passInt(myIndex++);
				passSource(st);
				done();
				break;

			// A number of cases because Else's "optional" & optim goto's
			case ST_IF:
				Statement stn;
				call = Call.addAfter("WatchLog.printIfResult", st, true);

				// No Then path occurs when If-inst itself does break/contin.
				// So just display "past if" rather than if-false since code
				// can possibly reach "past if" without going thru If.
				if (call==null)
				{
					stn = st.next();
					call = Call.addBefore("WatchLog.printIfResult", stn.next());
					call.passInt(2);
					passSource(st, stn);	// Present if (xxx) brk/contin
					done();
					break;
				}

				// A real-Then, display True when it is reached
				call.passInt(1);
				passSource(st);
				done();

				// Insert code on Else-path if any, unless it's 1st code is
				// If. In that case dont because the If will show flow anyway
				stn = st.firstOfElse();
				if (stn==null || stn.getType()==Trek.ST_IF)
					break;
				call = Call.addAfter("WatchLog.printIfResult", st, false);
				if (call==null)
					break;
				call.passInt(0);
				passSource(st);
				done();
				break;

			case ST_SWITCH:
				call = Call.addAfter("WatchLog.printSwitch", st);
				if (call==null) continue;
				call.pass();
				passSource(st);
				done();
				break;

			// Other control pts
			//case ST_TRY:
			case ST_CATCH:
			case ST_FINALLY:
				call = Call.addAfter("WatchLog.printStatement", st);
				if (call==null) continue;
				passSource(st);
				done();
				break;
			}
		}								// End loop

		// Do method-level trace and context options & mark that it was done
		watchMethod(meth, (opts&(trm_argret|trm_this)) | ctx_entexit);
		watchPts += 1;
		meth.userDefined = tracing;
	}

	// Process a watch failure
	private void watchFail(String action, Statement st, Exception e)
	{
		// Discard call and classify error
		call.discard();
		boolean softerr = e instanceof TrekException;
		String reason = softerr ? e.getMessage() : e.toString();

		// Display error
		if (!quiet)
		{
			System.out.println("Cannot " + action + " at "
				+ st.toLocation() + " because:");
			if (softerr)
				System.out.println("  " + reason);
		}

		// Save error for user potentially
		if (firstError==null)
			firstError = "Cannot " + action + " because: " + reason;

		// Show bug in console window regardless
		if (!softerr)
			e.printStackTrace();
	}

	// If 1st iter of loop would not be traced -- because value changes are
	// being watched and the loop has no loop var, add disp of loop top
	// top = the loop top to chk
	private void watchFirstIter(Statement top)
	{
		if (!top.hasChange(null))
		{
			call = Call.addBefore("val-t:WatchLog.printStatement", top);
			passSource(top);
			done();
		}
	}

	// Watch variable modified by specified instruction
	// inst = instruction that matched watch term
	// idx = term it matched (or -1 if part of traceMethod
	private void watchInst(Instruction inst, int idx)
	{
		Statement st = inst.getStatement();
		Method meth = st.getMethod();

		// If already tracing or watching, dont need to do anything.
		// If watching and already have watchpt, ditto.
		if (idx<0 && st.hasChange("val-*")
			|| (idx>=0 && st.hasChange("val-w")))
			return;

		// Watchpts subsume tracing of assignment stmts. Conversely
		// mark this change as either watch or trace
		String chid;
		if (idx>0)
		{
			chid = "val-w:";
			st.undoChange("val-t*");
		}
		else chid = "val-tr:";

		// Start building watch-method call for either field or local.
		// Do if-test if requested.
		call = Call.addAt(chid + "WatchLog.printValue", inst);
		if (call==null) return;
		try {
			if (idx>=0)
				insertIfTest(meth, idx);

			// Pass variable's new value, and array index if at array element
			call.pass();
			int type = inst.getType();
			if (type==IN_LOADELEMENT || type==IN_STOREELEMENT)
				call.passIndex();
		}

		// Context or user error. Report it unless quiet switch was specified
		catch (Exception e) {
			watchFail("insert watchpoint", st, e);
			watchFails += 1;
			return;
		}

		// Source code of For and While at start of loop. (While can change
		// variable via something like: while (x++<10))
		Statement src;				// Source for where this change is
		int type = st.getType();
		if (type==ST_ENDFOR || type==ST_ENDWHILE)
		{
			int looplev = st.getLevel();
			for (src=st.prev(); src.getLevel()!=looplev; src=src.prev());

		}
		else src = st;

		// Pass source code assoc with this stmt, and complete the change
		passSource(src);
		done();

		// if st is loop bottom, the above printValue subsumes loop-tracing
		// if it was set. However if the loop has no loop var, the 1st iter of
		// the loop would not be displayed. So add display of loop start stmt.
		if (src!=st && src.next().undoChange("while"))
			watchFirstIter(src);

		// If primary watch pt (as opposed to part of trace),
		// do its Shows and mark that method has been at least touched
		if (idx<0) return;
		watchValues(idx, st, inst);
		watchPts += 1;
		if (meth.userDefined==null)
			meth.userDefined = touched;
	}

	// Look thruout loc's method for insts that modify this local
	// loc = local to watch
	// idx = index of its related info (eg. show's)
	private void watchLocal(Local loc, int idx)
	{
		Method meth = loc.getMethod();
		for (Statement st=meth.firstStatement(); st!=null; st=st.next()) {
			for (Instruction in=st.firstInstruction(); in!=null; in=in.next()) {
				check(in, loc, idx);
			}
		}
	}

	// Insert specified tracing if havent yet
	// meth = method to watch
	// opts = watch context options specific to this method
	// Notes:
	//	Thread and stack context options involve no work beyond passing
	//	context. All the real work is at runtime
	private void watchMethod(Method meth, int opts)
	{
		// Chk if already inserted entry
		if (meth.userDefined==tracing)
			return;
		String mname = meth.toName();

		// If any method-level info wanted, insert entry/exit calls
		opts |= context;
		if (opts==0)
			return;

		// Output call for method entry
		call = Call.addBefore("WatchLog.enterMethod", meth);
		if (call==null) return;
		if (datalog)
			call.passString(meth.toObjectId());
		else call.passString(mname);
		call.passString(trekId);
		call.passInt(context);
		done();

		// Output arg-value info when method is entered.
		// Note that "this" is by convention the first local.
		// Then comes all the method's user-visible args.
		// Build info string by going thru args one by one.
		Local arg0 = meth.firstLocal();
		if ((opts&ctx_argret)!=0)
		{
			for (Local arg=arg0; arg!=null; arg=arg.next()) {
				if (arg.isThis()) continue;
				if (!arg.isArg()) break;
				call = Call.addBefore("WatchLog.argValue", meth);
				if (call==null) continue;
				call.pass(arg.toName());
				call.passString(arg.toName());
				done();
			}
		}

		// If there is "this", watch it and related stuff if so requested
		if (arg0!=null && arg0.isThis())
		{
			if ((opts&(ctx_argret|ctx_this))!=0)
			{
				call = Call.addBefore("WatchLog.argValue", meth);
				if (call!=null)
				{
					call.pass(arg0.toName());
					call.passString(arg0.toName());
					done();
				}
			}

			// Watch this.* if so requested (and not final or array)
			if ((opts&ctx_this)!=0)
			{
				ClassFile cf = meth.getClassFile();
				for (Field fld=cf.firstField(); fld!=null; fld=fld.next()) {
					if (fld.isArray() || fld.hasModifier(Trek.ACC_FINAL))
						continue;

					// Put out the call unless there is problem
					call = Call.addBefore("WatchLog.argValue", meth);
					if (call==null) continue;
					String name = fld.toName();
					call.pass(name);
					call.passString("  " + name);
					done();
				}
			}
		}

		// Output line for each method exit
		boolean isvoid = meth.toType().equals("void");
		for (Statement st=meth.firstStatement(); st!=null; st=st.next()) {
			String hack;
			int type = st.getType();
			if (type!=ST_RETURN && type!=ST_THROW)
				continue;

			// At a Return or throw stmt. Pass its val if any & desired.
			// Put encoded 1st char (& method name) in context arg so
			// WatchLog.printValue knows this is exit method.
			// (Done to avoid 7 more methods/data type for exiting method)
			if ((opts&ctx_argret)!=0
				&& (!isvoid || type==ST_THROW))
			{
				call = Call.addAfter("WatchLog.printValue", st);
				if (call==null) continue;
				call.pass();
				call.passString("?" + mname);
			}

			// No retval or not requested
			else {
				call = Call.addAfter("WatchLog.exitMethod", st);
				if (call==null) continue;
				call.passString(mname);
			}
			done();
		}
	}									// End watchMethod

	// Watch (before) a statement
	// st = statement that matched watch term
	// idx = term it matched
	private void watchStmt(Statement st, int idx)
	{
		st.undoChange("val-ti1");		// Iter 1 tracing redundant now
		call = Call.addBefore("WatchLog.printStatement", st);
		if (call==null) return;
		try {
			insertIfTest(st.getMethod(), idx);
		}

		// Context or user error. Report it unless quiet switch was specified
		catch (Exception e) {
			watchFail("insert watchpoint", st, e);
			watchFails += 1;
			return;
		}

		// Complete outputting of stmt source
		passSource(st);
		done();

		// Output any vals being watched and mark that watchpt done here
		watchValues(idx, st, null);
		Method meth = st.getMethod();
		if (meth.userDefined==null)
			meth.userDefined = touched;
		watchPts += 1;
	}

	// Display a specific set of values at the watchpt
	// idx = watch term it is for
	// st/inst = one or other defines location of watchpt
	private void watchValues(int idx, Statement st, Instruction inst)
	{
		boolean failed = false;
		Method meth = st.getMethod();
		String vars = show[idx];
		if (vars==null) return;

		// Display vars are specified, sep them
		int cp[] = {0, ','};
		while (true) {
			String var = Trek.nextItem(vars, cp);
			if (var==null) return;

			// Initiate call
			if (inst!=null)
				call = Call.addAt("WatchLog.printValue", inst);
			else call = Call.addBefore("WatchLog.printValue", st);
			if (call==null) continue;

			// Pass the variable
			try {
				insertIfTest(meth, idx);
				call.pass(var);
			}

			// Context or user error. Report it unless quiet switch was specified
			catch (Exception e) {
				watchFail("display " + var, st, e);
				if (!failed)
				{
					failed = true;
					showFails += 1;
				}
				continue;
			}

			// Pass its context and term the call
			if (datalog)
				call.passString(var);
			else call.passString(st.toIndent(1, 1, 40) + "  SHOW " + var);
			done();
		}
	}										// End watchValues
}

